Package: randomForestRHF
Version: 1.0.1
Date: 2026-04-27
Title: Random Hazard Forests
Authors@R: c(person("Hemant", "Ishwaran", email = "hemant.ishwaran@gmail.com", role = "aut"),
  person("Udaya B.", "Kogalur", email = "ubk@kogalur.com", role = c("aut", "cre")))
Author: Hemant Ishwaran [aut],
  Udaya B. Kogalur [aut, cre]
Maintainer: Udaya B. Kogalur <ubk@kogalur.com>
Depends: R (>= 4.3.0),
Imports: randomForestSRC (>= 3.3.1), varPro (>= 3.0.0), survival
Suggests: mlbench, interp, glmnet
SystemRequirements: OpenMP
Description: Random Hazard Forests (RHF) extend Random Survival
  Forests (RSF) by directly estimating the hazard function and by
  accommodating time-dependent covariates through counting-process
  style inputs. The package fits tree ensembles for dynamic survival
  prediction, returning hazard, cumulative hazard, integrated hazard,
  and related performance summaries for training and test data. The
  methods build on Random Survival Forests described by Ishwaran et
  al. (2008) <doi:10.1214/08-AOAS169> and on nonparametric hazard
  modeling with time-dependent covariates described by Lee et
  al. (2021) <doi:10.1214/20-AOS2028>.
License: GPL (>= 3)
URL: https://www.randomforestsrc.org/ https://ishwaran.org/
NeedsCompilation: yes
Packaged: 2026-04-27 21:39:07 UTC; kogalur
Repository: CRAN
Date/Publication: 2026-04-28 08:10:03 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-28 10:57:40 UTC; unix
Archs: randomForestRHF.so.dSYM

Package: randomForestRHF
Version: 1.0.1
Date: 2026-04-27
Title: Random Hazard Forests
Authors@R: c(person("Hemant", "Ishwaran", email = "hemant.ishwaran@gmail.com", role = "aut"),
  person("Udaya B.", "Kogalur", email = "ubk@kogalur.com", role = c("aut", "cre")))
Author: Hemant Ishwaran [aut],
  Udaya B. Kogalur [aut, cre]
Maintainer: Udaya B. Kogalur <ubk@kogalur.com>
Depends: R (>= 4.3.0),
Imports: randomForestSRC (>= 3.3.1), varPro (>= 3.0.0), survival
Suggests: mlbench, interp, glmnet
SystemRequirements: OpenMP
Description: Random Hazard Forests (RHF) extend Random Survival
  Forests (RSF) by directly estimating the hazard function and by
  accommodating time-dependent covariates through counting-process
  style inputs. The package fits tree ensembles for dynamic survival
  prediction, returning hazard, cumulative hazard, integrated hazard,
  and related performance summaries for training and test data. The
  methods build on Random Survival Forests described by Ishwaran et
  al. (2008) <doi:10.1214/08-AOAS169> and on nonparametric hazard
  modeling with time-dependent covariates described by Lee et
  al. (2021) <doi:10.1214/20-AOS2028>.
License: GPL (>= 3)
URL: https://www.randomforestsrc.org/ https://ishwaran.org/
NeedsCompilation: yes
Packaged: 2026-04-27 21:39:07 UTC; kogalur
Repository: CRAN
Date/Publication: 2026-04-28 08:10:03 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-05-01 15:28:12 UTC; unix
Archs: randomForestRHF.so.dSYM

## Dynamic C-code Library.
useDynLib(randomForestRHF, .registration = TRUE)

importFrom("parallel", mclapply)
importFrom("randomForestSRC", "rfsrc")

importFrom("grDevices", "adjustcolor")

importFrom("graphics", "legend", "lines", "par", "rug", "mtext", "points")
importFrom("graphics", "matplot", "boxplot", "abline", "polygon")


importFrom("stats", "runif", "as.formula", "formula", "na.omit", "model.matrix", "sd", "var", "supsmu")
importFrom("stats", "quantile", "rbinom", "setNames", "lowess", "uniroot")
importFrom("stats", "complete.cases")
importFrom("stats", "plogis")

importFrom("survival", "Surv")

importFrom("utils", "tail", "combn", "capture.output", "head", "flush.console")

importFrom("varPro", "varpro", "get.orgvimp", "get.vimp", "cv.varpro",
           "varpro.strength", "get.varpro.strengthArray",
           "local.importance", "importance.varpro.workhorse")

export(as.data.frame.importance.rhf,
       auct.rhf,
       auct,
       convert.counting,
       convert.standard.counting,
       dotmatrix.importance.rhf,
       dotmatrix.importance,
       hazard.simulation,
       importance.rhf,
       plot.auct.rhf,
       plot.importance.rhf,
       plot.rhf,
       plot.tune.treesize.rhf,
       predict.rhf,
       print.auct.rhf,
       print.importance.rhf,
       print.rhf,       
       rhf,
       rhf.news,
       smoothed.hazard.rhf,
       smoothed.hazard,
       tune.treesize.rhf,
       tune.rhf,
       tune.iAUC.rhf,
       tune.iAUC,
       varpro.cache.rhf,
       varpro.cache,
       xvar.wt.rhf)

S3method(as.data.frame, importance.rhf)

S3method(plot, auct.rhf)
S3method(plot, importance.rhf)
S3method(plot, rhf)
S3method(plot, tune.treesize.rhf)


S3method(predict, rhf)

S3method(print, auct.rhf)
S3method(print, importance.rhf)
S3method(print, rhf)

predict.rhf.workhorse <-  function(object,
                                   newdata,
                                   get.tree = NULL,
                                   block.size = NULL,
                                   seed = NULL,
                                   membership = FALSE,
                                   do.trace = FALSE,
                                   ...)
{
  ## confirm this is a rhf object
  if (!inherits(object, "rhf")) {
    stop("this function only works for objects of class 'rhf'")
  }
  ## hidden options (used later)
  dots <- list(...)
  ## initialize the seed
  seed <- get.seed(seed)
  ## uniform versus endpoint estimation hazard
  experimental.bits <- get.experimental.bits(dots$experimental.bits, FALSE)
  ## hard coded options (some are legacy)
  perf.type <- NULL
  ## set restore.mode and the ensemble option
  if (missing(newdata)) {##restore: no data
    restore.mode <- TRUE
    ensemble <- "all"
  }
  else {##not restore: test data present
    restore.mode <- FALSE
    ensemble <- "inbag"
  }
  ## check if this is an anonymous object and process accordingly
  if (inherits(object, "anonymous")) {
    anonymize.bits <- 2^26
    if (restore.mode) {
      stop("in order to predict with anonymous forests please provide a test data set")
    }
  }
  else {
    anonymize.bits <- 0
  }
  ## set the family
  family <- object$family
  subj  <- object$id
  subj.unique.count <- length(unique(subj))
  case.wt  <- get.weight(NULL, subj.unique.count)
  event.info <- object$event.info
  ## get time-map information
  max.time <- object$max.time 
  time.map <- object$time.map
  if (is.null(time.map)) {
    time.map <- list(method = "legacy",
                     max.time = as.double(max.time),
                     tau = as.double(max.time))
  }
  ## obtain y-outcome information from the grow object; scale y to [0,1]
  yvar <- .iscale.yvar(object$yvar, time.map)
  yvar.names <- object$yvar.names
  yvar.types <- object$yvar.types
  yvar.nlevels  <- object$yvar.factor$nlevels
  yvar.nlevels  <- NULL
  ## obtain x information from the grow object 
  xvar <- object$xvar
  xvar.names <- object$xvar.names
  xvar.types <- object$xvar.types
  xvar.nlevels  <- rep(0, length(xvar.types))
  xvar.nlevels  <- NULL
  ## get the subject name
  subj.names  <- object$subj.names  
  ## hard coded options
  xvar.augment.newdata <- augmentXlist <- hcutCnt <- NULL
  hcut  <- 1
  ## recover the split information
  splitinfo <- object$splitinfo
  splitrule <- object$splitinfo$splitrule
  ## get the subjects -- this might be the test subjects eventually TBD
  id = object$id  
  ## initialize the seed
  seed <- get.seed(seed)
  ## REASSIGN OBJECT: hereafter we only need the forest 
  object <- object$forest
  ## confirm version coherence
  if (is.null(object$version)) {
      stop(
          paste(
              "This function only works with objects created with the following minimum version of the package:",
                     "  Minimum version:  0.0.0.0",
              paste0("  Your version:     unknown"),
              sep = "\n"
          ),
          call. = FALSE
      )
  }
  else {
    object.version <- as.integer(unlist(strsplit(object$version, "[.]")))
    installed.version <- as.integer(unlist(strsplit("1.0.1", "[.]")))
    minimum.version <- as.integer(unlist(strsplit("0.0.0.0", "[.]")))
    object.version.adj <- object.version[1] + (object.version[2]/10) + (object.version[3]/100)
    installed.version.adj <- installed.version[1] + (installed.version[2]/10) + (installed.version[3]/100)
    minimum.version.adj <- minimum.version[1] + (minimum.version[2]/10) + (minimum.version[3]/100)
    ## Minimum object version must be satisfied for us to proceed.  This is the only way
    ## terminal node restoration is guaranteed, due to RNG coherency.
    if (object.version.adj >= minimum.version.adj) {
      ## We are okay
    }
    else {
      stop(
          paste(
              "This function only works with objects created with the following minimum version of the package:",
              "  Minimum version:  0.0.0.0",
              paste0("  Your version:     ", object$version),
              sep = "\n"
          ),
          call. = FALSE
      )
    }
  }
  ##--------------------------------------------------------
  ##
  ## process x and y: test data is present
  ##
  ##--------------------------------------------------------
  if (!restore.mode) {
    ## clean up test data (handle missingness, scale time, scale y)
    nd <- cleanup.counting.newdata(newdata    = newdata,
                                   xvar.names = xvar.names,
                                   yvar.names = yvar.names,
                                   subj.names = subj.names,
                                   time.map   = time.map,
                                   max.time   = max.time)
    ## overwrite with cleaned version
    newdata   <- nd$newdata
    n.newdata <- nrow(newdata)
    ## get the test subjects
    subj.newdata  <- nd$subj
    subj.newdata.unique.count <- length(unique(subj.newdata))
    ## restrict xvar to the training xvar.names
    xvar.newdata <- nd$xvar
    ## hard coded 
    xvar.augment.newdata <- NULL
    ## extract test yvar (if present); already scaled to [0,1]
    yvar.newdata <- nd$yvar
    yvar.present <- nd$yvar.present
    if (yvar.present) {
      perf.type <- get.perf(perf.type, family)
    } else {
      ## keep subj.newdata.unique.count for dimensionality,
      ## but drop subj/y for performance calculations
      subj.newdata <- yvar.newdata <- NULL
      perf.type <- "none"
    }
  }
  ##--------------------------------------------------------
  ##
  ##  process x and y: no test data is present
  ##
  ##--------------------------------------------------------
  else {
    ## There cannot be test data in restore mode
    ## The native code switches based on n.newdata being zero (0).  Be careful.
    n.newdata <- 0
    xvar.newdata <- yvar.newdata <-  subj.newdata  <- NULL
    ## perf type
    perf.type <- get.perf(perf.type, family)
  }
  ## ------------------------------------------------------------
  ##
  ## restore/non-restore x/y processing completed
  ## finalize and make C call
  ##
  ## ------------------------------------------------------------
  ## set the performance bits
  perf.bits <-  get.perf.bits(perf.type)
  ## initialize the number of trees in the forest
  ntree <- object$ntree
  ## process the get.tree vector that specifies which trees we want
  ## to extract from the forest.  This is only relevant to restore mode.
  ## The parameter is ignored in predict mode.
  get.tree <- get.tree.index(get.tree, ntree)
  bootstrap.bits <- get.bootstrap.bits(object$parms$bootstrap)
  ## initialize the low bits
  ensemble.bits <- get.ensemble.bits(ensemble)
  ## sample related
  samptype <- object$parms$samptype
  sampsize <- object$parms$sampsize
  samp <- object$parms$samp
  ## Initalize the high bits
  samptype.bits <- get.samptype.bits(samptype)
  membership.bits <-  get.membership.bits(membership)
  ##  jitt.bits <- get.jitt.bits(jitt)
  ## We over-ride block-size in the case that get.tree is user specified
  block.size <- min(get.block.size.bits(block.size, ntree), sum(get.tree))
  ## Default target. We don't support more than one, and this is a bit of a legacy issue.
  m.target.idx <- 1
  ## do.trace
  do.trace <- get.trace(do.trace)
  ## WARNING: Note that the maximum number of slots in the following
  ## foreign function call is 64.  Ensure that this limit is not
  ## exceeded.  Otherwise, the program will error on the call.
  ## Real time prediction option:
  real.time  <- is.hidden.rt(dots)    
  real.time.bits  <- get.rt.bits(real.time)
  if (real.time) {
    real.time.options  <- is.hidden.rt.opt(dots)
  }
  else {
    real.time.options  <- NULL
  }
  ## Start the C external timer.
  ctime.external.start  <- proc.time()
  nativeOutput <- tryCatch({.Call("entryPred",
                                  as.integer(do.trace),
                                  as.integer(seed),
                                  as.integer(bootstrap.bits +   
                                             perf.bits +
                                             ensemble.bits +
                                             anonymize.bits),   ## low option word
                                  as.integer(membership.bits +  ## high option word
                                             2^19 + 2^18 +      ## TERM_INCG and TERM_OUTG
                                             samptype.bits),
                                  as.integer(experimental.bits),  ## rhf (local and experimental) option word
                                  as.integer(ntree),
                                  as.integer(object$n),
                                  list(as.integer(subj.unique.count),
                                       if (is.null(case.wt)) NULL else as.double(case.wt),
                                       as.integer(sampsize),
                                       if (is.null(samp)) NULL else as.integer(samp)),
                                  as.integer(hcut),
                                  as.integer(splitinfo$index),
                                  list(if (is.null(m.target.idx)) as.integer(0) else as.integer(length(m.target.idx)),
                                       if (is.null(m.target.idx)) NULL else as.integer(m.target.idx)),
                                  list(as.integer(length(yvar.types)),
                                       if (is.null(yvar.types)) NULL else as.character(yvar.types),
                                       if (is.null(yvar.nlevels)) NULL else as.integer(yvar.nlevels),
                                       if (is.null(yvar.nlevels)) NULL else sapply(1:length(yvar.nlevels), function(nn) {as.integer(length(yvar.nlevels[[nn]]))}),
                                       if (is.null(subj)) NULL else as.integer(subj),
                                       if (is.null(event.info)) as.integer(0) else as.integer(length(event.info$event.type)),
                                       if (is.null(event.info)) NULL else as.integer(event.info$event.type)),
                                  if (is.null(yvar.nlevels)) {
                                    NULL
                                  }
                                  else {
                                    lapply(1:length(yvar.nlevels),
                                           function(nn) {as.integer(yvar.nlevels[[nn]])})
                                  },
                                  if (is.null(yvar.types)) NULL else as.double(as.vector(data.matrix(yvar))),
                                  list(as.integer(ncol(xvar)),
                                       if (is.null(xvar.types)) NULL else as.character(xvar.types),
                                       if (is.null(xvar.nlevels)) NULL else as.integer(xvar.nlevels),
                                       if (is.null(xvar.nlevels)) NULL else sapply(1:length(xvar.nlevels), function(nn) {as.integer(length(xvar.nlevels[[nn]]))}),
                                       NULL,
                                       NULL),
                                  if (is.null(xvar.nlevels)) {
                                    NULL
                                  }
                                  else {
                                    lapply(1:length(xvar.nlevels),
                                           function(nn) {as.integer(xvar.nlevels[[nn]])})
                                  },
                                  as.double(as.vector(data.matrix(xvar))),
                                  ## assignment of augmented variables (list of two values: dimension, data)
                                  augmentXlist,
                                  as.integer(n.newdata),
                                  if (is.null(subj.newdata)) NULL else as.integer(subj.newdata),
                                  if (is.null(yvar.newdata)) NULL else as.double(as.vector(data.matrix(yvar.newdata))),
                                  if (is.null(xvar.newdata)) NULL else as.double(as.vector(data.matrix(xvar.newdata))),
                                  as.integer(object$totalNodeCount),
                                  as.integer(object$leafCount),
                                  as.integer(object$seed),
                                  as.integer((object$nativeArray)$treeID),
                                  as.integer((object$nativeArray)$nodeID),
                                  as.integer((object$nativeArray)$nodeSZ),
                                  as.integer((object$nativeArray)$brnodeID),
                                  ## This is hc_zero.  It is never NULL.
                                  list(as.integer((object$nativeArray)$parmID),
                                  as.double((object$nativeArray)$contPT),
                                  as.integer((object$nativeArray)$mwcpSZ),
                                  as.integer((object$nativeArray)$fsrecID),
                                  if (is.null((object$nativeFactorArray)$mwcpPT)) NULL else as.integer((object$nativeFactorArray)$mwcpPT)),
                                  as.integer(object$trmbrCaseCt),
                                  as.integer(object$timbrCaseCt),
                                  as.integer(object$tombrCaseCt),
                                  as.integer(object$trmbrCaseId),
                                  as.integer(object$timbrCaseId),
                                  as.integer(object$tombrCaseId),
                                  as.integer(get.tree),
                                  as.integer(block.size),
                                  if (real.time) list(as.integer(real.time.options$port), as.integer(real.time.options$time.out)) else NULL,
                                  as.integer(get.rf.cores()))}, error = function(e){NULL})
  ## Stop the C external timer.
  ctime.external.stop <- proc.time()
  ## check for error return condition in the native code
  if (is.null(nativeOutput)) {
    if (real.time == TRUE) {
      ## This is acceptable, for now.  Real time mode returns null,
      ## but we can revist this as the code evolves.
      return (NULL)  
    }
    else {
      stop("An error has occurred in prediction.  Please turn trace on for further analysis.")
    }
  }
  ## sample size used for the return predict object
  n.observed = ifelse(restore.mode, nrow(xvar), n.newdata)
  if (!restore.mode) {
      hazard.ibg <- NULL
      hazard.oob <- NULL
      chf.ibg <- NULL
      chf.oob <- NULL
      risk.ibg     <- NULL
      risk.oob     <- NULL
      int.haz.ibg <- NULL
      int.haz.oob <- NULL
      inbag.out <- NULL
      pseudo.membership <- NULL
  } 
  if (restore.mode) {
    if (membership) {
      pseudo.membership <- matrix(nativeOutput$nodeMembership, c(n.observed, ntree))
      inbag.out <- matrix(nativeOutput$bootstrapCount, c(subj.unique.count, ntree))
      nativeOutput$nodeMembership <- NULL
      nativeOutput$bootstrapCount <- NULL
    } 
    else {
      pseudo.membership <- NULL
      inbag.out  <- NULL
    }
    if (!is.null(nativeOutput$ensembleID)) {
      ensemble.id <- nativeOutput$ensembleID
    }
    else {
      ensemble.id <- NULL
    }
    if (!is.null(nativeOutput$ibgEnsbHazard)) {
      hazard.ibg  <- array(nativeOutput$ibgEnsbHazard, c(subj.unique.count, length(event.info$time.interest)))
    } else {
      hazard.ibg <- NULL
    }
    if (!is.null(nativeOutput$oobEnsbHazard)) {
      hazard.oob  <- array(nativeOutput$oobEnsbHazard, c(subj.unique.count, length(event.info$time.interest)))
    } else {
      hazard.oob <- NULL
    }
    if (!is.null(nativeOutput$ibgEnsbNlsnAaln)) {
      chf.ibg  <- array(nativeOutput$ibgEnsbNlsnAaln, c(subj.unique.count, length(event.info$time.interest)))
    } else {
      chf.ibg <- NULL
    }
    if (!is.null(nativeOutput$oobEnsbNlsnAaln)) {
      chf.oob  <- array(nativeOutput$oobEnsbNlsnAaln, c(subj.unique.count, length(event.info$time.interest)))
    } else {
      chf.oob <- NULL
    }
    if (!is.null(nativeOutput$ibgRisk)) {
      risk.ibg <- nativeOutput$ibgRisk
    } else {
      risk.ibg <- NULL
    }
    if (!is.null(nativeOutput$oobRisk)) {
      risk.oob <- nativeOutput$oobRisk
    } else {
      risk.oob <- NULL
    }
    if (!is.null(nativeOutput$ibgWCase)) {
        int.haz.ibg <- nativeOutput$ibgWCase
    } else {
        int.haz.ibg <- NULL
    }
    if (!is.null(nativeOutput$oobWCase)) {
        int.haz.oob <- nativeOutput$oobWCase
    } else {
        int.haz.oob <- NULL
    }
    if (!is.null(nativeOutput$absWCaseTimeLeft)) {
        int.haz.left <- .inverse.time(nativeOutput$absWCaseTimeLeft, time.map)
    } else {
        int.haz.left <- NULL
    }
    if (!is.null(nativeOutput$absWCaseTimeRight)) {
        int.haz.right <- .inverse.time(nativeOutput$absWCaseTimeRight, time.map)
    } else {
        int.haz.right <- NULL
    }
    hazard.tst  <- NULL
    chf.tst  <- NULL
    unscaled.risk.tst  <- NULL
    risk.tst     <- NULL
    int.haz.tst <- NULL
    ttmbrCaseCt <- NULL
    ttmbrCaseId <- NULL
  } else {
    if (!is.null(nativeOutput$ensembleID)) {
      ensemble.id <- nativeOutput$ensembleID
    }
    else {
      ensemble.id <- NULL
    }
    if (!is.null(nativeOutput$ibgEnsbHazard)) {
      hazard.tst  <- array(nativeOutput$ibgEnsbHazard, c(subj.newdata.unique.count, length(event.info$time.interest)))
    } else {
      hazard.tst <- NULL
    }
    if (!is.null(nativeOutput$ibgEnsbNlsnAaln)) {
      chf.tst  <- array(nativeOutput$ibgEnsbNlsnAaln, c(subj.newdata.unique.count, length(event.info$time.interest)))
    } else {
      chf.tst <- NULL
    }
    if (!is.null(nativeOutput$ibgRisk)) {
      risk.tst <- nativeOutput$ibgRisk
    } else {
      risk.tst <- NULL
    }
    if (!is.null(nativeOutput$ibgWCase)) {
        int.haz.tst <- nativeOutput$ibgWCase
    } else {
        int.haz.tst <- NULL
    }
    if (!is.null(nativeOutput$absWCaseTimeLeft)) {
      int.haz.left <- .inverse.time(nativeOutput$absWCaseTimeLeft, time.map)
    } else {
      int.haz.left <- NULL
    }
    if (!is.null(nativeOutput$absWCaseTimeRight)) {
      int.haz.right <- .inverse.time(nativeOutput$absWCaseTimeRight, time.map)
    } else {
      int.haz.right <- NULL
    }
    ## We use the existing oob data structure to output the test case counts and ids.
    ttmbrCaseCt = nativeOutput$tombrCaseCt
    ttmbrCaseId = nativeOutput$tombrCaseId
  }
  if (!is.null(nativeOutput$tNelsonAalen)) {
      t.chf <- vector("list", ntree)
      offset = 0
      for (i in 1:ntree) {
        t.chf[[i]] <- matrix(nativeOutput$tNelsonAalen[(offset+1):(offset + (nativeOutput$leafCount[i] * length(event.info$time.interest)))],
                                    c(nativeOutput$leafCount[i], length(event.info$time.interest)), byrow=TRUE)
          offset = offset + (nativeOutput$leafCount[i] * length(event.info$time.interest))
      }
  } else {
      t.chf <- NULL
  }
  if (!is.null(nativeOutput$tHazard)) {
      t.hazard <- vector("list", ntree)
      offset = 0
      for (i in 1:ntree) {
        t.hazard[[i]] <- matrix(nativeOutput$tHazard[(offset+1):(offset + (nativeOutput$leafCount[i] * length(event.info$time.interest)))],
                                    c(nativeOutput$leafCount[i], length(event.info$time.interest)), byrow=TRUE)
          offset = offset + (nativeOutput$leafCount[i] * length(event.info$time.interest))
      }
  } else {
      t.hazard <- NULL
  }
  if (!is.null(nativeOutput$tHazardTimeCnt)) {
      t.haz.time.cnt  <- nativeOutput$tHazardTimeCnt
  }
  else {
      t.haz.time.cnt  <- NULL
  }
  if (!is.null(nativeOutput$tHazardTimeIdx)) {
      t.haz.time.idx <- vector("list", n.observed)
      offset = 0
      for (i in 1:n.observed) {
          if (t.haz.time.cnt[i] != 0) {
              t.haz.time.idx[[i]]  <- nativeOutput$tHazardTimeIdx[(offset+1):(offset + t.haz.time.cnt[i])]
              offset = offset + t.haz.time.cnt[i]
          }
          else {
              t.haz.time.idx[[i]] <- NA_integer_
          }
      }
  }
  else {
      t.haz.time.idx  <- NULL
  }
  ## set this to NULL 
  err.rate <- NULL
  ## scale values back to original time scale
  hz.scale <- .hazard.scale(event.info$time.interest, time.map)
  if (!is.null(hazard.ibg)) hazard.ibg <- sweep(hazard.ibg, 2L, hz.scale, "*")
  if (!is.null(hazard.oob)) hazard.oob <- sweep(hazard.oob, 2L, hz.scale, "*")
  if (!is.null(hazard.tst)) hazard.tst <- sweep(hazard.tst, 2L, hz.scale, "*")
  if (!is.null(t.hazard)) {
    t.hazard <- lapply(t.hazard, function(h) sweep(h, 2L, hz.scale, "*"))
  }
  ## DO NOT rescale chf or t.chf
  ## make the output object
  rhfOutput <- list(
    forest = object,
    family = family,
    n = n.observed,
    ntree = ntree,
    yvar = if (restore.mode) as.data.frame(.scale.yvar(yvar, time.map)) else as.data.frame(.scale.yvar(yvar.newdata, time.map)),
    xvar = if (restore.mode) xvar else xvar.newdata,
    xvar.time = object$xvar.time, 
    hcut = hcut,
    max.time = max.time,
    time.map = time.map,
    time.interest = .inverse.time(event.info$time.interest, time.map),
    event.info = event.info,
    ensemble.id = ensemble.id,
    hazard.inbag = hazard.ibg,
    hazard.oob = hazard.oob,
    hazard.test = hazard.tst,
    chf.inbag = chf.ibg,
    chf.oob   = chf.oob,
    chf.test = chf.tst,
    risk.inbag = risk.ibg,
    risk.oob   = risk.oob,
    risk.test  = risk.tst,
    int.haz.inbag = int.haz.ibg,
    int.haz.oob   = int.haz.oob,
    int.haz.test  = int.haz.tst,
    int.haz.left = int.haz.left,
    int.haz.right = int.haz.right,
    t.chf = t.chf,
    t.hazard = t.hazard,
    t.haz.time.cnt = t.haz.time.cnt,
    t.haz.time.idx = t.haz.time.idx,
    ttmbrCaseCt = ttmbrCaseCt,
    ttmbrCaseId = ttmbrCaseId,
    id = if (restore.mode) id else subj.newdata,
    pseudo.membership = pseudo.membership,
    inbag = inbag.out,
    err.rate = err.rate,
    ctime.internal = nativeOutput$cTimeInternal,
    ctime.external = ctime.external.stop - ctime.external.start
  )
  ## memory management
  nativeOutput$leafCount <- NULL
  remove(object)
  class(rhfOutput) <- c("rhf", "predict",   family)
  return(rhfOutput)
}

rhf <- function(formula,
                data,
                ntree = 500,
                nsplit = 10,
                treesize = NULL,
                nodesize = NULL,
                block.size = 10,
                bootstrap = c("by.root", "none", "by.user"),
                samptype = c("swor", "swr"),
                samp = NULL,
                case.wt = NULL,
                membership = TRUE,
                sampsize = if (samptype == "swor") function(x){x * .632} else function(x){x},
                xvar.wt = NULL,
                ntime = 50,
                min.events.per.gap = 10,
                seed = NULL,
                do.trace = FALSE,
                ...
                )
{
  bootstrap <- match.arg(bootstrap, c("by.root", "none", "by.user"))
  samptype <- match.arg(samptype, c("swor", "swr"))
  do.call("rhf.workhorse", c(list(
                             formula=formula,
                             data=data,
                             ntree = ntree,
                             nsplit = nsplit,
                             treesize = treesize,
                             nodesize = nodesize,
                             block.size = block.size,
                             bootstrap = bootstrap,
                             samptype = samptype,
                             samp = samp,
                             case.wt = case.wt,
                             membership = membership,
                             sampsize = sampsize,
                             xvar.wt = xvar.wt,
                             ntime = ntime,
                             min.events.per.gap = min.events.per.gap,
                             seed = seed,
                             do.trace = do.trace),
                             list(...)))
}

########################################################################
## window engine helpers
########################################################################
.rhf_build_window_plan <- function(start, stopv, grid) {
  K <- length(grid)
  scale.time <- max(1, abs(c(start, stopv, grid)), na.rm = TRUE)
  eps.time <- sqrt(.Machine$double.eps) * scale.time
  ## Window k is (left[k], right[k]] with:
  ##   left  = c(0, grid[-K])
  ##   right = grid
  left.idx <- findInterval(start, grid) + 1L
  left.idx <- pmin(left.idx, K + 1L)
  stop.adj <- pmax(stopv - eps.time, 0)
  right.idx <- findInterval(stop.adj, grid) + 1L
  right.idx <- pmax(0L, pmin(right.idx, K))
  valid <- left.idx <= right.idx
  diff.count <- numeric(K + 1L)
  if (any(valid)) {
    diff.count <- tabulate(left.idx[valid], nbins = K + 1L) -
      tabulate(right.idx[valid] + 1L, nbins = K + 1L)
  }
  n.risk <- as.integer(cumsum(diff.count)[seq_len(K)])
  window.left <- c(0, grid[-K])
  window.right <- grid
  window.mid <- (window.left + window.right) / 2
  window.label <- paste0("(",
                         .rhf_format_time_names(window.left),
                         ", ",
                         .rhf_format_time_names(window.right),
                         "]")
  list(
    K = K,
    left.idx = left.idx,
    right.idx = right.idx,
    n.risk = n.risk,
    window.left = window.left,
    window.right = window.right,
    window.mid = window.mid,
    window.label = window.label
  )
}
.rhf_empty_membership_series <- function(K) {
  list(
    count_all = integer(K),
    count_valid = integer(K),
    sum = numeric(K),
    sumsq = numeric(K)
  )
}
.rhf_group_sums2 <- function(group, value1, value2, nbins) {
  out1 <- numeric(nbins)
  out2 <- numeric(nbins)
  if (!length(group)) {
    return(list(sum = out1, sumsq = out2))
  }
  if (length(group) == 1L) {
    gg <- as.integer(group)[1L]
    out1[gg] <- as.numeric(value1)[1L]
    out2[gg] <- as.numeric(value2)[1L]
    return(list(sum = out1, sumsq = out2))
  }
  rs <- rowsum(cbind(value1, value2), group = group, reorder = FALSE)
  idx <- as.integer(rownames(rs))
  out1[idx] <- rs[, 1L]
  out2[idx] <- rs[, 2L]
  list(sum = out1, sumsq = out2)
}
.rhf_membership_window_series <- function(idx,
                                          y,
                                          left.idx,
                                          right.idx,
                                          K,
                                          y.has.na = FALSE) {
  if (!length(idx)) {
    return(.rhf_empty_membership_series(K))
  }
  idx <- as.integer(idx)
  idx <- idx[idx >= 1L & idx <= length(left.idx)]
  if (!length(idx)) {
    return(.rhf_empty_membership_series(K))
  }
  li <- left.idx[idx]
  ri <- right.idx[idx]
  active <- li <= ri
  if (!any(active)) {
    return(.rhf_empty_membership_series(K))
  }
  idx <- idx[active]
  li <- as.integer(li[active])
  ri1 <- as.integer(ri[active] + 1L)
  K1 <- K + 1L
  diff.count.all <- tabulate(li, nbins = K1) - tabulate(ri1, nbins = K1)
  count.all <- as.integer(cumsum(diff.count.all)[seq_len(K)])
  yv <- y[idx]
  if (isTRUE(y.has.na)) {
    ok <- !is.na(yv)
    if (!any(ok)) {
      return(list(
        count_all = count.all,
        count_valid = integer(K),
        sum = numeric(K),
        sumsq = numeric(K)
      ))
    }
    li.valid <- li[ok]
    ri1.valid <- ri1[ok]
    yv <- yv[ok]
    diff.count.valid <- tabulate(li.valid, nbins = K1) - tabulate(ri1.valid, nbins = K1)
    count.valid <- as.integer(cumsum(diff.count.valid)[seq_len(K)])
  }
  else {
    li.valid <- li
    ri1.valid <- ri1
    count.valid <- count.all
  }
  add <- .rhf_group_sums2(li.valid,
                          value1 = yv,
                          value2 = yv * yv,
                          nbins = K1)
  drop <- .rhf_group_sums2(ri1.valid,
                           value1 = yv,
                           value2 = yv * yv,
                           nbins = K1)
  sum.series <- cumsum(add$sum - drop$sum)[seq_len(K)]
  sumsq.series <- cumsum(add$sumsq - drop$sumsq)[seq_len(K)]
  list(
    count_all = count.all,
    count_valid = count.valid,
    sum = sum.series,
    sumsq = sumsq.series
  )
}
.rhf_local_welch_from_stats <- function(n1,
                                        sum1,
                                        sumsq1,
                                        n2,
                                        sum2,
                                        sumsq2) {
  K <- length(n1)
  out <- rep(NA_real_, K)
  ok <- (n1 >= 2L) & (n2 >= 2L)
  if (!any(ok)) {
    return(out)
  }
  n1.ok <- as.numeric(n1[ok])
  n2.ok <- as.numeric(n2[ok])
  sum1.ok <- as.numeric(sum1[ok])
  sum2.ok <- as.numeric(sum2[ok])
  sumsq1.ok <- as.numeric(sumsq1[ok])
  sumsq2.ok <- as.numeric(sumsq2[ok])
  m1 <- sum1.ok / n1.ok
  m2 <- sum2.ok / n2.ok
  v1 <- (sumsq1.ok - (sum1.ok * sum1.ok) / n1.ok) / pmax(1, n1.ok - 1)
  v2 <- (sumsq2.ok - (sum2.ok * sum2.ok) / n2.ok) / pmax(1, n2.ok - 1)
  tiny <- sqrt(.Machine$double.eps)
  v1[is.finite(v1) & v1 < 0 & abs(v1) < tiny] <- 0
  v2[is.finite(v2) & v2 < 0 & abs(v2) < tiny] <- 0
  se <- sqrt(v1 / n1.ok + v2 / n2.ok)
  out.ok <- rep(NA_real_, length(n1.ok))
  good <- is.finite(se) & (se > 0)
  out.ok[good] <- abs((m1[good] - m2[good]) / se[good])
  out[ok] <- out.ok
  out
}
.rhf_base_branch_id <- function(strength,
                                base.rules) {
  if (!length(base.rules)) {
    return(integer(0))
  }
  tree <- strength$treeID[base.rules]
  node <- strength$nodeID[base.rules]
  as.integer(cumsum(c(TRUE,
                      (tree[-1L] != tree[-length(tree)]) |
                      (node[-1L] != node[-length(node)]))))
}
.rhf_precompute_rule_windows <- function(strength,
                                         base.rules,
                                         oobMembership,
                                         compMembership,
                                         y,
                                         left.idx,
                                         right.idx,
                                         K,
                                         verbose = FALSE) {
  n.base <- length(base.rules)
  if (!n.base) {
    return(list(
      keep = matrix(FALSE, nrow = 0L, ncol = K),
      n.oob = matrix(integer(0), nrow = 0L, ncol = K),
      local.imp = matrix(numeric(0), nrow = 0L, ncol = K),
      n.rules = integer(K)
    ))
  }
  t.start <- proc.time()[3L]
  if (isTRUE(verbose)) {
    message("[varpro.cache.rhf] precomputing window-local rule statistics")
  }
  y.has.na <- anyNA(y)
  branch.id <- .rhf_base_branch_id(strength = strength,
                                   base.rules = base.rules)
  n.branch <- max(branch.id)
  branch.first <- match(seq_len(n.branch), branch.id)
  oob.count.all <- matrix(0L, nrow = n.branch, ncol = K)
  oob.count.valid <- matrix(0L, nrow = n.branch, ncol = K)
  oob.sum <- matrix(0, nrow = n.branch, ncol = K)
  oob.sumsq <- matrix(0, nrow = n.branch, ncol = K)
  for (bb in seq_len(n.branch)) {
    rule.index <- base.rules[branch.first[bb]]
    ser <- .rhf_membership_window_series(
      idx = oobMembership[[rule.index]],
      y = y,
      left.idx = left.idx,
      right.idx = right.idx,
      K = K,
      y.has.na = y.has.na
    )
    oob.count.all[bb, ] <- ser$count_all
    oob.count.valid[bb, ] <- ser$count_valid
    oob.sum[bb, ] <- ser$sum
    oob.sumsq[bb, ] <- ser$sumsq
  }
  keep <- matrix(FALSE, nrow = n.base, ncol = K)
  n.oob <- matrix(0L, nrow = n.base, ncol = K)
  local.imp <- matrix(NA_real_, nrow = n.base, ncol = K)
  for (ii in seq_len(n.base)) {
    bb <- branch.id[ii]
    rule.index <- base.rules[ii]
    ser.comp <- .rhf_membership_window_series(
      idx = compMembership[[rule.index]],
      y = y,
      left.idx = left.idx,
      right.idx = right.idx,
      K = K,
      y.has.na = y.has.na
    )
    n.oob[ii, ] <- oob.count.all[bb, ]
    keep[ii, ] <- (oob.count.all[bb, ] > 0L) & (ser.comp$count_all > 0L)
    local.imp[ii, ] <- .rhf_local_welch_from_stats(
      n1 = oob.count.valid[bb, ],
      sum1 = oob.sum[bb, ],
      sumsq1 = oob.sumsq[bb, ],
      n2 = ser.comp$count_valid,
      sum2 = ser.comp$sum,
      sumsq2 = ser.comp$sumsq
    )
  }
  if (isTRUE(verbose)) {
    message("[varpro.cache.rhf] rule-window statistics ready: ",
            n.base, " rules x ", K, " windows",
            ", elapsed ",
            .rhf_format_elapsed(proc.time()[3L] - t.start))
  }
  list(
    keep = keep,
    n.oob = n.oob,
    local.imp = local.imp,
    n.rules = as.integer(colSums(keep))
  )
}

####################################################################
##
## TDC Helper Functions
##
####################################################################
## converts a standard survival data set to counting process format
## scale is ignored but kept in place for legacy
convert.counting <- function(f, dta, scale = FALSE) {
  ## coerce formula
  f <- as.formula(f)
  ## extract names
  time.nm  <- all.vars(f)[1]
  event.nm <- all.vars(f)[2]
  ## remove missing values
  dta <- na.omit(dta)
  ## raw time
  time <- as.numeric(dta[[time.nm]])
  if (any(!is.finite(time))) {
    stop("Non-finite time values found.")
  }
  if (any(time < 0, na.rm = TRUE)) {
    stop("time values cannot be negative")
  }
  if (isTRUE(scale)) {
    warning("'scale=TRUE' is deprecated for RHF workflows: rhf() now maps time internally. Returning raw times.")
  }
  data.frame(id = seq_len(nrow(dta)),
             start = 0,
             stop = time,
             event = dta[[event.nm]],
             dta[, !(colnames(dta) %in% all.vars(f)[1:2]), drop = FALSE])
}
## clean up counting process data
cleanup.counting <- function(dta,
                             xvar.names = NULL,
                             yvar.names,
                             subj.names,
                             sorted = FALSE,
                             eps = 1e-6) {
  ## work out candidate x variables
  if (is.null(xvar.names)) {
    xvar.candidate <- setdiff(colnames(dta), c(subj.names, yvar.names))
  } else {
    missing.x <- setdiff(xvar.names, colnames(dta))
    if (length(missing.x) > 0L) {
      stop("The following xvar.names are not columns of 'dta': ",
           paste(missing.x, collapse = ", "))
    }
    xvar.candidate <- setdiff(unique(xvar.names), c(subj.names, yvar.names))
  }
  xvar.all   <- xvar.candidate
  xvar.names <- xvar.candidate
  ## coerce start/stop once, early
  start.nm <- yvar.names[1L]
  stop.nm  <- yvar.names[2L]
  for (nm in c(start.nm, stop.nm)) {
    v <- dta[[nm]]
    if (is.factor(v)) {
      v <- as.numeric(as.character(v))
    } else {
      v <- as.numeric(v)
    }
    dta[[nm]] <- v
  }
  startv <- dta[[start.nm]]
  stopv  <- dta[[stop.nm]]
  ## check if time is non-negative (ignoring NAs)
  if (any(startv < 0 | stopv < 0, na.rm = TRUE)) {
    stop("time values cannot be negative")
  }
  ## drop x variables with all missing values
  if (length(xvar.names) > 0L) {
    all.na <- vapply(
      xvar.names,
      function(z) all(is.na(dta[[z]])),
      logical(1L)
    )
    if (any(all.na)) {
      drop.x     <- xvar.names[all.na]
      dta        <- dta[, !(colnames(dta) %in% drop.x), drop = FALSE]
      xvar.names <- xvar.names[!all.na]
      warning("Dropping x variable(s) with all missing values: ",
              paste(drop.x, collapse = ", "))
    }
  }
  ## case-wise deletion of remaining missing data
  cols.to.check <- intersect(c(subj.names, yvar.names, xvar.names),
                             colnames(dta))
  if (length(cols.to.check) == 0L) {
    stop("No variables remain in data after removing predictors with all missing values.")
  }
  cc <- complete.cases(dta[, cols.to.check, drop = FALSE])
  if (!all(cc)) {
    n.rem <- sum(!cc)
    dta   <- dta[cc, , drop = FALSE]
    warning(sprintf("Removing %d row(s) with missing values.", n.rem))
  }
  ## checks after NA cleaning
  if (nrow(dta) == 0L) {
    stop("After processing (including removing missing values) there are no observations left (NULL data set).")
  }
  if (length(xvar.names) == 0L) {
    if (length(xvar.all) == 0L) {
      stop("No x variables were supplied (xvar.names is empty and no extra variables were found).")
    } else {
      stop("After removing missing values there are no x variables left.")
    }
  }
  ## one-pass stable sort by subject encounter-order, then start time.
  ## This preserves the original grouping semantics without the O(n * n_subjects)
  ## cost of lapply(unique(id), which(id == i), ...).
  if (isFALSE(sorted)) {
    id  <- dta[[subj.names]]
    ord <- order(match(id, id), dta[[start.nm]], method = "radix")
    if (!identical(ord, seq_len(nrow(dta)))) {
      dta <- dta[ord, , drop = FALSE]
    }
  }
  ## transform time to [0,1] using modified logit with tau = training max.time
  max.time <- max(dta[[stop.nm]], na.rm = TRUE)
  if (!is.finite(max.time) || max.time <= 0) {
    stop("Invalid max time in training data: max(stop) must be positive and finite.")
  }
  time.map <- list(method   = "mlogit",
                   tau      = as.double(max.time),
                   max.time = as.double(max.time))
  startv <- .forward.time(dta[[start.nm]], time.map)
  stopv  <- .forward.time(dta[[stop.nm]],  time.map)
  ## remove non-finite time values and any data points where stop<=start within epsilon tolerance
  bad <- (!is.finite(startv) | !is.finite(stopv) | (stopv <= startv + eps))
  bad[is.na(bad)] <- TRUE
  if (any(bad)) {
    keep <- !bad
    dta   <- dta[keep, , drop = FALSE]
    startv <- startv[keep]
    stopv  <- stopv[keep]
  }
  if (nrow(dta) == 0L) {
    stop("After processing there are no valid counting-process intervals left.")
  }
  dta[[start.nm]] <- startv
  dta[[stop.nm]]  <- stopv
  ## store helpful attributes
  attr(dta, "max.time")              <- max.time
  attr(dta, "time.map")              <- time.map
  attr(dta, "xvar.names")            <- xvar.names
  attr(dta, "sorted.by.subj.start")  <- TRUE
  dta
}
## helper to clean and scale new (test) counting-process data for predict.rhf
cleanup.counting.newdata <- function(newdata,
                                     xvar.names,
                                     yvar.names,
                                     subj.names,
                                     time.map,
                                     max.time,
                                     sorted = FALSE,
                                     eps = 1e-6,
                                     nonfinite.action = c("stop", "drop")) {
  nonfinite.action <- match.arg(nonfinite.action)
  if (missing(newdata) || is.null(newdata)) {
    stop("Argument 'newdata' must be a non-null data.frame.")
  }
  cn <- colnames(newdata)
  ## check subject variable present
  missing.subj <- setdiff(subj.names, cn)
  if (length(missing.subj) > 0L) {
    stop("Subject variable(s) not found in 'newdata': ",
         paste(missing.subj, collapse = ", "))
  }
  ## check predictor variables present (must match training forest)
  missing.x <- setdiff(xvar.names, cn)
  if (length(missing.x) > 0L) {
    stop("The following predictor variable(s) used to fit the forest are missing in 'newdata': ",
         paste(missing.x, collapse = ", "))
  }
  ## determine whether y is present in newdata
  y.in <- yvar.names %in% cn
  if (any(y.in) && !all(y.in)) {
    stop("Argument 'yvar.names' must refer to variables that are either all present in 'newdata' or all absent.")
  }
  yvar.present <- all(y.in)
  start.nm <- stop.nm <- NULL
  ## coerce start/stop once, early (if y present)
  if (yvar.present) {
    start.nm <- yvar.names[1L]
    stop.nm  <- yvar.names[2L]
    for (nm in c(start.nm, stop.nm)) {
      v <- newdata[[nm]]
      if (is.factor(v)) {
        v <- as.numeric(as.character(v))
      } else {
        v <- as.numeric(v)
      }
      newdata[[nm]] <- v
    }
    if (any(newdata[[start.nm]] < 0 | newdata[[stop.nm]] < 0, na.rm = TRUE)) {
      stop("time values in 'newdata' cannot be negative")
    }
    ## retained for legacy/interface sanity checks
    if (length(max.time) != 1L || !is.finite(max.time) || max.time <= 0) {
      stop("Invalid 'max.time' from training data: must be a positive finite scalar.")
    }
  }
  ## case-wise deletion of missing values in id, x, and (if present) y
  cols.to.check <- c(subj.names, xvar.names)
  if (yvar.present) {
    cols.to.check <- c(cols.to.check, yvar.names)
  }
  cols.to.check <- intersect(cols.to.check, cn)
  cc <- complete.cases(newdata[, cols.to.check, drop = FALSE])
  if (!all(cc)) {
    n.drop <- sum(!cc)
    warning(sprintf(
      "Removing %d row(s) from 'newdata' due to missing values in subject, y, or x.",
      n.drop
    ))
  }
  if (!any(cc)) {
    stop("After removing missing values from 'newdata' there are no observations left.")
  }
  dta <- newdata[cc, , drop = FALSE]
  ## one-pass stable sort by subject encounter-order, then start time.
  ## If y is absent, group by subject while preserving within-subject order.
  if (isFALSE(sorted)) {
    id <- dta[[subj.names]]
    if (yvar.present) {
      ord <- order(match(id, id), dta[[start.nm]], method = "radix")
    } else {
      ord <- order(match(id, id), seq_len(nrow(dta)), method = "radix")
    }
    if (!identical(ord, seq_len(nrow(dta)))) {
      dta <- dta[ord, , drop = FALSE]
    }
  }
  start.scaled <- stop.scaled <- NULL
  if (yvar.present) {
    ## transform start/stop outcomes to the internal time scale
    start.scaled <- .forward.time(dta[[start.nm]], time.map)
    stop.scaled  <- .forward.time(dta[[stop.nm]],  time.map)
    ## coherence check
    nonfinite <- (!is.finite(start.scaled) | !is.finite(stop.scaled))
    bad <- nonfinite | (stop.scaled <= start.scaled + eps)
    bad[is.na(bad)] <- TRUE
    if (any(nonfinite)) {
      msg <- sprintf("Found %d row(s) in 'newdata' with non-finite start/stop after scaling.",
                     sum(nonfinite))
      if (nonfinite.action == "stop") {
        stop(msg)
      } else {
        warning(msg)
      }
    }
    if (any(bad)) {
      dta <- dta[!bad, , drop = FALSE]
      start.scaled <- start.scaled[!bad]
      stop.scaled  <- stop.scaled[!bad]
    }
    if (nrow(dta) == 0L) {
      stop("After removing invalid/non-positive length intervals from 'newdata' there are no observations left.")
    }
  }
  attr(dta, "sorted.by.subj.start") <- isTRUE(yvar.present)
  ## initialise outputs only after final filtering to avoid extra copies
  subj.newdata <- dta[, subj.names]
  xvar.newdata <- dta[, xvar.names, drop = FALSE]
  yvar.newdata <- NULL
  if (yvar.present) {
    ## preserve original coercion behavior for non-time y columns
    yraw <- dta[, yvar.names, drop = FALSE]
    for (nm in yvar.names[1:2]) {
      v <- yraw[[nm]]
      if (is.factor(v)) v <- as.numeric(as.character(v))
      yraw[[nm]] <- as.numeric(v)
    }
    yvar.newdata <- as.matrix(yraw)
    storage.mode(yvar.newdata) <- "double"
    yvar.newdata[, 1L] <- start.scaled
    yvar.newdata[, 2L] <- stop.scaled
  }
  list(
    newdata      = dta,
    subj         = subj.newdata,
    xvar         = xvar.newdata,
    yvar         = yvar.newdata,
    yvar.present = yvar.present
  )
}
get.duplicated <- function(x) {
  c(apply(x, 2, function(xx) {
    1 * !all(xx == xx[1])
  }))
}
get.tdc.cov <- function(dta, subj.names, yvar.names,
                        presorted = isTRUE(attr(dta, "sorted.by.subj.start"))) {
  ## extract covariates
  x <- dta[, !(colnames(dta) %in% c(subj.names, yvar.names)), drop = FALSE]
  p <- ncol(x)
  if (p == 0L) {
    return(numeric(0L))
  }
  id <- dta[, subj.names]
  n  <- length(id)
  if (n <= 1L) {
    out <- rep(0L, p)
    names(out) <- colnames(x)
    return(out)
  }
  if (presorted) {
    ord <- seq_len(n)
    id.ord <- id
  } else {
    ord <- order(id, method = "radix")
    id.ord <- id[ord]
  }
  same <- id.ord[-1L] == id.ord[-n]
  if (!any(same, na.rm = TRUE)) {
    out <- rep(0L, p)
    names(out) <- colnames(x)
    return(out)
  }
  out <- vapply(
    x,
    function(v) {
      v <- v[ord]
      chg <- (v[-1L] != v[-n]) & same
      as.integer(any(chg, na.rm = TRUE))
    },
    integer(1L)
  )
  out
}
get.tdc.subj.time <- function(dta, subj.names, yvar.names,
                              presorted = isTRUE(attr(dta, "sorted.by.subj.start"))) {
  ## extract covariates
  x <- dta[, !(colnames(dta) %in% c(subj.names, yvar.names)), drop = FALSE]
  p <- ncol(x)
  ## preserve original behavior in trivial cases
  if (p == 0L) {
    return(rep(0L, nrow(dta)))
  }
  id <- dta[, subj.names]
  n  <- length(id)
  if (n <= 1L) {
    return(rep(0L, nrow(dta)))
  }
  if (presorted) {
    ord <- seq_len(n)
    id.ord <- id
  } else {
    ord <- order(id, method = "radix")
    id.ord <- id[ord]
  }
  same <- id.ord[-1L] == id.ord[-n]
  if (!any(same, na.rm = TRUE)) {
    return(rep(0L, nrow(dta)))
  }
  cov.tdc <- integer(p)
  names(cov.tdc) <- colnames(x)
  pair.chg <- logical(n - 1L)
  for (j in seq_len(p)) {
    v <- x[[j]][ord]
    chg <- (v[-1L] != v[-n]) & same
    cov.tdc[j] <- as.integer(any(chg, na.rm = TRUE))
    pair.chg <- pair.chg | (!is.na(chg) & chg)
  }
  if (sum(cov.tdc) == 0) {
    return(rep(0L, nrow(dta)))
  }
  pc <- as.integer(pair.chg)
  cs <- c(0L, cumsum(pc))
  rr <- rle(id.ord)
  ends <- cumsum(rr$lengths)
  starts <- ends - rr$lengths + 1L
  subj.has.chg <- (cs[ends] - cs[starts]) > 0L
  if (presorted) {
    out <- as.integer(subj.has.chg)
  } else {
    id.unq <- unique(id)
    by_id <- setNames(as.integer(subj.has.chg), as.character(rr$values))
    out <- unname(by_id[as.character(id.unq)])
  }
  out
}
#########################################################################################
##
## process survival information, set time.interest
##
#########################################################################################
timegrid.min.events <- function(data,
                                stop.col = "stop",
                                event.col = "event",
                                ntime = 50L,
                                min.events.per.gap = 10L,
                                include.first.event = FALSE,
                                merge.tail = TRUE) {
  if (!stop.col %in% names(data))  stop("stop.col not found in data")
  if (!event.col %in% names(data)) stop("event.col not found in data")
  ntime <- as.integer(ntime)
  if (is.na(ntime) || ntime < 0L) stop("ntime must be >= 0")
  min.events.per.gap <- as.integer(min.events.per.gap)
  if (is.na(min.events.per.gap) || min.events.per.gap < 1L) {
    stop("min.events.per.gap must be >= 1")
  }
  # Robust conversion for stop
  stopv <- data[[stop.col]]
  if (is.factor(stopv)) stopv <- as.numeric(as.character(stopv))
  stopv <- as.numeric(stopv)
  evv <- data[[event.col]]
  evv <- as.integer(!is.na(evv) & evv != 0L)
  ev_time <- stopv[evv == 1L]
  ev_time <- ev_time[is.finite(ev_time) & ev_time > 0]
  if (!length(ev_time)) stop("No positive event times found.")
  # Unique event times + event counts at each time
  ut <- sort(unique(ev_time))
  counts <- tabulate(match(ev_time, ut), nbins = length(ut))
  n_events <- sum(counts)
  # If ntime is 0: all unique event times
  if (is.null(ntime) || ntime == 0L) {
    grid <- ut
    attr(grid, "n_events") <- n_events
    attr(grid, "n_unique_event_times") <- length(ut)
    attr(grid, "min.events.per.gap_eff") <- NA_integer_
    return(grid)
  }
  # Effective events per gap combines both controls:
  # - never < min.events.per.gap
  # - tends to produce ~ntime grid points when possible
  m_eff <- max(min.events.per.gap, as.integer(ceiling(n_events / ntime)))
  m_eff <- max(1L, min(m_eff, n_events))
  grid_idx <- integer(0)
  acc <- 0L
  for (k in seq_along(ut)) {
    acc <- acc + counts[k]
    if (acc >= m_eff) {
      grid_idx <- c(grid_idx, k)
      acc <- 0L
    }
  }
  tmax <- ut[length(ut)]
  if (!length(grid_idx)) {
    grid <- tmax
  } else {
    grid <- ut[grid_idx]
    if (tail(grid, 1) < tmax) {
      if (merge.tail && acc > 0L && acc < m_eff) {
        grid[length(grid)] <- tmax
      } else {
        grid <- c(grid, tmax)
      }
    }
  }
  if (include.first.event) grid <- c(ut[1L], grid)
  grid <- sort(unique(grid))
  ## do not attach attributes, native code does not like that 
  #attr(grid, "n_events") <- n_events
  #attr(grid, "n_unique_event_times") <- length(ut)
  #attr(grid, "min.events.per.gap_eff") <- m_eff
  grid
}
get.grow.event.info <- function(yvar, fmly, need.deaths = TRUE, ntime, min.events.per.gap) {
  if (grepl("surv", fmly)) {
    ##-----------------------------------------------------------
    ## survival, competing risks, or time dependent covariates
    ##-----------------------------------------------------------
    if (dim(yvar)[2] == 2) {
      ##---------------------------------
      ## survival or competing risks:
      ##---------------------------------
      r.dim <- 2
      time <- yvar[, 1]
      cens <- yvar[, 2]
      start.time <- NULL
      ## censoring must be coded coherently
      if (!all(floor(cens) == abs(cens), na.rm = TRUE)) {
        stop("for survival families censoring variable must be coded as a non-negative integer (perhaps the formula is set incorrectly?)")
      }
      ## check if deaths are available (if user specified)
      if (need.deaths && (all(na.omit(cens) == 0))) {
        stop("no deaths in data!")
      }
      ## Check for event time consistency.
      ## we over-ride this now to allow for negative time (see Stute)
      ##if (!all(na.omit(time) >= 0)) {
      ##  stop("time must be  positive")
      ##}
      ## Extract the unique event types.
      event.type <- unique(na.omit(cens))
      ## Ensure they are all greater than or equal to zero.
      if (sum(event.type >= 0) != length(event.type)) {
        stop("censoring variable must be coded as NA, 0, or greater than 0.")
      }
      ## Discard the censored state, if it exists.
      event <- na.omit(cens)[na.omit(cens) > 0]
      event.type <- unique(event)
      ## Set grid of time points.
      nonMissingOutcome <- which(!is.na(cens) & !is.na(time))
      nonMissingDeathFlag <- (cens[nonMissingOutcome] != 0)
      time.interest <- sort(unique(time[nonMissingOutcome[nonMissingDeathFlag]]))
      ## trim the time points if the user has requested it
      ## we also allow the user to pass requested time points
      if (!is.null(ntime) && length(ntime) == 1 && ntime < 0) {
        ntime <- 0
      }
      if (!is.null(ntime) && !((length(ntime) == 1) && ntime == 0)) {
        if (length(ntime) == 1 && length(time.interest) > ntime) {
          time.interest <- time.interest[
            seq(1, length(time.interest), length = ntime)]
        }
        if (length(ntime) > 1) {
          time.interest <- unique(sapply(ntime, function(tt) {
            time.interest[max(1, sum(tt >= time.interest, na.rm = TRUE))]
          }))
        }
      }
    }
    ##-------------------------------
    ## time dependent covariates:
    ##-------------------------------
    else {
      r.dim <- 3
      start.time <- yvar[, 1]
      time <- yvar[, 2]
      cens <- yvar[, 3]
      ## censoring must be coded coherently
      if (!all(floor(cens) == abs(cens), na.rm = TRUE)) {
        stop("for survival families censoring variable must be coded as a non-negative integer (perhaps the formula is set incorrectly?)")
      }
      ## check if deaths are available (if user specified)
      if (need.deaths && (all(na.omit(cens) == 0))) {
        stop("no deaths in data!")
      }
      ## Check for event time consistency.
      if (!all(na.omit(time) >= 0)) {
        stop("time must be  positive")
      }
      ## Extract the unique event types.
      event.type <- unique(na.omit(cens))
      ## Ensure they are all greater than or equal to zero.
      if (sum(event.type >= 0) != length(event.type)) {
        stop("censoring variable must be coded as NA, 0, or greater than 0.")
      }
      ## Discard the censored state, if it exists.
      event <- na.omit(cens)[na.omit(cens) > 0]
      event.type <- unique(event)
      ## Set grid of time points.
      nonMissingOutcome <- which(!is.na(cens) & !is.na(time))
      nonMissingDeathFlag <- (cens[nonMissingOutcome] != 0)
      time.interest <- sort(unique(time[nonMissingOutcome[nonMissingDeathFlag]]))
      ## trim the time points if the user has requested it
      ## we also allow the user to pass requested time points
      if (!is.null(ntime) && length(ntime) == 1 && ntime < 0) {
        ntime <- 0
      }
      if (!is.null(ntime) && !((length(ntime) == 1) && ntime == 0)) {
        if (length(ntime) == 1) {
          time.interest <- timegrid.min.events(data.frame(stop=time, event=cens),
                                               ntime = ntime,
                                               min.events.per.gap = min.events.per.gap)
        }
        else {
          time.interest <- unique(sapply(ntime, function(tt) {
            time.interest[max(1, sum(tt >= time.interest, na.rm = TRUE))]
          }))
        }
      }
    }
  }
  ##---------------------
  ## other families
  ##---------------------
  else {
    if ((fmly == "regr+") | (fmly == "class+") | (fmly == "mix+")) {
      r.dim <- dim(yvar)[2]
    }
    else {
      if (fmly == "unsupv") {
        r.dim <- 0
      }
      else {
        r.dim <- 1
      }
    }
    event <- event.type <- cens <- time.interest <- cens <- time <- start.time <- NULL
  }
  return(list(event = event, event.type = event.type, cens = cens,
              time.interest = time.interest,
              time = time, start.time = start.time, r.dim = r.dim))
}
####################################################################
##
## TDC Helper Functions for hazards
## - only coherent for single or time static trees
##
####################################################################
hazard.to.chf <- function(o, max.time = 1) {
  tme.delta <- diff(c(0, o$time.interest))
  hz <- if (!is.null(o$hazard.oob)) o$hazard.oob else o$hazard
  t(apply(hz, 1, function(h) cumsum(h * tme.delta)))
}
#########################################################################################
##
## convert.standard.counting()
##
## Collapse counting-process (start/stop) data to a single-row-per-ID dataset.
##
## Original purpose:
##   - Quickly convert counting-process survival data to a standard survival
##     data set using a baseline snapshot of covariates.
##
## Extension (landmarking):
##   - Provide a landmark snapshot at a user-specified time (landmark.time)
##     using the covariate row whose [start, stop) interval contains t0^-.
##   - Optionally return the row index used for the snapshot (row_index)
##     which is useful for extracting record-level predictions (e.g., CHF)
##     from survival forests fit on the long (record-level) data.
##
## Notes:
##   - The function supports a "pseudo" Surv() with an ID argument:
##       Surv(id, start, stop, event)
##     as well as the standard counting-process Surv():
##       Surv(start, stop, event)
##     In the latter case, id.default is used.
##   - landmark.time is assumed to be on the *same scale* as the start/stop
##     columns in `data`.
##
#########################################################################################
convert.standard.counting <- function(formula, data,
                                      scale             = FALSE,
                                      rescale.from.attr = FALSE,
                                      keep.id           = FALSE,
                                      keep.row_index    = FALSE,
                                      sorted            = FALSE,
                                      id.default        = "id",
                                      eps               = 1e-8,
                                      landmark.time     = NULL,
                                      landmark.use.tminus = TRUE,
                                      return.type       = c("survival", "x"),
                                      keep.landmark.cols = FALSE) {
  return.type <- match.arg(return.type)
  ## ---- parse Surv(...) on LHS ----
  f   <- as.formula(formula)
  lhs <- f[[2]]
  if (!is.call(lhs) || !identical(as.character(lhs[[1]]), "Surv")) {
    stop("Left-hand side of formula must be a Surv(...) call.")
  }
  getnm <- function(z) paste(deparse(z), collapse = "")
  L <- as.list(lhs)[-1]  # drop 'Surv'
  if (length(L) == 4L) {
    ## Surv(id, start, stop, event)
    subj.nm  <- getnm(L[[1]])
    start.nm <- getnm(L[[2]])
    stop.nm  <- getnm(L[[3]])
    event.nm <- getnm(L[[4]])
  } else if (length(L) == 3L) {
    ## Surv(start, stop, event) -> use default id column
    subj.nm  <- id.default
    start.nm <- getnm(L[[1]])
    stop.nm  <- getnm(L[[2]])
    event.nm <- getnm(L[[3]])
  } else {
    stop("Expecting Surv(id, start, stop, event) or Surv(start, stop, event).")
  }
  ## required columns
  req  <- c(subj.nm, start.nm, stop.nm, event.nm)
  miss <- setdiff(req, names(data))
  if (length(miss)) {
    stop("Missing required columns in data: ", paste(miss, collapse = ", "))
  }
  id         <- data[[subj.nm]]
  start.time <- data[[start.nm]]
  stop.time  <- data[[stop.nm]]
  event.val  <- data[[event.nm]]
  if (any(!is.finite(start.time) | !is.finite(stop.time))) {
    stop("Non-finite start/stop values found.")
  }
  if (any(stop.time < start.time - eps, na.rm = TRUE)) {
    stop("Found rows with stop < start beyond tolerance.")
  }
  ## Optional: rescale start/stop times (and landmark.time) using attr(data, 'max.time')
  ## This is mainly useful if start/stop were stored on [0,1] and you want
  ## to work in original units.
  if (isTRUE(rescale.from.attr)) {
    tm <- attr(data, "time.map")
    if (!is.null(tm)) {
      start.time <- .inverse.time(start.time, tm)
      stop.time  <- .inverse.time(stop.time,  tm)
      if (!is.null(landmark.time)) {
        landmark.time <- .inverse.time(landmark.time, tm)
      }
    } else {
      mt <- attr(data, "max.time")
      if (!is.null(mt) && is.finite(mt) && mt > 0) {
        start.time <- start.time * mt
        stop.time  <- stop.time  * mt
        if (!is.null(landmark.time)) {
          landmark.time <- landmark.time * mt
        }
      }
    }
  }
  ## Original code silently drops NA ids because which(id == NA) returns integer(0).
  ## Preserve that behavior, but do it up front to avoid carrying empty groups around.
  valid.id <- !is.na(id)
  if (!any(valid.id)) {
    return(data.frame())
  }
  if (!all(valid.id)) {
    row.map    <- which(valid.id)
    id         <- id[valid.id]
    start.time <- start.time[valid.id]
    stop.time  <- stop.time[valid.id]
    event.val  <- event.val[valid.id]
  } else {
    row.map <- seq_along(id)
  }
  n <- length(id)
  x.cols <- setdiff(names(data), req)
  ## One-pass grouping strategy:
  ## - preserve subject encounter order via match(id, id)
  ## - if sorted=FALSE, sort within subject by (start, stop)
  ## - if sorted=TRUE, preserve original within-subject row order
  presorted <- isTRUE(attr(data, "sorted.by.subj.start"))
  grp <- match(id, id)
  if (sorted && presorted) {
    ord <- seq_len(n)
  } else if (sorted) {
    ord <- order(grp, seq_len(n), method = "radix")
  } else {
    ord <- order(grp, start.time, stop.time, method = "radix")
  }
  id.ord    <- id[ord]
  start.ord <- start.time[ord]
  stop.ord  <- stop.time[ord]
  event.ord <- event.val[ord]
  row.ord   <- row.map[ord]
  rr <- rle(id.ord)
  uid.ord <- rr$values
  ends   <- cumsum(rr$lengths)
  starts <- ends - rr$lengths + 1L
  n.id   <- length(uid.ord)
  keep.vec <- rep(TRUE, n.id)
  time.vec <- numeric(n.id)
  event.vec <- integer(n.id)
  base.row <- integer(n.id)
  if (!is.null(landmark.time)) {
    t0 <- as.numeric(landmark.time)
    if (!is.finite(t0)) {
      stop("landmark.time must be finite.")
    }
    t.eff <- if (isTRUE(landmark.use.tminus)) (t0 - eps) else t0
    if (isTRUE(keep.landmark.cols)) {
      landmark.time.vec <- rep(NA_real_, n.id)
      t.end.vec         <- rep(NA_real_, n.id)
      t.end.event.vec   <- rep(NA_real_, n.id)
    }
  } else {
    t0 <- NULL
  }
  ## Scan each subject exactly once using group boundaries.
  for (ii in seq_len(n.id)) {
    s <- starts[ii]
    e <- ends[ii]
    start.slice <- start.ord[s:e]
    stop.slice  <- stop.ord[s:e]
    ## end of follow-up for this id
    t.end <- max(stop.slice, na.rm = TRUE)
    ## last row at end time (used for event indicator)
    last.local <- tail(which(stop.slice >= t.end - eps), 1L)
    last.idx   <- s + last.local - 1L
    event.end  <- as.integer(event.ord[last.idx] > 0)
    ## default: baseline snapshot row
    base.local <- 1L
    time.i     <- t.end
    event.i    <- event.end
    if (!is.null(t0)) {
      ## exclude those not at risk at landmark
      if (t.end <= t0 + eps) {
        keep.vec[ii] <- FALSE
        next
      }
      ## find interval containing t_eff: start <= t_eff < stop
      cand <- which(start.slice <= t.eff & stop.slice > t.eff)
      if (length(cand)) {
        ## if multiple, take the one with largest start
        base.local <- cand[which.max(start.slice[cand])]
      } else {
        ## fallback: last row with stop <= t0 (if exists), else first row
        cand2 <- which(stop.slice <= t0 + eps)
        if (length(cand2)) {
          base.local <- cand2[which.max(stop.slice[cand2])]
        } else {
          base.local <- 1L
        }
      }
      ## define outcome relative to landmark (standard landmarking)
      time.i  <- t.end - t0
      event.i <- as.integer(event.end == 1L && t.end > t0 + eps)
      if (isTRUE(keep.landmark.cols)) {
        landmark.time.vec[ii] <- t0
        t.end.vec[ii]         <- t.end
        t.end.event.vec[ii]   <- if (isTRUE(event.end == 1L)) t.end else NA_real_
      }
    }
    base.row[ii]  <- row.ord[s + base.local - 1L]
    time.vec[ii]  <- time.i
    event.vec[ii] <- event.i
  }
  uid.keep   <- uid.ord[keep.vec]
  base.row   <- base.row[keep.vec]
  time.vec   <- time.vec[keep.vec]
  event.vec  <- event.vec[keep.vec]
  if (!length(time.vec)) {
    return(data.frame())
  }
  if (!is.null(t0) && isTRUE(keep.landmark.cols)) {
    landmark.time.vec <- landmark.time.vec[keep.vec]
    t.end.vec         <- t.end.vec[keep.vec]
    t.end.event.vec   <- t.end.event.vec[keep.vec]
  }
  if (isTRUE(scale)) {
    mx <- max(time.vec, na.rm = TRUE)
    if (is.finite(mx) && mx > 0) {
      time.vec <- time.vec / mx
    }
  }
  out <- NULL
  if (return.type == "survival") {
    out <- data.frame(time = time.vec, event = event.vec, check.names = FALSE)
  }
  ## bind covariates with one direct subset instead of per-id pieces + rbind
  if (length(x.cols)) {
    cov.rows <- data[base.row, x.cols, drop = FALSE]
    rownames(cov.rows) <- NULL
    if (is.null(out)) {
      out <- data.frame(cov.rows, check.names = FALSE)
    } else {
      out <- data.frame(out, cov.rows, check.names = FALSE)
    }
  } else {
    if (is.null(out)) {
      out <- data.frame(check.names = FALSE)
    }
  }
  ## optional helper columns
  if (isTRUE(keep.row_index)) {
    out <- data.frame(row_index = base.row, out, check.names = FALSE)
  }
  if (!is.null(t0) && isTRUE(keep.landmark.cols)) {
    out <- data.frame(landmark_time = landmark.time.vec,
                      t_end = t.end.vec,
                      t_end_event = t.end.event.vec,
                      out,
                      check.names = FALSE)
  }
  if (isTRUE(keep.id)) {
    id.df <- setNames(data.frame(uid.keep, check.names = FALSE), subj.nm)
    out   <- data.frame(id.df, out, check.names = FALSE)
  }
  out
}

dotmatrix.importance.rhf <- function(x,
       vars = NULL,
       top_n_union = 15L,
       variable.labels = NULL,
       time.labels = NULL,
       sort_by = c("q90", "sum", "max", "mean", "median", "alphabetical", "cluster", "none"),
       sort_abs = TRUE,
       transform = c("none", "log10"),
       color_by = c("value", "sign", "single", "none"),
       point_color = "steelblue4",
       value_colors = c("grey85", "steelblue4"),
       sign_colors = c("firebrick3", "grey90", "steelblue4"),
       cex.range = c(0.6, 3.2),
       size.cap = 0.99,
       color.cap = 0.99,
       alpha = 0.9,
       show.grid = TRUE,
       grid.col = "grey92",
       legend = TRUE,
       display.note = TRUE,
       xlab = "",
       ylab = "",
       main = "RHF time-localized VarPro importance",
       axis.cex = 0.7,
       var.cex = 0.7,
       time.label.srt = 45,
       save_plot = FALSE,
       out.file = "rhf_time_varpro_dotmatrix.pdf",
       width = 11,
       height = NULL,
       mar = NULL,
       legend.width = 0.7,
       ...) {
  if (!inherits(x, "importance.rhf")) {
    stop("This function only works for objects of class 'importance.rhf'.")
  }
  sort_by <- match.arg(sort_by)
  transform <- match.arg(transform)
  color_by <- match.arg(color_by)
  cex.range <- as.numeric(cex.range)
  if (length(cex.range) != 2L || any(!is.finite(cex.range)) || any(cex.range < 0)) {
    stop("cex.range must be a length-2 nonnegative numeric vector.")
  }
  mat <- x$importance.matrix
  if (!is.matrix(mat) || !length(mat)) {
    stop("No importance values available to plot.")
  }
  mat <- .rhf_select_dotmatrix_variables(mat = mat,
                                         vars = vars,
                                         top_n_union = top_n_union,
                                         sort_abs = sort_abs)
  var.order <- .rhf_order_dotmatrix_variables(mat = mat,
                                              variable.labels = variable.labels,
                                              sort_by = sort_by,
                                              sort_abs = sort_abs)
  var.order <- .rhf_unique_in_order(var.order)
  mat <- mat[var.order, , drop = FALSE]
  var.codes <- rownames(mat)
  time.codes <- colnames(mat)
  time.values <- x$window.info$time
  var.labels <- .rhf_label_lookup(var.codes, variable.labels, infer_prefix = TRUE)
  var.labels <- .rhf_make_unique_labels(var.labels, var.codes)
  if (is.null(time.labels)) {
    x.labels <- .rhf_format_time_labels(time.values)
  }
  else {
    x.labels <- .rhf_label_lookup(time.codes, time.labels, infer_prefix = FALSE)
  }
  if (color_by == "sign") {
    size.metric <- abs(mat)
    color.metric <- mat
    size.title <- if (transform == "log10") "log10(|Importance| + 1)" else "|Importance|"
    color.title <- "Importance"
  }
  else {
    size.metric <- pmax(mat, 0)
    color.metric <- if (color_by == "value") size.metric else NULL
    size.title <- if (transform == "log10") "log10(Importance + 1)" else "Importance"
    color.title <- size.title
  }
  if (transform == "log10") {
    size.metric <- log10(size.metric + 1)
    if (color_by == "value") {
      color.metric <- size.metric
    }
  }
  size.cap.info <- .rhf_cap_values(size.metric,
                                   cap = size.cap,
                                   symmetric = FALSE,
                                   positive.only = TRUE,
                                   arg = "size.cap")
  size.metric.display <- matrix(size.cap.info$values,
                                nrow = nrow(mat),
                                ncol = ncol(mat),
                                dimnames = dimnames(mat))
  if (color_by == "value") {
    color.cap.info <- .rhf_cap_values(color.metric,
                                      cap = color.cap,
                                      symmetric = FALSE,
                                      positive.only = FALSE,
                                      arg = "color.cap")
    color.metric.display <- matrix(color.cap.info$values,
                                   nrow = nrow(mat),
                                   ncol = ncol(mat),
                                   dimnames = dimnames(mat))
  }
  else if (color_by == "sign") {
    color.cap.info <- .rhf_cap_values(color.metric,
                                      cap = color.cap,
                                      symmetric = TRUE,
                                      positive.only = FALSE,
                                      arg = "color.cap")
    color.metric.display <- matrix(color.cap.info$values,
                                   nrow = nrow(mat),
                                   ncol = ncol(mat),
                                   dimnames = dimnames(mat))
  }
  else {
    color.cap.info <- list(applied = FALSE,
                           label = "none",
                           cap = NA_real_,
                           range = c(NA_real_, NA_real_))
    color.metric.display <- color.metric
  }
  draw <- is.finite(size.metric.display) & (size.metric.display > 0)
  cex.mat <- matrix(0, nrow = nrow(mat), ncol = ncol(mat))
  size.range <- if (any(draw)) range(size.metric.display[draw], na.rm = TRUE) else c(0, 1)
  if (any(draw)) {
    cex.mat[draw] <- .rhf_rescale_from_range(size.metric.display[draw],
                                             from = size.range,
                                             to = cex.range)
  }
  if (color_by == "none") {
    col.mat <- matrix("black", nrow = nrow(mat), ncol = ncol(mat))
  }
  else if (color_by == "single") {
    col.mat <- matrix(point_color, nrow = nrow(mat), ncol = ncol(mat))
  }
  else if (color_by == "value") {
    col.mat <- matrix(.rhf_map_palette(color.metric.display,
                                       value_colors,
                                       symmetric = FALSE),
                      nrow = nrow(mat), ncol = ncol(mat))
  }
  else {
    col.mat <- matrix(.rhf_map_palette(color.metric.display,
                                       sign_colors,
                                       symmetric = TRUE),
                      nrow = nrow(mat), ncol = ncol(mat))
  }
  col.mat[draw] <- grDevices::adjustcolor(col.mat[draw], alpha.f = alpha)
  display.note.text <- NULL
  if (isTRUE(display.note)) {
    if (isTRUE(size.cap.info$applied) &&
        isTRUE(color.cap.info$applied) &&
        identical(size.cap.info$label, color.cap.info$label)) {
      display.note.text <- paste0("plot size/color capped at ", size.cap.info$label)
    }
    else {
      note.parts <- character(0)
      if (isTRUE(size.cap.info$applied)) {
        note.parts <- c(note.parts, paste0("size capped at ", size.cap.info$label))
      }
      if (isTRUE(color.cap.info$applied)) {
        note.parts <- c(note.parts, paste0("color capped at ", color.cap.info$label))
      }
      if (length(note.parts)) {
        display.note.text <- paste0("plot ", paste(note.parts, collapse = "; "))
      }
    }
  }
  if (is.null(height)) {
    height <- max(5.5, 0.22 * nrow(mat) + 1.8)
  }
  if (isTRUE(save_plot)) {
    .rhf_open_plot_device(out.file = out.file, width = width, height = height)
    on.exit(grDevices::dev.off(), add = TRUE)
  }
  old.par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old.par), add = TRUE)
  if (is.null(mar)) {
    mar <- .rhf_dotmatrix_default_mar(var.labels = var.labels,
                                      x.labels = x.labels,
                                      var.cex = var.cex,
                                      axis.cex = axis.cex,
                                      time.label.srt = time.label.srt,
                                      legend = legend)
  }
  if (isTRUE(legend)) {
    graphics::layout(matrix(c(1, 2), nrow = 1L), widths = c(5.6, legend.width))
    on.exit(graphics::layout(1), add = TRUE)
  }
  graphics::par(mar = mar, mgp = c(0.9, 0.12, 0), xpd = NA)
  x.pos <- seq_len(ncol(mat))
  y.pos <- rev(seq_len(nrow(mat)))
  graphics::plot(NA,
                 xlim = c(0.5, ncol(mat) + 0.5),
                 ylim = c(0.5, nrow(mat) + 0.5),
                 xlab = xlab,
                 ylab = ylab,
                 xaxt = "n",
                 yaxt = "n",
                 bty = "n",
                 xaxs = "i",
                 yaxs = "i",
                 main = main)
  if (isTRUE(show.grid)) {
    guide.height <- min(0.16, max(0.07, 0.010 * nrow(mat)))
    graphics::segments(x0 = x.pos,
                       y0 = 0.5,
                       x1 = x.pos,
                       y1 = pmin(nrow(mat) + 0.5, 0.5 + guide.height),
                       col = grid.col,
                       lty = 3)
    graphics::abline(h = seq_len(nrow(mat)), col = grid.col, lty = 3)
    graphics::abline(v = seq_len(ncol(mat)), col = grid.col, lty = 3)
  }
  ## rewrite title over dashed lines
  graphics::title(main = main)
  if (!is.null(display.note.text) && nzchar(display.note.text)) {
    graphics::mtext(display.note.text, side = 3, line = 0.30, adj = 1, cex = 0.75)
  }
  xx <- rep(x.pos, each = nrow(mat))
  yy <- rep(y.pos, times = ncol(mat))
  keep <- as.vector(draw)
  graphics::points(xx[keep],
                   yy[keep],
                   pch = 16,
                   cex = as.vector(cex.mat)[keep],
                   col = as.vector(col.mat)[keep],
                   ...)
  graphics::axis(2, at = y.pos, labels = var.labels, las = 1, tick = FALSE, cex.axis = var.cex)
  .rhf_draw_dotmatrix_xlabels(at = x.pos,
                              labels = x.labels,
                              cex = axis.cex,
                              gap.lines = 0.10,
                              srt = time.label.srt)
  graphics::box()
  if (isTRUE(legend)) {
    graphics::par(mar = c(mar[1L], 0.10, mar[3L], 0.10), xpd = NA)
    size.breaks <- if (any(draw)) {
      .rhf_pretty_breaks(size.metric.display[draw],
                         n = 10L,
                         positive.only = TRUE,
                         symmetric = FALSE)
    }
    else {
      numeric(0)
    }
    color.range <- NULL
    if (color_by == "value" && any(draw)) {
      color.range <- range(color.metric.display[draw], na.rm = TRUE)
    }
    if (color_by == "sign") {
      finite.color <- color.metric.display[is.finite(color.metric.display)]
      if (length(finite.color)) {
        lim <- max(abs(finite.color), na.rm = TRUE)
        color.range <- c(-lim, lim)
      }
    }
    .rhf_draw_dotmatrix_legend(size.breaks = size.breaks,
                               size.range = size.range,
                               size.title = size.title,
                               cex.range = cex.range,
                               alpha = alpha,
                               color_by = color_by,
                               color.range = color.range,
                               color.title = color.title,
                               point_color = point_color,
                               value_colors = value_colors,
                               sign_colors = sign_colors)
  }
  invisible(list(matrix = mat,
                 variables = var.codes,
                 times = time.values,
                 size.metric = size.metric,
                 size.metric.display = size.metric.display,
                 color.metric = color.metric,
                 color.metric.display = color.metric.display,
                 size.cap = size.cap.info,
                 color.cap = color.cap.info))
}
dotmatrix.importance <- dotmatrix.importance.rhf
########################################################################
## plotting method
########################################################################
plot.importance.rhf <- function(x,
                 type = c("dotmatrix", "lines"),
                 vars = NULL,
                 top = 10L,
                 rank.by = c("q90", "median", "mean", "max"),
                 curve = c("step", "line", "lowess"),
                 smooth.f = 2/3,
                 display.cap = 0.99,
                 display.note = TRUE,
                 xlab = NULL,
                 ylab = NULL,
                 lty = 1,
                 lwd = 2.0,
                 ...) {
  type <- match.arg(type)
  if (type == "dotmatrix") {
    if (is.null(xlab)) {
      xlab <- ""
    }
    if (is.null(ylab)) {
      ylab <- ""
    }
    return(dotmatrix.importance.rhf(x = x, vars = vars, xlab = xlab, ylab = ylab, ...))
  }
  rank.by <- match.arg(rank.by)
  curve <- match.arg(curve)
  top <- as.integer(top)[1L]
  if (!is.finite(top) || top < 1L) {
    stop("'top' must be a positive integer.")
  }
  mat <- x$importance.matrix
  if (!is.matrix(mat) || !length(mat)) {
    stop("No importance values available to plot.")
  }
  if (is.null(vars)) {
    score <- .rhf_row_summary(mat, rank.by = rank.by)
    ord <- order(score, decreasing = TRUE, na.last = TRUE)
    vars <- rownames(mat)[head(ord, top)]
  }
  vars <- intersect(as.character(vars), rownames(mat))
  if (!length(vars)) {
    stop("No requested variables found in the importance object.")
  }
  zz.raw <- mat[vars, , drop = FALSE]
  xx <- x$window.info$time
  if (is.null(xlab)) {
    xlab <- "Time"
  }
  if (is.null(ylab)) {
    ylab <- "Localized Importance"
  }
  cap.info <- .rhf_cap_values(zz.raw,
                              cap = display.cap,
                              symmetric = any(zz.raw[is.finite(zz.raw)] < 0),
                              positive.only = FALSE,
                              arg = "display.cap")
  zz <- matrix(cap.info$values,
               nrow = nrow(zz.raw),
               ncol = ncol(zz.raw),
               dimnames = dimnames(zz.raw))
  ylim <- range(zz, na.rm = TRUE)
  if (!all(is.finite(ylim))) {
    ylim <- c(0, 1)
  }
  cols <- seq_len(nrow(zz))
  lty <- rep_len(lty, nrow(zz))
  lwd <- rep_len(lwd, nrow(zz))
  if (curve %in% c("step", "line")) {
    graphics::matplot(xx,
                      t(zz),
                      type = if (curve == "step") "s" else "l",
                      lty = lty,
                      lwd = lwd,
                      col = cols,
                      xlab = xlab,
                      ylab = ylab,
                      ylim = ylim,
                      ...)
  }
  else {
    graphics::plot(NA,
                   xlim = range(xx, na.rm = TRUE),
                   ylim = ylim,
                   xlab = xlab,
                   ylab = ylab,
                   type = "n",
                   ...)
    for (i in seq_len(nrow(zz))) {
      ok <- is.finite(xx) & is.finite(zz[i, ])
      if (sum(ok) < 2L) {
        next
      }
      sm <- stats::lowess(xx[ok], zz[i, ok], f = smooth.f, iter = 0)
      graphics::lines(sm$x, sm$y, col = cols[i], lty = lty[i], lwd = lwd[i])
    }
  }
  if (isTRUE(display.note) && isTRUE(cap.info$applied)) {
    graphics::mtext(paste0("plot capped at ", cap.info$label),
                    side = 3,
                    line = 0.30,
                    adj = 1,
                    cex = 0.75)
  }
  graphics::legend("topright",
                   legend = vars,
                   col = cols,
                   lty = lty,
                   lwd = lwd,
                   bty = "n")
  invisible(zz)
}

get.data.pass.bits <- function (data.pass) {
  if (!is.null(data.pass)) {
    if (data.pass == TRUE) {
      data.pass <- 2^15
    }
    else if (data.pass == FALSE) {
      data.pass <- 0
    }
    else {
      stop("Invalid choice for 'data.pass' option:  ", data.pass)
    }
  }
  else {
    stop("Invalid choice for 'data.pass' option:  ", data.pass)
  }
  return (data.pass)
}
get.experimental.bits  <- function(experimental.bits, trace) {
  ## New protocol is to use the uniform hazard estimator. This is
  ## UBK's version of the hazard. It avoids the issue of having OOB
  ## unscaled risk inf values, that we experienced due to log(0) where
  ## the column (corresponding to a time interest point) is zero for
  ## the OOB subject.
  if (!is.null(experimental.bits)) {
    if (experimental.bits < 0L) {
      stop("bits must be between 0 and 255.")
    }
    else if (experimental.bits > 255L) {
      stop("bits must be between 0 and 255.")
    }
    else if (experimental.bits == 0) {
        ## This is for RSF, no frills.
        experimental.bits <- 0
    }
    else if (is.bit(experimental.bits, 0) && is.bit(experimental.bits, 1)) {
        stop(paste(paste0("Endpoint Estimation:  ", is.bit(experimental.bits, 0), "\n"),
                   paste0("Uniform Estimation:   ", is.bit(experimental.bits, 1), "\n"),
                   paste0("only one protocol bit allowed."),
                   sep = "\n"))
    }
  }
  else {
    ## Default is UBK's rule when null.
    experimental.bits <- 2^1
  }
  if (isTRUE(trace)) {
      message(
          paste(
              paste0("Endpoint Estimation:  ", is.bit(experimental.bits, 0)),
              paste0("Uniform Estimation:   ", is.bit(experimental.bits, 1)),
              paste0("Uniform Head:         ", is.bit(experimental.bits, 2)),
              paste0("Uniform Tail:         ", is.bit(experimental.bits, 3)),
              sep = "\n"
          )
      )
  }
  return (experimental.bits)
}
# Returns TRUE if the given 0-7 bit is set in an 8-bit integer, FALSE otherwise.
# - value: integer (vectorized); values outside 0..255 are masked to 8 bits
# - bit: single integer in 0..7 (0 = least significant bit)
is.bit <- function(value, bit) {
  if (length(bit) != 1L || is.na(bit) || bit != as.integer(bit) || bit < 0L || bit > 7L) {
    stop("`bit` must be a single integer in 0..7.")
  }
  v <- bitwAnd(as.integer(value), 255L)         # clamp to 8 bits
  m <- bitwShiftL(1L, as.integer(bit))          # mask for the requested bit
  bitwAnd(v, m) != 0L                           # TRUE if bit set, FALSE otherwise (NA propagates)
}
get.tree.index <- function(get.tree, ntree) {
  ## NULL --> default setting
  if (is.null(get.tree)) {
    rep(1, ntree)
  }
  ## the user has specified a subset of trees
  else {
    pt <- get.tree >=1 & get.tree <= ntree
    if (sum(pt) > 0) {
      get.tree <- get.tree[pt]
      get.tree.temp <- rep(0, ntree)
      get.tree.temp[get.tree] <- 1
      get.tree.temp
    }
    else {
      rep(1, ntree)
    }
  }
}
get.block.size.bits <- function (block.size, ntree) {
    ## Check for user silliness.
    if (!is.null(block.size)) {
        if ((block.size < 1) || (block.size > ntree)) {
            block.size <- ntree
        }
        else {
            block.size <- round(block.size)
        }
    }
    else {
        block.size <- ntree
    }
    return (block.size)
}
get.bootstrap.bits <- function (bootstrap) {
  if (bootstrap == "none") {
    bootstrap <- 0
  }
  else if (bootstrap == "by.root") {
    bootstrap <- 2^19
  }
  else if (bootstrap == "by.user") {
    bootstrap <- 2^20
  }
  else {
    stop("Invalid choice for 'bootstrap' option:  ", bootstrap)
  }
  return (bootstrap)
}
## convert ensemble option into native code parameter.
get.ensemble.bits <- function (ensemble) {
  if (ensemble == "oob") {
    ensemble <- 2^1
  }
  else if (ensemble == "inbag") {
    ensemble <- 2^0
  }
  else if (ensemble == "all") {
    ensemble <- 2^0 + 2^1
  }    
  else {
      ## For testing purposes, we allow the ensemble to be omitted altogether.
      ensemble <- 0
  }
  return (ensemble)
}
get.empirical.risk.bits <-  function (empirical.risk) {
  if (empirical.risk) {
    return (2^18)
  }
  else {
    return (0)
  }
}
get.forest.bits <- function (forest) {
  ## Convert forest option into native code parameter.
  if (!is.null(forest)) {
    if (forest == TRUE) {
      forest <- 2^5
    }
    else if (forest == FALSE) {
      forest <- 0
    }
    else {
      stop("Invalid choice for 'forest' option:  ", forest)
    }
  }
  else {
    stop("Invalid choice for 'forest' option:  ", forest)
  }
  return (forest)
}
get.membership.bits <- function (membership) {
  ## Convert option into native code parameter.
  bits <- 0
  if (!is.null(membership)) {
    if (membership == TRUE) {
      bits <- 2^6
    }
    else if (membership != FALSE) {
      stop("Invalid choice for 'membership' option:  ", membership)
    }
  }
  else {
    stop("Invalid choice for 'membership' option:  ", membership)
  }
  return (bits)
}
get.perf <-  function (perf, family) {
  ## Deal with non-classification
  if (family != "class") {
    if (is.null(perf)) {
      return("default")
    }
    perf <- match.arg(perf, c("none", "default", "standard"))
    if (perf == "standard") {
      perf <- "default"
    }
    return(perf)
  }
  else {
      ## Deal with classification
      if (is.null(perf)) {
          return("default")
      }
      perf <- match.arg(perf, c("none", "default", "standard", "misclass", "brier", "gmean"))
      if (perf == "standard" || perf == "misclass") {
          perf <- "default"
      }
      return(perf)
  }
}
get.perf.bits <- function (perf) {
  if (perf == "default") {
    return (2^2)
  }
  else if (perf == "gmean") {
    return (2^2 + 2^14)
  }
  else if (perf == "brier") {
    return (2^2 + 2^3)
  }
  else {#everything else becomes "none"
    return (0)
  }
}
get.rf.cores <- function () {
    ## PART I:  Two ways for the user to specify cores:
    ## (1) R-option "rf.cores"
    ## (2) Shell-environment-option "RF_CORES"
    if (is.null(getOption("rf.cores", NULL))) {
        if (!is.na(as.numeric(Sys.getenv("RF_CORES")))) {
            options(rf.cores = as.integer(Sys.getenv("RF_CORES")))
        }
    }
    ## If the user has set the cores using either of the two methods, we respect it.
    if (!is.null(getOption("rf.cores", NULL))) {
        return (getOption("rf.cores"))
    }
    ## PART II:  Respect R CMD check limit
    chk <- tolower(Sys.getenv("_R_CHECK_LIMIT_CORES_", ""))
    if (nzchar(chk) && chk != "false") {
        ## under R CMD check --as-cran (CRAN sets this)
        return(2L)
    }
    ## PART III:  Use everything.
    return (-1L)
}
get.weight <- function(weight, n) {
  ## set the default weight
  if (!is.null(weight)) {
    if (any(weight < 0)      ||
      length(weight) != n  ||
      any(is.na(weight))) {
        stop("Invalid weight vector specified.")
    }
  }
  else {
    weight <- rep(1, n)
  }
  return (weight)
}
## Real time predicton option:
is.hidden.rt <-  function(dots) {
  if (is.null(dots$real.time)) {
    FALSE
  }
  else {
    as.logical(as.character(dots$real.time))
  }
}
get.rt.bits  <- function(real.time) {
  if (real.time == TRUE) {
    bits  <- 2^7
  }
  else {
    bits  <- 0
  }
  return (bits)
}
is.hidden.rt.opt  <- function(dots) {
  if (is.null(dots$real.time.options)) {
    list(port = 6666, time.out = 15)      
  }
  else {
    dots$real.time.options
  }
}
is.hidden.wmode <- function(dots) {
    if (is.null(dots$wmode)) {
        wmode <- 2
    }
    else {
        wmode <- dots$wmode
    }
    return (wmode)
}
get.wmode.bits <- function(wmode) {
    ## Only (1, 2, 3) are valid modes. We shift them to the local option bits. No error checking.
    wmode <- bitwShiftL(wmode, 16)
    return (wmode)
}
## convert samptype option into native code parameter.
get.samptype.bits <- function (samptype) {
  if (samptype == "swr") {
    bits <- 0
  }
  else if (samptype == "swor") {
    bits <- 2^12
  }
  else {
    stop("Invalid choice for 'samptype' option:  ", samptype)
  }
  return (bits)
}
get.seed <- function (seed) {
  if ((is.null(seed)) || (abs(seed) < 1)) {
    seed <- runif(1,1,1e6)
  }
  seed <- -round(abs(seed))
  return (seed)
}
get.trace <- function (do.trace) {
  ## Convert trace into native code parameter.
  if (!is.logical(do.trace)) {
    if (do.trace >= 1) {
      do.trace <- round(do.trace)
    }
    else {
      do.trace <- 0
    }
  }
  else {
    do.trace <- 1 * do.trace
  }
  return (do.trace)
}

plot.rhf <- function(x, idx = NULL, scale.hazard = FALSE, 
                     ngrid = 30, bass = 0, q = 0.99, grid = FALSE, 
                     col = NULL, lty = NULL, legend.loc = "topright", 
                     jitter.factor = 1, lwd = 4, 
                     hazard.only = TRUE, legend.show = TRUE, ...) {
  ## Preliminary
  time.interest <- x$time.interest
  if (is.null(idx)) idx <- x$ensemble.id[1]
  if (sum(idx %in% x$ensemble.id) == 0L) stop("idx does not match id values in x$id")
  idx <- match(idx, x$ensemble.id)
  n.idx <- length(idx)
  dots <- list(...)
  if (is.null(col)) col <- seq_len(n.idx)
  if (is.null(lty)) lty <- rep(1, n.idx)
  ## pull the ensemble estimators
  get_mat <- function(name_oob, name_test) {
      if (!is.null(x[[name_oob]])) return(x[[name_oob]])
      if (!is.null(x[[name_test]])) return(x[[name_test]])
      NULL
  }
  chf <- get_mat("chf.oob", "chf.test")
  haz <- get_mat("hazard.oob", "hazard.test")
  ## missing values are set to zero
  chf[is.na(chf)] <- 0
  haz[is.na(haz)] <- 0
  if (is.logical(scale.hazard)) {
    if (scale.hazard) {
      haz <- diff(c(0, time.interest)) * haz
    }
  } else if (is.numeric(scale.hazard) && length(scale.hazard) == 1 && is.finite(scale.hazard)) {
    haz <- scale.hazard * haz
  } else {
    stop("scale.hazard must be either logical or a finite numeric scalar.")
  }
  ## Compute plot ranges
  ranges <- lapply(idx, function(i) {
    h <- haz[i, ]
    pt <- h <= quantile(h, q, na.rm = TRUE) 
    h.smooth <- supsmu(time.interest[pt], h[pt], bass = bass)$y
    ## auxiliary equally spaced grid (used for overlays)
    time.grid <- unique(c(0, seq(0, max(time.interest), length = ngrid)))
    if (!hazard.only) {
      H <- chf[i, ]
      H.grid <- c(0, H)[1 + sIndex(time.interest, time.grid)]
      H.grid.smooth <- supsmu(time.grid, H.grid)$y
      delta <- max(diff(time.grid))
      h.grid.smooth <- supsmu(time.grid, c(0, diff(H.grid.smooth) / delta), bass = bass)$y
    } else {
      H <- H.grid.smooth <- NULL
      ## NEW: make ngrid meaningful in hazard-only mode - smooth hazard on a uniform grid
      h.on.grid <- stats::approx(x = time.interest[pt], y = h[pt],
                          xout = time.grid, method = "linear", rule = 2, ties = mean)$y
      h.grid.smooth <- supsmu(time.grid, h.on.grid, bass = bass)$y
    }
    list(H = H,
         h.smooth = h.smooth,
         h.grid.smooth = h.grid.smooth,
         H.grid.smooth = H.grid.smooth,
         time.grid = time.grid)  ## NEW: carry the grid back for plotting
  })
  ## Extract or compute xlim/ylim
  xlim <- if (!is.null(dots$xlim)) dots$xlim else range(time.interest)
  ylim <- if (!is.null(dots$ylim)) dots$ylim else {
    range(unlist(
      lapply(ranges, function(xi) {
        if (hazard.only) {
          c(xi$h.smooth, xi$h.grid.smooth)     ## includes grid smoother in hazard-only
        } else {
          c(xi$H, xi$H.grid.smooth, xi$h.smooth, xi$h.grid.smooth)
        }
      })
    ), na.rm = TRUE)
  }
  dots$xlim <- NULL
  dots$ylim <- NULL
  ## Main plot
  ylab.default <- if (hazard.only) "Hazard" else "CHF + hazard"
  do.call("plot", c(list(x = NA, y = NA,
                         xlim = xlim, ylim = c(max(0, ylim[1]), ylim[2]),
                         xlab = "time", ylab = ylab.default), dots))
  ## Loop through cases
  invisible(lapply(seq_along(idx), function(k) {
    i <- idx[k]
    h <- haz[i, ]
    pt <- h <= quantile(h, q, na.rm = TRUE)
    h.smooth <- supsmu(time.interest[pt], h[pt], bass = bass)$y
    ## plot hazard
    t.jittered <- jitter(time.interest[pt], factor = jitter.factor)
    lines(t.jittered, h[pt], type = "h", col = adjustcolor(col[k], alpha.f = 0.5), lwd = lwd)
    lines(time.interest[pt], h.smooth, type = "s", col = col[k], lty = if (hazard.only) 1 else 3)
    if (!hazard.only) {
      H <- chf[i, ]
      rr <- ranges[[k]]
      lines(time.interest, H, type = "s", col = col[k])
      if (grid) {
        lines(rr$time.grid, rr$H.grid.smooth, type = "s", col = col[k], lty = 3)
        lines(rr$time.grid, rr$h.grid.smooth, type = "s", col = col[k], lty = 2)
      }
    } else {
      ## NEW: allow a dashed grid-based hazard overlay when hazard-only = TRUE
      if (grid) {
        rr <- ranges[[k]]
        lines(rr$time.grid, rr$h.grid.smooth, type = "s", col = col[k], lty = 2)
      }
    }
  }))
  my_rug(x)
  ## Optional legend
  if (legend.show && n.idx > 1) {
    labels <- if (!is.null(names(idx))) names(idx) else paste("Case", idx)
    legend(legend.loc, legend = labels, col = col, lty = lty, bty = "n")
  }
}
####################################################################
##
## TDC Helper Graphical Functions
##
####################################################################
## Time alignment helper
sIndex <- function(x, y) {
  sapply(seq_along(y), function(j) sum(x <= y[j]))
}
## Rug function with clipping protection
my_rug <- function(x, ticksize = 0.02, nmax = 350) {
  rug.time <- unique(x$yvar$stop[x$yvar$event == 1])
  rug.time <- sample(rug.time, size = min(length(rug.time), nmax), replace = FALSE)
  xlims <- par("usr")[1:2]
  rug.time <- rug.time[rug.time >= xlims[1] & rug.time <= xlims[2]]
  suppressWarnings(rug(rug.time, ticksize = ticksize))
}

get.lot <- function(hcut = 1,
                    treesize = NULL,
                    tune = TRUE,
                    lag = NULL,
                    strikeout = NULL,
                    max.two   = NULL,
                    max.three = NULL,
                    max.four  = NULL) {
  if (is.null(treesize) || (!is.function(treesize) && !is.numeric(treesize))) {
    treesize <- function(n, tdc) {
      if (tdc) {
        min(30, n / 5)
      }
      else {
        max(100, n / 5)
      }
    }
  }
  if (is.null(lag)) {
    lag <- 50
  }
  if (tune == FALSE) {
    lag <- 0
  }
  if (is.null(strikeout)) {
    strikeout <- 7
  }
  if (is.null(max.two)) {
    max.two <- 200
  }
  if (is.null(max.three)) {
    max.three <- 60
  }
  if (is.null(max.four)) {
    max.four <- 30
  }
  list(hcut = as.integer(hcut),  
       treesize = ifelse(is.numeric(treesize), as.integer(treesize), treesize),
       lag = as.integer(lag),
       strikeout = as.integer(strikeout),
       max.two = as.integer(max.two),
       max.three = as.integer(max.three),
       max.four = as.integer(max.four)
       )
}
get.splitinfo <- function(formula.detail, splitrule, nsplit) {
  ## CAUTION: HARD CODED ON NATIVE SIDE
  splitrule.names <- c("sg.regr",             ##  1
                       "sg.class",            ##  2
                       "cart.regr",           ##  3
                       "cart.class",          ##  4
                       "cart.random",         ##  5
                       "sg.tdc",              ##  6
                       "sg.nelson.aalen")     ##  7  
  ## set the split rule
  ## Note no error checking for now. TBD
  if (is.null(splitrule)) {
      splitrule <- "sg.tdc"
  }
  splitrule.idx <- which(splitrule.names == splitrule)
  if (length(splitrule.idx) == 0) {
    stop("invalid split rule:  ", splitrule)
  }
  ## set nsplit
  if(!is.null(nsplit)) {
    nsplit <- round(nsplit)    
    if (nsplit < 0) {
      stop("invalid nsplit value: set nsplit >= 0")
    }
  }
  else {
    nsplit <- 10
  }
  list(splitrule = splitrule, index = splitrule.idx, nsplit = nsplit)
}

###################################################################
### Hazard simulation machinery 
###################################################################
####################################################################
## basic theory
## given for t, but applies for x-variables by conditioning on x
##
## relationship between F and CHF
## F(t) = 1 - exp(-H(t))
##
## probability integral transform
## Y           ~  F
## F(Y)        ~  U[0,1]
## 1 - F(Y)    ~  U[0,1]
## exp(-H(T))  ~  U[0,1]
## H(T)        ~  -log(U[0,1])
## T           ~  H^{-1}(-log{U[0,1]})
##
## illustration for Cox PH model
## h(t|x) = h_0(t) exp(b x)
## H(t|x) = H_0(t) exp(b x)
## T|x ~ H_0^{-1}(-log{U[0,1]} * exp(-b x))
##
## time scaling convention (used below)
## Let t* = s * t be scaled time with scale factor s.
## Then h*(t*) = h(t* / s) / s and H*(t*) = H(t* / s).
## Our code applies this consistently when scale != 1.
####################################################################
hazard.simulation <- function(type = 1,
                              n = 500, p = 10, nrecords = 7,
                              scale = FALSE, ngrid = 1e5, ...) {
  ## Allow numeric or character type specification
  hazard.types <- c("hazard.1", "hazard.2", "hazard.3")
  if (is.numeric(type)) {
    if (type < 1 || type > length(hazard.types)) {
      stop("Invalid numeric type. Must be 1, 2, or 3.")
    }
    type <- hazard.types[type]
  } else {
    type <- match.arg(type, choices = hazard.types)
  }
  ## Dispatch to selected hazard function
  sim_data <- switch(type,
    hazard.1 = hazard.1(n = n, p = p, nrecords = nrecords, scale = scale, ngrid = ngrid, ...),
    hazard.2 = hazard.2(n = n, p = p, nrecords = nrecords, scale = scale, ...),
    hazard.3 = hazard.3(n = n, p = p, nrecords = nrecords, scale = scale, ...)
  )
  sim_data
}
##----------------------------------------------------------
## SIMULATION 1
##
## h(t|x) = (1 + z_2 * t) * exp(z_1 + z_2 * t)
## where z_1 = 1.5 * x_1,  z_2 = 1.5 * (x_4 + x_5)
##
## integration by parts gives
## H(t|x) = t * exp(z_1 + z_2 * t)
##
## write z2(t) = z_2 * t = 1.5 * x_4(t) + 1.5 * x_5(t)
## where x_4(t) = x_4 * t, x_5(t) = x_5 * t
##
## p = dimension
## nrecords = number of records (nr) per case: 1 + Bin(nr - 1, 0.7)
## scale:
##   FALSE -> no scaling
##   TRUE  -> scale times to [0,1] by dividing by max(stop)
##   numeric s -> multiply times by s (e.g., s=365 to map to days)
## ngrid kept for backward compatibility (not used by new inversion)
##----------------------------------------------------------
hazard.1 <- function(n = 500, p = 10, nrecords = 7, scale = FALSE, ngrid = 1e5) {
  ## static covariates
  p <- max(5, p)
  x <- matrix(runif(n * p), n)
  ## numerical inversion for CHF using uniroot with automatic bracketing
  ## Solve t * exp(Z1 + Z2 * t) = -log(U)
  invert.H <- function(Z1, Z2, U, upper = 100) {
    rhs <- -log(U)
    if (Z2 == 0) return(rhs * exp(-Z1))          # closed form when Z2=0
    f   <- function(t) t * exp(Z1 + Z2 * t) - rhs
    ub  <- upper
    ## expand upper bound until f(ub) >= 0 (monotone increasing)
    while (f(ub) < 0) ub <- ub * 2
    stats::uniroot(f, c(0, ub), tol = 1e-10)$root
  }
  ## simulate number of records per case
  if (nrecords > 1) {
    id.nrecords <- 1 + rbinom(n, size = (nrecords - 1), prob = .7)
  } else {
    id.nrecords <- rep(1, n)
  }
  ## build up the data, one case at a time
  dta <- do.call(rbind, lapply(seq_len(n), function(i) {
    ## simulate the event and censoring time 
    x4 <- x[i, 4]
    x5 <- x[i, 5]
    z1 <- 1.5 * x[i, 1]
    z2 <- 1.5 * (x4 + x5)
    Tm <- invert.H(z1, z2, runif(1))
    Ce <- -1.5 * log(runif(1))
    ## observed time 
    observed.time <- pmin(Tm, Ce)
    ## time dependent covariates
    ## sample random points between 0 and the observed time
    tseq <- c(sort(runif(max(1, id.nrecords[i] - 1), 0, observed.time)), observed.time)
    xtd  <- (x4 + x5) * tseq
    ## start and stop times
    start <- c(0, head(tseq, -1))
    stop  <- tseq
    ## event occurrence flags per interval
    event <- 1L * ((Tm <= Ce) & Tm <= stop)
    ## assemble the data for the case
    dta.i <- data.frame(id = i,
                        start = start,
                        stop  = stop,
                        event = event,
                        xtd   = xtd)
    data.frame(dta.i, x = do.call(cbind, lapply(x[i, ], function(xx) rep(xx, id.nrecords[i]))))
  }))
  ## compute scale factor s:
  ## FALSE -> 1, TRUE -> 1 / max(stop), numeric -> that numeric
  s <- if (isTRUE(scale)) {
    1 / max(dta$stop)
  } else if (isFALSE(scale)) {
    1
  } else if (is.numeric(scale) && length(scale) == 1L) {
    as.numeric(scale)
  } else {
    1
  }
  dta$start <- s * dta$start
  dta$stop  <- s * dta$stop
  ## true hazard for data (ID-aligned; accounts for time scaling)
  haz <- function(time, x) {
    ids  <- x[, "id"]
    ## original-time evaluation point for theory-based hazard
    time <- sort(time)
    t0   <- time / s                         # map scaled time back to original
    X    <- x[match(unique(ids), ids), grepl("^x\\.", names(x)), drop = FALSE]
    out  <- do.call(rbind, lapply(seq_len(nrow(X)), function(i) {
      x4 <- X[i, 4]; x5 <- X[i, 5]
      z1 <- 1.5 * X[i, 1]
      z2 <- 1.5 * (x4 + x5)
      h0 <- (1 + z2 * t0) * exp(z1 + z2 * t0)  # hazard in original time
      h0 / s                                   # transform to scaled-time hazard
    }))
    rownames(out) <- as.character(unique(ids))
    out
  }
  list(dta = dta, haz = haz, scale = s)
}
##----------------------------------------------------------
## SIMULATION 2 
##
## if x_1 <= .5 & x_2<=.5:
## h(t|x) = 0 if 0.5 <= t <= 2.5 else h(t|x) = exp(.2 * x_2)
## i.e. H(t|x)= H_0(t) * g1(x), where g_1(x) = exp(.2 * x_2)
##
## H_0(t) = t                      for 0<t<0.5
## H_0(t) = .5                     for 0.5<=t<=2.5
## H_0(t) = t - 2                  for 2.5<t
##
## otherwise if x_1 and x_2 are not as above: 
## h(t|x) = 0 if 2.5 <= t <= 4.5 else h(t|x) = exp(.2 * x_3 + x_4(t))
## where x_4(t) is the time dependent covariate
##
## x_4(t) = x_4 * log(t), x_4 is a 0/1 binary variable
##
## i.e. if  t < 2.5 or  t > 4.5
## h(t|x) = t ^ (x_4) * exp(.2 * x_3)
##
## So H(t|x) = H_0(t) * g2(x), g2(x) = exp(.2 * x_3)
##
## thus if x_4 = 0:
## H_0(t) = t                      for 0<t<2.5
## H_0(t) = 2.5                    for 2.5<=t<=4.5
## H_0(t) = t - 2                  for 4.5<t
## 
## thus if x_4 = 1:
## H_0(t) = 0.5 * t^2                                for 0<t<2.5
## H_0(t) = 3.125                                    for 2.5<=t<=4.5
## H_0(t) = 0.5 t^2 - 7                               for 4.5<t
##----------------------------------------------------------
H2.case1.inverse <- function(x) {
  ifelse(x < 0.5, x,
         ifelse(x == 0.5, 0.5, x + 2))
}
H2.case20.inverse <- function(x) {
  ifelse(x < 2.5, x,
         ifelse(x == 2.5, 2.5, x + 2))
}
H2.case21.inverse <- function(x) {
  ifelse(x < 3.125, sqrt(2 * x),
         ifelse(x == 3.125, 2.5, sqrt(2 * x + 14)))
}
hazard.2 <- function(n = 500, p = 10, nrecords = 7, scale = FALSE) {
  ## static covariates
  p <- max(4, p)
  x <- matrix(runif(n * p), n)
  x[, 4] <- rbinom(n, 1, .8)
  ## simulate number of records per case
  if (nrecords > 1) {
    id.nrecords <- 1 + rbinom(n, size = (nrecords - 1), prob = .7)
  } else {
    id.nrecords <- rep(1, n)
  }
  ## build up the data, one case at a time
  dta <- do.call(rbind, lapply(seq_len(n), function(i) {
    ## simulate the censoring time and event time
    x4  <- x[i, 4]
    x.pt <- x[i, 1] <= 0.5 & x[i, 2] <= 0.5
    ## case 1
    if (x.pt) {
      z  <- -log(runif(1)) * exp(-.2 * x[i, 2])
      Tm <- H2.case1.inverse(z)
    }
    ## case 2
    else {
      z  <- -log(runif(1)) * exp(-.2 * x[i, 3])
      Tm <- if (x4 == 0) H2.case20.inverse(z) else H2.case21.inverse(z)
    }
    ## censoring 
    Ce <- -5.5 * log(runif(1))
    ## observed time
    observed.time <- pmin(Tm, Ce)
    ## time dependent covariates
    ## sample random points between 0 and the observed time
    tseq <- c(sort(runif(max(1, id.nrecords[i] - 1), 0, observed.time)), observed.time)
    eps  <- 1e-8
    xtd  <- x4 * log(pmax(tseq, eps))
    ## start and stop times
    start <- c(0, head(tseq, -1))
    stop  <- tseq
    ## event occurrence flags per interval
    event <- 1L * ((Tm <= Ce) & Tm <= stop)
    ## assemble the data for the case
    dta.i <- data.frame(id    = i,
                        start = start,
                        stop  = stop,
                        event = event,
                        xtd   = xtd)
    data.frame(dta.i, x = do.call(cbind, lapply(x[i, ], function(xx) rep(xx, id.nrecords[i]))))
  }))
  ## compute scale factor s
  s <- if (isTRUE(scale)) {
    1 / max(dta$stop)
  } else if (isFALSE(scale)) {
    1
  } else if (is.numeric(scale) && length(scale) == 1L) {
    as.numeric(scale)
  } else {
    1
  }
  dta$start <- s * dta$start
  dta$stop  <- s * dta$stop
  ## true hazard for data (ID-aligned; accounts for time scaling)
  haz <- function(time, x) {
    ids  <- x[, "id"]
    time <- sort(time)
    t0   <- time / s
    X    <- x[match(unique(ids), ids), grepl("^x\\.", names(x)), drop = FALSE]
    out  <- do.call(rbind, lapply(seq_len(nrow(X)), function(i) {
      x.pt <- X[i, 1] <= 0.5 & X[i, 2] <= 0.5
      sapply(t0, function(tm) {
        if (x.pt) {
          if (0.5 <= tm && tm <= 2.5) 0 else exp(.2 * X[i, 2])
        } else {
          if (2.5 <= tm && tm <= 4.5) 0 else (tm ^ X[i, 4]) * exp(.2 * X[i, 3])
        }
      })
    }))
    ## transform to scaled-time hazard
    out <- out / s
    rownames(out) <- as.character(unique(ids))
    out
  }
  list(dta = dta, haz = haz, scale = s)
}
##----------------------------------------------------------
## SIMULATION 3
##
## Time-varying covariates with stochastic trajectories:
##   X4_i(t) = X4_i + A4_i * t
##   X5_i(t) = X5_i + A5_i * t
## where (A4_i, A5_i) are subject-specific random slopes.
##
## We draw
##   A4_i ~ Normal(a4, a4_sd^2),   A5_i ~ Normal(a5, a5_sd^2),
## independently across subjects. Setting a4_sd = a5_sd = 0 recovers the
## original deterministic-trajectory version.
##
## Hazard:
##   h3(t | X_i(t)) = (1 + t) * exp( beta1 * X1 + beta2 * X2
##                                   + beta3 * X4_i(t) + beta4 * X5_i(t)
##                                   + beta5 * X1 * X4_i(t)
##                                   + beta6 * X2^2 )
##
## Linear predictor:
##   eta_i(t) = theta0_i + lambda_i * t
## where
##   theta0_i = beta1*X1 + beta2*X2 + beta3*X4 + beta4*X5
##              + beta5*X1*X4 + beta6*X2^2
##   lambda_i = beta3*A4_i + beta4*A5_i + beta5*A4_i*X1
##
## CHF:
##   H_i(t) = exp(theta0_i) * J(t; lambda_i),
##   with closed-form J(t; lambda) as below.
##
## Event times obtained by inversion H_i(T_i) = -log U.
##----------------------------------------------------------
hazard.3 <- function(n = 500, p = 10, nrecords = 7, scale = FALSE,
                     a4 = 2.5, a5 = -1.5,
                     a4_sd = 1.25, a5_sd = 1.25) {
  if (a4_sd < 0 || a5_sd < 0) stop("hazard.3: 'a4_sd' and 'a5_sd' must be nonnegative.")
  ## static covariates
  p <- max(5, p)
  x <- matrix(runif(n * p), n)
  ## number of records per subject
  if (nrecords > 1) {
    id.nrecords <- 1 + rbinom(n, size = (nrecords - 1), prob = 0.7)
  } else {
    id.nrecords <- rep(1, n)
  }
  ## coefficients
  beta1 <- 0.5
  beta2 <- -0.5
  beta3 <- 0.8
  beta4 <- 0.5
  beta5 <- 0.7
  beta6 <- 0.5
  ## subject-specific slopes (stochastic trajectories)
  a4_i <- stats::rnorm(n, mean = a4, sd = a4_sd)
  a5_i <- stats::rnorm(n, mean = a5, sd = a5_sd)
  ## optional truncation (helps avoid extreme slopes with small n)
  if (a4_sd > 0) a4_i <- pmin(pmax(a4_i, a4 - 4 * a4_sd), a4 + 4 * a4_sd)
  if (a5_sd > 0) a5_i <- pmin(pmax(a5_i, a5 - 4 * a5_sd), a5 + 4 * a5_sd)
  ## name slopes by id for safe lookup inside haz()
  names(a4_i) <- as.character(seq_len(n))
  names(a5_i) <- as.character(seq_len(n))
  ## subject-specific CHF in original time (unscaled)
  H_fun_i <- function(t, xi, a4i, a5i) {
    if (t <= 0) return(0)
    x1 <- xi[1]
    x2 <- xi[2]
    x4 <- xi[4]
    x5 <- xi[5]
    theta0 <- beta1 * x1 + beta2 * x2 +
              beta3 * x4 + beta4 * x5 +
              beta5 * x1 * x4 + beta6 * x2^2
    lambda_i <- beta3 * a4i + beta4 * a5i + beta5 * a4i * x1
    ## J(t; lambda) = Integral_0^t (1 + u) exp(lambda u) du
    if (abs(lambda_i) < 1e-8) {
      ## limit lambda -> 0
      J <- t + 0.5 * t^2
    } else {
      J <- ((lambda_i * t + lambda_i - 1) * exp(lambda_i * t) - (lambda_i - 1)) / (lambda_i^2)
    }
    exp(theta0) * J
  }
  ## build counting-process data
  dta <- do.call(rbind, lapply(seq_len(n), function(i) {
    xi  <- x[i, ]
    a4i <- a4_i[i]
    a5i <- a5_i[i]
    ## simulate event time via inversion
    Ui     <- runif(1)
    target <- -log(Ui)
    t_upper <- 1
    H_upper <- H_fun_i(t_upper, xi, a4i, a5i)
    while (H_upper < target && t_upper < 1e3) {
      t_upper <- 2 * t_upper
      H_upper <- H_fun_i(t_upper, xi, a4i, a5i)
    }
    if (H_upper < target) {
      Tm <- Inf
    } else {
      Tm <- stats::uniroot(function(t) H_fun_i(t, xi, a4i, a5i) - target,
                           interval = c(0, t_upper),
                           tol = 1e-8)$root
    }
    ## independent censoring
    Ce <- -4 * log(runif(1))
    observed.time <- min(Tm, Ce)
    ## sampling record times
    nrec <- id.nrecords[i]
    if (nrec > 1) {
      tseq <- sort(runif(nrec - 1, 0, observed.time))
      tseq <- c(tseq, observed.time)
    } else {
      tseq <- observed.time
    }
    start <- c(0, head(tseq, -1))
    stop  <- tseq
    event <- as.integer((Tm <= Ce) & (Tm <= stop))
    ## time-varying covariates (original time scale)
    xtd1 <- xi[4] + a4i * tseq
    xtd2 <- xi[5] + a5i * tseq
    dta.i <- data.frame(id    = i,
                        start = start,
                        stop  = stop,
                        event = event,
                        xtd1  = xtd1,
                        xtd2  = xtd2)
    ## replicate static covariates across records
    data.frame(dta.i,
               x = do.call(cbind, lapply(x[i, ], function(xx) rep(xx, nrec))))
  }))
  ## scale factor
  s <- if (isTRUE(scale)) {
    1 / max(dta$stop)
  } else if (isFALSE(scale)) {
    1
  } else if (is.numeric(scale) && length(scale) == 1L) {
    as.numeric(scale)
  } else {
    1
  }
  dta$start <- s * dta$start
  dta$stop  <- s * dta$stop
  ## ------------------------------------------------------
  ## true hazard (scaled time) as a subject x time matrix
  ## ------------------------------------------------------
  haz <- function(time, x) {
    ## time: vector on the scaled time axis
    time <- as.numeric(time)
    if (!length(time)) stop("hazard.3$haz: 'time' must be non-empty.")
    ## original (unscaled) time points
    t0 <- time / s
    ## subject IDs and unique subjects
    ids      <- as.character(x[, "id"])
    id_unq   <- unique(ids)
    n_subj   <- length(id_unq)
    K        <- length(t0)
    ## slopes aligned with id_unq
    a4_vec <- a4_i[id_unq]
    a5_vec <- a5_i[id_unq]
    if (anyNA(a4_vec) || anyNA(a5_vec)) stop("hazard.3$haz: slope lookup failed (id mismatch).")
    ## pull one static covariate row per subject
    Xmat <- as.matrix(x[match(unique(ids), ids), grepl("^x\\.", names(x)), drop = FALSE])
    out <- matrix(NA_real_, nrow = n_subj, ncol = K)
    rownames(out) <- id_unq
    colnames(out) <- NULL   ## keep times implicit; columns aligned with 'time'
    for (i in seq_len(n_subj)) {
      xrow <- Xmat[i, ]
      x1   <- xrow[1]
      x2   <- xrow[2]
      x4   <- xrow[4]
      x5   <- xrow[5]
      theta0 <- as.numeric(beta1 * x1 + beta2 * x2 +
                           beta3 * x4 + beta4 * x5 +
                           beta5 * x1 * x4 + beta6 * x2^2)
      lambda_i <- as.numeric(beta3 * a4_vec[i] + beta4 * a5_vec[i] + beta5 * a4_vec[i] * x1)
      ## hazard in original time
      h0 <- (1 + t0) * exp(theta0 + lambda_i * t0)
      ## transform to scaled-time hazard: h*(t*) = h(t*/s) / s
      out[i, ] <- h0 / s
    }
    out
  }
  list(dta = dta, haz = haz, scale = s)
}

##########################################################################
##
## main wrapper function
##
##########################################################################
auct.rhf <- function(object,
                     marker = c("cumhaz",  "hazard", "chf", "haz"),
                     method = c("cumulative", "incident"),
                     tau = NULL,
                     riskset = c("subject", "record"),
                     min.controls = 25,
                     nfrac.controls = 0.10,
                     min.cases = 1,
                     g.floor = 0.10,
                     g.floor.q = NULL,
                     power = 2,
                     ydata = NULL,
                     winsor.q = NULL,
                     eps = 1e-12,
                     bootstrap.rep = 0L,
                     bootstrap.refit = FALSE,
                     bootstrap.conf = 0.95,
                     bootstrap.seed = NULL,
                     verbose = TRUE) {
  marker <- match.arg(marker)
  if (marker == "chf") marker <- "cumhaz"
  if (marker == "haz") marker <- "hazard"
  method  <- match.arg(method)
  riskset <- match.arg(riskset)
  ## --- 1) Time grid and marker matrix (RHF-specific) ---
  times.full <- object$time.interest
  if (is.null(times.full) || length(times.full) < 1L)
    stop("time.interest missing")
  keep.time <- if (is.null(tau)) rep(TRUE, length(times.full)) else times.full <= tau
  if (!any(keep.time))
    stop("No time.interest <= tau.")
  times <- times.full[keep.time]
  ## Pick the requested marker; no conversions/fallbacks
  get.mat <- function(name.oob, name.test) {
    if (!is.null(object[[name.oob]])) return(object[[name.oob]])
    if (!is.null(object[[name.test]])) return(object[[name.test]])
    NULL
  }
  Z <- switch(marker,
              "cumhaz" = get.mat("chf.oob", "chf.test"),
              "hazard" = get.mat("hazard.oob", "hazard.test"))
  if (is.null(Z))
    stop("Requested marker matrix is missing in RHF object.")
  Z <- Z[, keep.time, drop = FALSE]
  ## --- 2) Outcomes and subject alignment (RHF-specific extraction) ---
  YY <- if (is.null(object$id) || is.null(object$yvar)) {
    if (is.null(ydata))
      stop("Predict object lacks id/yvar; supply counting-process data via ydata=.")
    ydata[, c("id","start","stop","event")]
  } else {
    yv <- as.data.frame(object$yvar)
    if (!is.null(object$yvar.names) && length(object$yvar.names) >= 3L) {
      colnames(yv)[1:3] <- object$yvar.names[1:3]
    }
    if (!all(colnames(yv)[1:3] %in% c("start","stop","event"))) {
      colnames(yv)[1:3] <- c("start","stop","event")
    }
    data.frame(id = object$id, yv)
  }
  YY$id <- as.character(YY$id)
  ## --- 3) Call model-agnostic core for AUC(t)/iAUC ---
  core <- .tdc_auct_core(times        = times,
                         Z            = Z,
                         YY           = YY,
                         method       = method,
                         riskset      = riskset,
                         min.controls = min.controls,
                         nfrac.controls = nfrac.controls,
                         min.cases    = min.cases,
                         g.floor      = g.floor,
                         g.floor.q    = g.floor.q,
                         power        = power,
                         winsor.q     = winsor.q,
                         eps          = eps)
  A             <- core$AUC.by.time
  iAUC.uno      <- core$iAUC.uno
  iAUC.std      <- core$iAUC.std
  diag.riskset  <- core$diag.riskset
  times         <- core$times
  g.floor       <- core$g.floor   ## possibly adjusted by g.floor.q
  aux           <- core$aux
  ## unpack aux objects needed for bootstrap branches
  Tstop      <- aux$Tstop
  Delta      <- aux$Delta
  entry      <- aux$entry
  subj.order <- aux$subj.order
  id.to.row  <- aux$id.to.row
  start.vec  <- aux$start.vec
  stop.vec   <- aux$stop.vec
  id.vec     <- aux$id.vec
  Nsubj      <- aux$Nsubj
  k.list     <- aux$k.list
  by.k       <- aux$by.k
  ## ---- Bootstrap SEs (unchanged logic, uses objects above) ----
  boot <- NULL
  if (isTRUE(bootstrap.rep > 0L)) {
    if (!is.null(bootstrap.seed)) {
      set.seed(as.integer(bootstrap.seed)[1L])
    }
    B <- as.integer(bootstrap.rep)
    base.times <- A$time
    if (isTRUE(bootstrap.refit)) {
      ## -------- Refit bootstrap (refit RHF each replicate) --------
      parms <- object$forest$parms
      if (is.null(parms)) {
        warning("bootstrap.refit=TRUE requested but object$forest$parms is missing; falling back to plug-in bootstrap.")
        bootstrap.refit <- FALSE
      } else {
        eval.df <- .rhf.build_data_from_object(object)
        is.predict.obj <- inherits(object, "predict") && !is.null(object$hazard.oob)
        train.df <- eval.df
        build.args <- function(parms, fml, dat) {
          allowed <- c("ntree","treesize","nodesize","nsplit","mtry","ntime",
                       "splitrule","experimental.bits","bootstrap","samptype",
                       "xvar.wt","sampsize","case.wt","membership","seed",
                       "do.trace","block.size")
          args <- list(formula = fml, data = dat)
          for (nm in intersect(names(parms), allowed)) {
            if (nm == "xvar.wt") {
              p <- ncol(dat) - 4L  # predictors only
              if (length(parms$xvar.wt) == p) args$xvar.wt <- parms$xvar.wt
            } else {
              args[[nm]] <- parms[[nm]]
            }
          }
          args
        }
        auc.mat    <- matrix(NA_real_, nrow = B, ncol = length(base.times))
        iauc.uno.v <- rep(NA_real_, B)
        iauc.std.v <- rep(NA_real_, B)
        fml <- as.formula("Surv(id, start, stop, event) ~ .")
        uid <- unique(train.df$id)
        if (isTRUE(verbose) && B > 0L) {
          t.start <- proc.time()[3]
          step    <- max(1L, floor(B / 10L))
          message("auct.rhf bootstrap (refit): starting ", B, " replicates...")
        }
        for (bb in seq_len(B)) {
          samp.ids <- sample(uid, length(uid), replace = TRUE)
          tr.b <- train.df[train.df$id %in% samp.ids, , drop = FALSE]
          args.fit <- build.args(parms, fml, tr.b)
          ofit <- try(do.call(rhf, args.fit), silent = TRUE)
          if (inherits(ofit, "try-error")) next
          ans.b <- try({
            if (is.predict.obj) {
              pr <- predict.rhf(ofit, eval.df)
              auct.rhf(pr, marker = marker, method = method, tau = tau,
                       riskset = riskset, min.controls = min.controls,
                       nfrac.controls = nfrac.controls, min.cases = min.cases,
                       g.floor = g.floor, g.floor.q = g.floor.q, power = power,
                       ydata = eval.df, winsor.q = winsor.q, eps = eps,
                       bootstrap.rep = 0L)
            } else {
              auct.rhf(ofit, marker = marker, method = method, tau = tau,
                       riskset = riskset, min.controls = min.controls,
                       nfrac.controls = nfrac.controls, min.cases = min.cases,
                       g.floor = g.floor, g.floor.q = g.floor.q, power = power,
                       winsor.q = winsor.q, eps = eps, bootstrap.rep = 0L)
            }
          }, silent = TRUE)
          if (inherits(ans.b, "try-error")) next
          at   <- ans.b$AUC.by.time
          aucb <- .rhf.align_auc_by_time(at$time, at$AUC, base.times)
          auc.mat[bb, ] <- aucb
          iauc.uno.v[bb] <- ans.b$iAUC.uno
          iauc.std.v[bb] <- ans.b$iAUC.std
          if (isTRUE(verbose) && B > 0L) {
            step <- max(1L, floor(B / 10L))
            if (bb == 1L || bb %% step == 0L || bb == B) {
              t.now   <- proc.time()[3]
              elapsed <- t.now - t.start
              rate    <- elapsed / bb
              remain  <- max(0, rate * (B - bb))
              message(sprintf("  [auct.rhf] refit bootstrap %d/%d  elapsed=%.1fs  approx.remaining=%.1fs", bb, B, elapsed, remain))
            }
          }
        }
        auc.se      <- apply(auc.mat, 2L, stats::sd, na.rm = TRUE)
        iauc.uno.se <- stats::sd(iauc.uno.v, na.rm = TRUE)
        iauc.std.se <- stats::sd(iauc.std.v, na.rm = TRUE)
        auc.lower <- auc.upper <- NULL
        if (is.finite(bootstrap.conf) && !is.na(bootstrap.conf) &&
            bootstrap.conf > 0 && bootstrap.conf < 1) {
          zc <- stats::qnorm((1 + bootstrap.conf) / 2)
          auc.lower <- A$AUC - zc * auc.se
          auc.upper <- A$AUC + zc * auc.se
        }
        boot <- list(
          AUC.se       = auc.se,
          iAUC.uno.se  = iauc.uno.se,
          iAUC.std.se  = iauc.std.se,
          conf.level   = bootstrap.conf,
          AUC.lower    = auc.lower,
          AUC.upper    = auc.upper,
          rep          = bootstrap.rep,
          mode         = "refit"
        )
      }
    }
    if (!isTRUE(bootstrap.refit)) {
      ## -------- Plug-in bootstrap (hold marker fixed; resample subjects) --------
      case.rows.by.k <- vector("list", length(k.list))
      ctrl.rows.by.k <- vector("list", length(k.list))
      names(case.rows.by.k) <- names(ctrl.rows.by.k) <- as.character(k.list)
      for (hh in seq_along(k.list)) {
        k  <- k.list[hh]
        tk <- times[k]
        if (method == "incident") {
          case.ids <- subj.order[ by.k[[as.character(k)]] ]
          cs <- as.integer(id.to.row[case.ids]); cs <- cs[is.finite(cs)]
          if (riskset == "subject") {
            cr0 <- which(entry < tk & Tstop >= tk)
          } else {
            at  <- which(start.vec < tk & stop.vec >= tk)
            ids <- unique(as.character(id.vec[at]))
            cr0 <- as.integer(id.to.row[ids]); cr0 <- cr0[is.finite(cr0)]
          }
          cr <- setdiff(cr0, cs)
        } else {
          cs <- which(Delta == 1 & Tstop <= tk)
          cr <- which(Tstop > tk)
        }
        case.rows.by.k[[hh]] <- cs
        ctrl.rows.by.k[[hh]] <- cr
      }
      auc.mat    <- matrix(NA_real_, nrow = B, ncol = length(k.list))
      iauc.uno.v <- rep(NA_real_, B)
      iauc.std.v <- rep(NA_real_, B)
      if (isTRUE(verbose) && B > 0L) {
        t.start <- proc.time()[3]
        step    <- max(1L, floor(B / 10L))
        message("auct.rhf bootstrap (plug-in): starting ", B, " replicates...")
      }
      for (bb in seq_len(B)) {
        w <- tabulate(sample.int(Nsubj, Nsubj, replace = TRUE), nbins = Nsubj)
        km.b  <- survival::survfit(survival::Surv(Tstop, 1 - Delta) ~ 1, weights = w)
        km.tb <- c(0, km.b$time)
        km.sb <- c(1, km.b$surv)
        G.left.b <- function(t) {
          idx <- findInterval(t, km.tb, left.open = TRUE, rightmost.closed = TRUE)
          pmax(km.sb[pmax(idx, 1L)], eps)
        }
        A.vec <- rep(NA_real_, length(k.list))
        W.vec <- rep(NA_real_, length(k.list))
        for (hh in seq_along(k.list)) {
          k  <- k.list[hh]
          tk <- times[k]
          cs <- case.rows.by.k[[hh]]
          cr <- ctrl.rows.by.k[[hh]]
          n.case.eff <- sum(w[cs])
          n.ctrl.eff <- sum(w[cr])
          n.ctrl.min <- max(min.controls, ceiling(nfrac.controls * Nsubj))
          if (n.case.eff < min.cases || n.ctrl.eff < n.ctrl.min) {
            A.vec[hh] <- NA_real_; W.vec[hh] <- NA_real_
            next
          }
          sc.case <- rep(Z[cs, k], times = w[cs])
          sc.ctrl <- rep(Z[cr, k], times = w[cr])
          sc.ctrl.sorted <- sort(sc.ctrl)
          less.idx <- findInterval(sc.case, sc.ctrl.sorted,
                                   left.open = TRUE,  rightmost.closed = TRUE)
          leq.idx  <- findInterval(sc.case, sc.ctrl.sorted,
                                   left.open = FALSE, rightmost.closed = TRUE)
          n.less   <- less.idx
          n.equal  <- pmax(leq.idx - less.idx, 0L)
          A.k      <- mean((n.less + 0.5 * n.equal) / length(sc.ctrl))
          Gk.raw <- G.left.b(tk)
          Gk     <- max(Gk.raw %||% 0, g.floor)
          Wk     <- n.case.eff / (Gk^power + eps)
          A.vec[hh] <- A.k
          W.vec[hh] <- Wk
        }
        auc.mat[bb, ] <- A.vec
        ok.b <- is.finite(A.vec) & is.finite(W.vec) & W.vec > 0
        iauc.uno.v[bb] <- if (any(ok.b)) {
          sum(A.vec[ok.b] * W.vec[ok.b]) / sum(W.vec[ok.b])
        } else {
          NA_real_
        }
        tt <- times[k.list]
        ok.t <- is.finite(A.vec) & is.finite(tt)
        if (sum(ok.t) >= 2L) {
          num <- sum((head(A.vec[ok.t], -1L) + tail(A.vec[ok.t], -1L)) *
                     diff(tt[ok.t]) / 2)
          den <- max(tt[ok.t]) - min(tt[ok.t])
          iauc.std.v[bb] <- if (is.finite(den) && den > 0) num / den else NA_real_
        } else {
          iauc.std.v[bb] <- NA_real_
        }
        if (isTRUE(verbose) && B > 0L) {
          step <- max(1L, floor(B / 10L))
          if (bb == 1L || bb %% step == 0L || bb == B) {
            t.now   <- proc.time()[3]
            elapsed <- t.now - t.start
            rate    <- elapsed / bb
            remain  <- max(0, rate * (B - bb))
            message(sprintf("  [auct.rhf] plug-in bootstrap %d/%d  elapsed=%.1fs  approx.remaining=%.1fs", bb, B, elapsed, remain))
          }
        }
      }
      auc.se <- apply(auc.mat, 2L, stats::sd, na.rm = TRUE)
      se.full <- rep(NA_real_, nrow(A))
      kk <- match(times[k.list], A$time)
      se.full[kk] <- auc.se
      iauc.uno.se <- stats::sd(iauc.uno.v, na.rm = TRUE)
      iauc.std.se <- stats::sd(iauc.std.v, na.rm = TRUE)
      auc.lower <- auc.upper <- NULL
      if (is.finite(bootstrap.conf) && !is.na(bootstrap.conf) &&
          bootstrap.conf > 0 && bootstrap.conf < 1) {
        zc <- stats::qnorm((1 + bootstrap.conf) / 2)
        auc.lower <- A$AUC - zc * se.full
        auc.upper <- A$AUC + zc * se.full
      }
      boot <- list(
        AUC.se       = se.full,
        iAUC.uno.se  = iauc.uno.se,
        iAUC.std.se  = iauc.std.se,
        conf.level   = bootstrap.conf,
        AUC.lower    = auc.lower,
        AUC.upper    = auc.upper,
        rep          = bootstrap.rep,
        mode         = "plug-in"
      )
    }
  }
  structure(list(
    AUC.by.time  = A,
    iAUC.uno     = iAUC.uno,
    iAUC.std     = iAUC.std,
    boot         = boot,
    marker       = marker,
    method       = method,
    riskset      = riskset,
    power        = power,
    g.floor      = g.floor,
    g.floor.q    = g.floor.q,
    winsor.q     = winsor.q,
    times        = times,
    diag.riskset = diag.riskset
  ), class = "auct.rhf")
}
auct <- auct.rhf
##########################################################################
##
## Model-agnostic core: AUC(t) + iAUC for counting-process data
##
##########################################################################
.tdc_auct_core <- function(times,
                           Z,
                           YY,
                           method        = c("cumulative", "incident"),
                           riskset       = c("subject", "record"),
                           min.controls  = 25,
                           nfrac.controls = 0.10,
                           min.cases     = 1,
                           g.floor       = 0.10,
                           g.floor.q     = NULL,
                           power         = 2,
                           winsor.q      = NULL,
                           eps           = 1e-12) {
  method  <- match.arg(method)
  riskset <- match.arg(riskset)
  ## basic checks
  if (is.null(times) || !length(times))
    stop("AUC core: 'times' must be a non-empty numeric vector.")
  if (!is.numeric(times))
    stop("AUC core: 'times' must be numeric.")
  if (!is.matrix(Z))
    stop("AUC core: 'Z' must be a matrix.")
  if (ncol(Z) != length(times))
    stop("AUC core: ncol(Z) must equal length(times).")
  needed <- c("id", "start", "stop", "event")
  if (!all(needed %in% names(YY)))
    stop("AUC core: 'YY' must contain columns: id, start, stop, event.")
  ## canonical y-data
  YY <- YY[, needed]
  YY$id <- as.character(YY$id)
  ## --- last record per subject (event/censor time) ---
  last <- YY[!duplicated(YY$id, fromLast = TRUE), , drop = FALSE]
  ## subject order: prefer rownames(Z) if present
  rn <- rownames(Z)
  if (!is.null(rn) && length(rn) == nrow(Z)) {
    subj.order <- as.character(rn)
  } else {
    subj.order <- as.character(YY$id[!duplicated(YY$id)])
  }
  last <- last[match(subj.order, last$id), , drop = FALSE]
  if (anyNA(last$id))
    stop("Cannot align subjects between predictions and outcomes.")
  Tstop <- last$stop
  Delta <- last$event
  ## subject entry time (for riskset = 'subject')
  entry <- tapply(YY$start, YY$id, min)
  entry <- entry[match(subj.order, names(entry))]
  Nsubj     <- nrow(Z)
  nmin.ctrl <- max(min.controls, ceiling(nfrac.controls * Nsubj))
  id.to.row <- stats::setNames(seq_len(Nsubj), subj.order)
  ## --- event-to-grid mapping for incident evaluation ---
  maxT  <- max(times, na.rm = TRUE)
  ev.idx <- which(Delta == 1 & Tstop <= maxT)
  if (!length(ev.idx))
    stop("No events within the evaluation grid.")
  ## nearest-right grid index k for each event time
  k.of.ev <- vapply(Tstop[ev.idx],
                    function(s) which(times >= s)[1L],
                    integer(1L))
  by.k   <- split(ev.idx, k.of.ev)
  k.list <- sort(as.integer(names(by.k)))
  k.list <- k.list[!is.na(k.list)]
  ## --- OPTIONAL aggregation for incident AUC: time windows with >= min.cases events ---
  ## This only activates for method = "incident" and min.cases > 1.
  if (identical(method, "incident") && is.finite(min.cases) && min.cases > 1L) {
    k.vec <- k.list
    if (length(k.vec)) {
      ## number of events at each k
      ev.per.k <- vapply(k.vec,
                         function(kk) length(by.k[[as.character(kk)]]),
                         integer(1L))
      total.ev <- sum(ev.per.k)
      if (total.ev > 0L) {
        blocks    <- list()
        start.pos <- 1L
        while (start.pos <= length(k.vec)) {
          cum.ev  <- 0L
          end.pos <- start.pos
          while (end.pos <= length(k.vec) && cum.ev < min.cases) {
            cum.ev  <- cum.ev + ev.per.k[end.pos]
            end.pos <- end.pos + 1L
          }
          if (cum.ev == 0L)
            break
          ## indices in k.vec that belong to this block
          idx.seq <- start.pos:(end.pos - 1L)
          ## If this is the last chunk and still < min.cases, merge into previous block
          if (cum.ev < min.cases && length(blocks) > 0L && end.pos > length(k.vec)) {
            blocks[[length(blocks)]]$k_idx <-
              c(blocks[[length(blocks)]]$k_idx, k.vec[idx.seq])
            break
          } else {
            block.k <- k.vec[idx.seq]
            blocks[[length(blocks) + 1L]] <- list(
              eval_k = block.k[1L],  ## evaluate at the earliest time in the block
              k_idx  = block.k       ## but include events from all these k's
            )
            start.pos <- end.pos
          }
        }
        if (length(blocks)) {
          ## rebuild k.list and by.k based on blocks
          new.k.list <- vapply(blocks, function(b) b$eval_k, integer(1L))
          new.by.k   <- vector("list", length(blocks))
          names(new.by.k) <- as.character(new.k.list)
          for (bb in seq_along(blocks)) {
            kk.seq <- blocks[[bb]]$k_idx
            idxs   <- unlist(by.k[as.character(kk.seq)], use.names = FALSE)
            ## just in case, drop duplicates
            new.by.k[[bb]] <- unique(idxs)
          }
          k.list <- new.k.list
          by.k   <- new.by.k
        }
      }
    }
  }
  ## --- KM IPCW (left-continuous) and optional floor augmentation ---
  km.fit <- survival::survfit(survival::Surv(Tstop, 1 - Delta) ~ 1)
  km.t   <- c(0, km.fit$time)
  km.s   <- c(1, km.fit$surv)
  G.left <- function(t) {
    idx <- findInterval(t, km.t, left.open = TRUE, rightmost.closed = TRUE)
    pmax(km.s[pmax(idx, 1L)], eps)
  }
  if (!is.null(g.floor.q)) {
    G.grid  <- vapply(times, G.left, numeric(1L))
    g.floor <- max(g.floor,
                   stats::quantile(G.grid, probs = g.floor.q,
                                   na.rm = TRUE))
  }
  ## --- loop over evaluation times and compute AUC(t) + weights ---
  res <- vector("list", length(k.list))
  names(res) <- k.list
  ## vectors used for record riskset
  start.vec <- YY$start
  stop.vec  <- YY$stop
  id.vec    <- YY$id
  for (hh in seq_along(k.list)) {
    k  <- k.list[hh]
    tk <- times[k]
    if (method == "incident") {
      ## Cases: fail at tk (mapped event-time index k.list)
      case.ids  <- subj.order[ by.k[[as.character(k)]] ]
      case.rows <- as.integer(id.to.row[case.ids])
      case.rows <- case.rows[is.finite(case.rows)]
      ## Controls: at risk at tk
      if (riskset == "subject") {
        ctrl.rows <- which(entry < tk & Tstop >= tk)
        ctrl.rows <- setdiff(ctrl.rows, case.rows)
      } else {
        at.risk.rows <- which(start.vec < tk & stop.vec >= tk)
        ctrl.ids     <- setdiff(unique(as.character(id.vec[at.risk.rows])),
                                case.ids)
        ctrl.rows    <- as.integer(id.to.row[ctrl.ids])
        ctrl.rows    <- ctrl.rows[is.finite(ctrl.rows)]
      }
    } else {  ## cumulative/dynamic
      ## Cases: failed by tk (T.i <= tk, Delta.i = 1)
      case.rows <- which(Delta == 1 & Tstop <= tk)
      ## Controls: event-free at tk (T.i > tk)
      ctrl.rows <- which(Tstop > tk)
      ## (Riskset ignored here by design.)
    }
    n.cases    <- length(case.rows)
    n.ctrl.raw <- length(ctrl.rows)
    if (n.cases < min.cases || n.ctrl.raw < nmin.ctrl) {
      res[[hh]] <- data.frame(time    = tk,
                              AUC     = NA_real_,
                              n.cases = n.cases,
                              n.ctrl  = n.ctrl.raw,
                              G       = NA_real_,
                              W       = NA_real_)
      next
    }
    sc.case <- Z[case.rows, k]
    sc.case <- sc.case[is.finite(sc.case)]
    sc.ctrl <- Z[ctrl.rows, k]
    sc.ctrl <- sc.ctrl[is.finite(sc.ctrl)]
    n.ctrl  <- length(sc.ctrl)
    if (!length(sc.case) || n.ctrl < nmin.ctrl) {
      res[[hh]] <- data.frame(time    = tk,
                              AUC     = NA_real_,
                              n.cases = length(sc.case),
                              n.ctrl  = n.ctrl,
                              G       = NA_real_,
                              W       = NA_real_)
      next
    }
    ## Unweighted within-time AUC with 0.5 for ties
    sc.ctrl.sorted <- sort(sc.ctrl)
    less.idx <- findInterval(sc.case, sc.ctrl.sorted,
                             left.open = TRUE,  rightmost.closed = TRUE)
    leq.idx  <- findInterval(sc.case, sc.ctrl.sorted,
                             left.open = FALSE, rightmost.closed = TRUE)
    n.less   <- less.idx
    n.equal  <- pmax(leq.idx - less.idx, 0L)
    AUC.k    <- mean((n.less + 0.5 * n.equal) / n.ctrl)
    ## Uno-style time weight using KM IPCW (stabilized)
    Gk.raw <- G.left(tk)
    Gk     <- max(Gk.raw %||% 0, g.floor)
    Wk     <- length(sc.case) / (Gk^power + eps)
    res[[hh]] <- data.frame(time    = tk,
                            AUC     = AUC.k,
                            n.cases = length(sc.case),
                            n.ctrl  = n.ctrl,
                            G       = Gk.raw,
                            W       = Wk)
  }
  A <- do.call(rbind, res)
  ## optional winsorization of time weights
  if (!is.null(winsor.q) && is.numeric(winsor.q) &&
      length(winsor.q) == 1L &&
      winsor.q > 0 && winsor.q < 1 &&
      "W" %in% names(A)) {
    cap <- stats::quantile(A$W, probs = winsor.q,
                           na.rm = TRUE, names = FALSE)
    A$W <- pmin(A$W, cap)
  }
  ## iAUC (Uno-style time weighting)
  ok <- is.finite(A$AUC) & is.finite(A$W) & A$W > 0
  iAUC.uno <- if (any(ok)) {
    sum(A$AUC[ok] * A$W[ok]) / sum(A$W[ok])
  } else {
    NA_real_
  }
  ## time-standardized mean of AUC(t)
  iAUC.std <- .rhf.trapz.std(A$time, A$AUC)
  ## ---------- risk-set diagnostic: subject vs record ----------
  diag.riskset <- NULL
  {
    J <- min(5L, max(1L, floor(length(times) / 10)))
    idx.sel <- unique(pmax(1L,
                           pmin(length(times),
                                round(stats::quantile(seq_along(times),
                                                      probs = seq(0.2, 0.8,
                                                                  length.out = J),
                                                      names = FALSE)))))
    t.sel <- times[idx.sel]
    subj.ctrl.ct <- integer(length(t.sel))
    rec.ctrl.ct  <- integer(length(t.sel))
    n.diff       <- integer(length(t.sel))
    for (j in seq_along(t.sel)) {
      t0 <- t.sel[j]
      ## subject-level controls
      ctrl.subj <- which(entry < t0 & Tstop >= t0)
      ## record-level controls
      at       <- which(start.vec < t0 & stop.vec >= t0)
      ctrl.ids <- unique(as.character(id.vec[at]))
      ctrl.rec <- as.integer(id.to.row[ctrl.ids])
      ctrl.rec <- ctrl.rec[is.finite(ctrl.rec)]
      subj.ctrl.ct[j] <- length(ctrl.subj)
      rec.ctrl.ct[j]  <- length(unique(ctrl.rec))
      n.diff[j]       <- length(setdiff(ctrl.subj, ctrl.rec)) +
                         length(setdiff(ctrl.rec,  ctrl.subj))
    }
    prop.diff <- if (length(t.sel)) mean(n.diff > 0) else NA_real_
    diag.riskset <- list(
      times              = t.sel,
      n.ctrl.subject     = subj.ctrl.ct,
      n.ctrl.record      = rec.ctrl.ct,
      n.diff             = n.diff,
      prop.times.different = prop.diff
    )
  }
  structure(list(
    AUC.by.time = A,
    iAUC.uno    = iAUC.uno,
    iAUC.std    = iAUC.std,
    diag.riskset = diag.riskset,
    times       = times,
    g.floor     = g.floor,
    aux = list(
      Tstop      = Tstop,
      Delta      = Delta,
      entry      = entry,
      subj.order = subj.order,
      id.to.row  = id.to.row,
      start.vec  = start.vec,
      stop.vec   = stop.vec,
      id.vec     = id.vec,
      Nsubj      = Nsubj,
      k.list     = k.list,
      by.k       = by.k
    )
  ), class = "auct.rhf")
}
##########################################################################
##
## Plot method
##
##########################################################################
plot.auct.rhf <- function(x,
                          bass = 10,
                          xlab = "Time",
                          ylab = NULL,
                          main = NULL,
                          ylim = NULL,
                          pch = 16,
                          alpha = .05,
                          ...) {
  A <- x$AUC.by.time
  ylab <- ylab %||% "AUC(t) [KM-IPCW]"
  ttl  <- paste0("Time-varying AUC(t) - ", x$method, ", marker=", x$marker)
  main <- main %||% ttl
  if (is.null(ylim)) {
    y <- A$AUC
    ymin <- suppressWarnings(min(y[is.finite(y)], na.rm = TRUE))
    ymax <- suppressWarnings(max(y[is.finite(y)], na.rm = TRUE))
    ylim <- if (is.finite(ymax) & is.finite(ymin)) c(ymin * .95, ymax * 1.05) else c(0, 1)
  }
  keep <- !is.na(A$AUC)
  plot(A$time[keep], A$AUC[keep], pch = pch, xlab = xlab, ylab = ylab, main = main, ylim = ylim, ...)
  ok <- is.finite(A$time) & is.finite(A$AUC)
  if (sum(ok) >= 5L) {
    sm <- stats::supsmu(A$time[ok], A$AUC[ok], bass = bass)
    lines(sm, lwd = 2)
  }
  ## optional bootstrap SE ribbon (base R polygon)
  if (!is.null(x$boot) && !is.null(x$boot$AUC.se)) {
    yhat <- A$AUC
    se   <- x$boot$AUC.se
    if (!is.null(x$boot$AUC.lower) && !is.null(x$boot$AUC.upper)) {
      yl <- x$boot$AUC.lower; yu <- x$boot$AUC.upper
    } else {
      zc <- stats::qnorm(1 - min(alpha/2, .5))
      yl <- yhat - zc * se
      yu <- yhat + zc * se
    }
    okb <- is.finite(A$time) & is.finite(yl) & is.finite(yu)
    if (sum(okb) >= 2L) {
      xt <- A$time[okb]
      polygon(c(xt, rev(xt)), c(yl[okb], rev(yu[okb])),
              border = NA, col = grDevices::adjustcolor("gray", alpha.f = 0.25))
      points(A$time[keep], A$AUC[keep], pch = pch)
    }
  }
}
##########################################################################
##
## Print method (right-justified; aligned colon incl. quantiles) 
##
##########################################################################
print.auct.rhf <- function(x, digits = 4, max.rows = 8, ...) {
  stopifnot(inherits(x, "auct.rhf"))
  A    <- x$AUC.by.time
  K    <- nrow(A)
  K.ok <- sum(is.finite(A$AUC))
  tr   <- range(x$times, na.rm = TRUE)
  fnum <- function(v) {
    if (is.numeric(v) && length(v) == 1L && is.finite(v)) {
      formatC(v, digits = digits, format = "f")
    } else {
      "NA"
    }
  }
  method <- if (!is.null(x$method)) x$method else x$definition
  riskln <- if (identical(method, "cumulative")) "ignored" else x$riskset
  main.labels <- c(
    "Method",
    "Marker",
    "Riskset",
    "Grid points (evaluated)",
    "Time range",
    "iAUC.uno",
    "iAUC.std (mean over time)",
    "IPCW"
  )
  main.values <- c(
    method,
    x$marker,
    riskln,
    sprintf("%d (%d with finite AUC)", K, K.ok),
    sprintf("[%.6g, %.6g]", tr[1], tr[2]),
    fnum(x$iAUC.uno),
    fnum(x$iAUC.std),
    paste0("KM  floor=", x$g.floor,
           if (!is.null(x$g.floor.q)) paste0(" (q=", x$g.floor.q, ")") else "",
           ", power=", x$power,
           if (!is.null(x$winsor.q)) paste0(", winsor.q=", x$winsor.q) else "")
  )
  q.label <- "AUC(t) quantiles (10%, 50%, 90%)"
  all.labels <- if (K.ok > 0) c(main.labels, q.label) else main.labels
  w <- max(nchar(all.labels), 0L)
  out <- "RHF time-varying AUC"
  for (i in seq_along(main.labels)) {
    out <- c(out, sprintf("  %*s : %s", w, main.labels[i], main.values[i]))
  }
  if (K.ok > 0) {
    yy <- A$AUC[is.finite(A$AUC)]
    qs <- stats::quantile(yy, probs = c(0.1, 0.5, 0.9), na.rm = TRUE, names = FALSE)
    out <- c(out,
             sprintf("  %*s : %s",
                     w, q.label, paste(formatC(qs, digits = digits, format = "f"), collapse = ", ")))
  }
  ## optional diagnostic summary
  if (!is.null(x$diag.riskset)) {
    dr <- x$diag.riskset
    if (is.list(dr) && length(dr$times)) {
      out <- c(
        out,
        sprintf(
          "  Risk-set diagnostic: subject vs record differ at %d/%d times (%.2f), median |Delta-ctrl| = %d",
          sum(dr$n.diff > 0), length(dr$times),
          if (is.finite(dr$prop.times.different)) dr$prop.times.different else NA_real_,
          stats::median(dr$n.diff, na.rm = TRUE)
        )
      )
    }
  }
  message(paste(out, collapse = "\n"))
  ## optional small table preview: finite-AUC rows only (safe printing)
  if (K.ok > 0 && max.rows > 0) {
    cols <- intersect(c("time","AUC","n.cases","n.ctrl","G","W"), colnames(A))
    A.show <- A[is.finite(A$AUC), cols, drop = FALSE]
    A.show <- head(A.show, max.rows)
    message("\nAUC(t) by time (head):")
    if (nrow(A.show) > 0) {
      print.data.frame(
        as.data.frame(lapply(A.show, function(col)
          if (is.numeric(col)) signif(col, digits) else col)),
        row.names = FALSE
      )
      message(" ..... ")
    } else {
      message("  (no finite AUC rows to display)")
    }
  }
  ## optional bootstrap information
  if (!is.null(x$boot)) {
    mode.txt <- if (!is.null(x$boot$mode)) x$boot$mode else "pointwise"
    se.txt <- paste0("iAUC.uno.se = ", fnum(x$boot$iAUC.uno.se),
                     ", iAUC.std.se = ", fnum(x$boot$iAUC.std.se),
                     " (B = ", as.integer(x$boot$rep), ")")
    message(sprintf("  %*s : %s", w, paste0("Bootstrap (", mode.txt, ")"), se.txt))
  }
  invisible(x)
}
##########################################################################
##
## helpers
##
##########################################################################
`%||%` <- function(a, b) if (!is.null(a)) a else b
.rhf.trapz <- function(t, y) {
  o <- order(t); t <- t[o]; y <- y[o]
  ok <- is.finite(t) & is.finite(y)
  t <- t[ok]; y <- y[ok]
  if (length(t) < 2L) return(NA_real_)
  sum((head(y, -1L) + tail(y, -1L)) * diff(t) / 2)
}
.rhf.trapz.std <- function(t, y) {
  num <- .rhf.trapz(t, y)
  if (!is.finite(num)) return(NA_real_)
  den <- max(t, na.rm = TRUE) - min(t, na.rm = TRUE)
  if (!is.finite(den) || den <= 0) return(NA_real_)
  num / den
}
.rhf.unique.subject.order <- function(id) {
  id.char <- as.character(id)
  id.char[!duplicated(id.char)]
}
.rhf.slice_by_subject <- function(Z, subj.order, keep.ids) {
  if (!is.matrix(Z)) stop("Z must be a matrix.")
  idx <- match(keep.ids, subj.order)
  Z[idx, , drop = FALSE]
}
.rhf.bootstrap_once <- function(object, Z, YY, times, subj.order,
                                marker, method, riskset,
                                min.cases, min.controls, nfrac.controls,
                                g.floor, g.floor.q, power, winsor.q, eps) {
  o.tmp <- list(
    time.interest = times,
    id  = YY$id,
    yvar = as.matrix(YY[, c("start","stop","event")])
  )
  if (marker == "cumhaz") {
    o.tmp$chf.oob  <- Z
  } else {
    o.tmp$hazard.oob <- Z
  }
  class(o.tmp) <- c("rhf", "grow")
  auct.rhf(object = o.tmp, marker = marker, method = method, tau = max(times),
           riskset = riskset, min.controls = min.controls, nfrac.controls = nfrac.controls,
           min.cases = min.cases, g.floor = g.floor, g.floor.q = g.floor.q,
           power = power, ydata = YY, winsor.q = winsor.q, eps = eps,
           bootstrap.rep = 0L)
}
.rhf.build_data_from_object <- function(object) {
  if (is.null(object$id) || is.null(object$yvar) || is.null(object$xvar)) {
    stop("object must carry id, yvar, and xvar for bootstrap.refit.")
  }
  idv <- as.character(object$id)
  yv  <- as.data.frame(object$yvar)
  xv  <- as.data.frame(object$xvar)
  ## apply provided names if available
  if (!is.null(object$yvar.names) && length(object$yvar.names) >= 3L) {
    colnames(yv)[1:3] <- object$yvar.names[1:3]
  }
  ## standardize y names to start/stop/event
  colnames(yv)[1:3] <- c("start","stop","event")
  if (!is.null(object$xvar.names) && length(object$xvar.names) == ncol(xv)) {
    colnames(xv) <- object$xvar.names
  }
  data.frame(id = idv, yv[, c("start","stop","event")], xv, check.names = FALSE)
}
.rhf.align_auc_by_time <- function(src.time, src.auc, base.time) {
  ## align src AUC(t) to base.time via left-continuous step function
  idx <- findInterval(base.time, src.time, left.open = TRUE, rightmost.closed = TRUE)
  out <- rep(NA_real_, length(base.time))
  ok  <- idx >= 1 & idx <= length(src.time)
  out[ok] <- src.auc[idx[ok]]
  out
}

predict.rhf <- function(object,
                        newdata,
                        get.tree = NULL,
                        block.size = 10,
                        membership = TRUE,
                        seed = NULL,
                        do.trace = FALSE,
                        ...)
{
    predict.rhf.workhorse(object,
                          newdata,
                          get.tree = get.tree,
                          block.size = block.size,
                          membership = membership,
                          seed = seed,
                          do.trace = do.trace, ...)
}

########################################################################
## plot helpers
########################################################################
.rhf_select_dotmatrix_variables <- function(mat,
                                            vars = NULL,
                                            top_n_union = 15L,
                                            sort_abs = TRUE) {
  if (!is.null(vars)) {
    vars <- intersect(as.character(vars), rownames(mat))
    if (!length(vars)) {
      stop("No requested variables found in the importance matrix.")
    }
    return(mat[vars, , drop = FALSE])
  }
  top_n_union <- as.integer(top_n_union)[1L]
  if (!is.finite(top_n_union) || top_n_union < 1L) {
    stop("top_n_union must be a positive integer.")
  }
  top.vars <- unique(unlist(lapply(seq_len(ncol(mat)), function(j) {
    v <- mat[, j]
    ord <- if (isTRUE(sort_abs)) {
      order(abs(v), decreasing = TRUE, na.last = TRUE)
    }
    else {
      order(v, decreasing = TRUE, na.last = TRUE)
    }
    rownames(mat)[head(ord, min(top_n_union, length(ord)))]
  })))
  if (!length(top.vars)) {
    stop("No variables available for the dot-matrix plot.")
  }
  mat[top.vars, , drop = FALSE]
}
.rhf_order_dotmatrix_variables <- function(mat,
                                           variable.labels = NULL,
                                           sort_by = c("q90", "sum", "max", "mean", "median", "alphabetical", "cluster", "none"),
                                           sort_abs = TRUE) {
  sort_by <- match.arg(sort_by)
  if (sort_by == "none") {
    return(rownames(mat))
  }
  if (sort_by %in% c("q90", "sum", "max", "mean", "median")) {
    score <- .rhf_row_summary(mat,
                              rank.by = sort_by,
                              abs = sort_abs)
    return(names(sort(score, decreasing = TRUE)))
  }
  if (sort_by == "alphabetical") {
    lab <- .rhf_label_lookup(rownames(mat), variable.labels, infer_prefix = TRUE)
    lab <- .rhf_make_unique_labels(lab, rownames(mat))
    return(rownames(mat)[order(lab)])
  }
  ## cluster
  mat.z <- t(scale(t(mat)))
  mat.z[!is.finite(mat.z)] <- 0
  hc <- stats::hclust(stats::dist(mat.z), method = "ward.D2")
  rownames(mat.z)[hc$order]
}
.rhf_rescale_from_range <- function(x,
                                    from,
                                    to = c(0.5, 3.0)) {
  x <- as.numeric(x)
  from <- as.numeric(from)
  out <- rep(0, length(x))
  ok <- is.finite(x)
  if (!any(ok)) {
    return(out)
  }
  if (length(from) != 2L || any(!is.finite(from))) {
    return(out)
  }
  if (from[1L] == from[2L]) {
    out[ok] <- mean(to)
    return(out)
  }
  out[ok] <- to[1L] + (x[ok] - from[1L]) / (from[2L] - from[1L]) * (to[2L] - to[1L])
  out
}
.rhf_pretty_breaks <- function(x,
                               n = 5L,
                               positive.only = FALSE,
                               symmetric = FALSE) {
  x <- as.numeric(x)
  x <- x[is.finite(x)]
  if (positive.only) {
    x <- x[x > 0]
  }
  if (!length(x)) {
    return(numeric(0))
  }
  n <- max(2L, as.integer(n)[1L])
  if (symmetric) {
    lim <- max(abs(x), na.rm = TRUE)
    if (!is.finite(lim) || lim <= 0) {
      return(0)
    }
    br <- pretty(c(-lim, lim), n = n)
    br <- br[br >= -lim & br <= lim]
    br <- unique(br[is.finite(br)])
    if (!0 %in% br) {
      br <- sort(unique(c(br, 0)))
    }
    return(br)
  }
  rng <- range(x, na.rm = TRUE)
  if (!is.finite(rng[1L]) || !is.finite(rng[2L])) {
    return(numeric(0))
  }
  if (rng[1L] == rng[2L]) {
    return(rng[1L])
  }
  br <- pretty(rng, n = n)
  br <- br[br >= rng[1L] & br <= rng[2L]]
  if (positive.only) {
    br <- br[br > 0]
  }
  br <- unique(br[is.finite(br)])
  if (!length(br)) {
    probs <- seq(0, 1, length.out = n)
    br <- as.numeric(stats::quantile(x, probs = probs, na.rm = TRUE, names = FALSE))
    if (positive.only) {
      br <- br[br > 0]
    }
    br <- unique(br[is.finite(br)])
  }
  br
}
.rhf_format_legend_values <- function(x,
                                      digits = 3L) {
  format(signif(as.numeric(x), digits = digits), trim = TRUE, scientific = FALSE)
}
.rhf_draw_dotmatrix_xlabels <- function(at,
                                        labels,
                                        cex = 0.9,
                                        gap.lines = 0.10,
                                        srt = 45) {
  if (!length(at) || !length(labels)) {
    return(invisible(TRUE))
  }
  usr <- graphics::par("usr")
  csi <- graphics::par("csi")
  din <- graphics::par("din")
  if (!is.finite(csi) || csi <= 0) {
    csi <- 0.2
  }
  if (!is.finite(din[2L]) || din[2L] <= 0) {
    din[2L] <- 7
  }
  ## Position labels using a fixed physical gap beneath the x-axis rather than
  ## a gap that scales with the number of variables in the plot.
  y.axis.ndc <- graphics::grconvertY(usr[3L], from = "user", to = "ndc")
  gap.in <- max(0, gap.lines) * csi * cex
  y.text.ndc <- y.axis.ndc - gap.in / din[2L]
  y.user <- graphics::grconvertY(y.text.ndc, from = "ndc", to = "user")
  graphics::text(at,
                 rep(y.user, length(at)),
                 labels = labels,
                 srt = srt,
                 adj = c(1, 1),
                 xpd = NA,
                 cex = cex)
  invisible(TRUE)
}
.rhf_dotmatrix_default_mar <- function(var.labels,
                                       x.labels,
                                       var.cex = 0.9,
                                       axis.cex = 0.9,
                                       time.label.srt = 45,
                                       legend = TRUE) {
  csi <- graphics::par("csi")
  if (!is.finite(csi) || csi <= 0) {
    csi <- 0.2
  }
  left.in <- if (length(var.labels)) {
    max(graphics::strwidth(var.labels, units = "inches", cex = var.cex), na.rm = TRUE)
  } else {
    0
  }
  lab.w.in <- if (length(x.labels)) {
    max(graphics::strwidth(x.labels, units = "inches", cex = axis.cex), na.rm = TRUE)
  } else {
    0
  }
  lab.h.in <- if (length(x.labels)) {
    max(graphics::strheight(x.labels, units = "inches", cex = axis.cex), na.rm = TRUE)
  } else {
    0
  }
  theta <- abs(as.numeric(time.label.srt)[1L]) * pi / 180
  rot.ext.in <- lab.w.in * sin(theta) + lab.h.in * cos(theta)
  ## Give the variable labels some extra room beyond their measured width so
  ## long names do not crowd the plotting region.
  left.mar <- 1.25 + left.in / csi + 1.15
  left.mar <- min(30, max(10.0, left.mar))
  ## Keep enough room for rotated time labels so they are not clipped, while
  ## controlling label proximity to the axis separately at draw time.
  bottom.mar <- 0.45 + rot.ext.in / csi + 0.65
  bottom.mar <- min(6.5, max(2.8, bottom.mar))
  top.mar <- 2.8
  right.mar <- if (isTRUE(legend)) 0.5 else 0.4
  c(bottom.mar, left.mar, top.mar, right.mar)
}
.rhf_draw_dotmatrix_legend <- function(size.breaks,
                                       size.range,
                                       size.title,
                                       cex.range,
                                       alpha,
                                       color_by,
                                       color.range = NULL,
                                       color.title = NULL,
                                       point_color = "steelblue4",
                                       value_colors = c("grey85", "steelblue4"),
                                       sign_colors = c("firebrick3", "grey90", "steelblue4"),
                                       cex.text = 0.85,
                                       cex.title = 0.9) {
  graphics::plot.new()
  graphics::plot.window(xlim = c(0, 0.70), ylim = c(0, 1), xaxs = "i", yaxs = "i")
  has.color <- color_by %in% c("value", "sign") &&
    length(color.range) == 2L && all(is.finite(color.range))
  size.breaks <- as.numeric(size.breaks)
  size.breaks <- sort(unique(size.breaks[is.finite(size.breaks)]), decreasing = TRUE)
  ## Compact legend geometry with titles placed close to the legend content.
  x.title <- 0.06
  x.dot <- 0.22
  x.size.lab <- 0.40
  if (length(size.breaks)) {
    size.y.top <- if (has.color) 0.79 else 0.74
    size.y.bottom <- if (has.color) 0.64 else 0.28
    size.y.seq <- if (length(size.breaks) == 1L) {
      mean(c(size.y.top, size.y.bottom))
    } else {
      seq(size.y.top, size.y.bottom, length.out = length(size.breaks))
    }
    ## adjust top panel
    size.y.seq <-  1.1 * size.y.seq
    size.y.top <- 1.1 * (size.y.top + 0.040)
    graphics::text(x.title,
                   size.y.top,
                   labels = size.title,
                   adj = c(0, 0.5),
                   font = 2,
                   cex = cex.title)
    cex.val <- .rhf_rescale_from_range(size.breaks,
                                       from = size.range,
                                       to = cex.range)
    graphics::points(rep(x.dot, length(size.breaks)),
                     size.y.seq,
                     pch = 16,
                     cex = cex.val,
                     col = grDevices::adjustcolor(point_color, alpha.f = alpha))
    graphics::text(rep(x.size.lab, length(size.breaks)),
                   size.y.seq,
                   labels = .rhf_format_legend_values(size.breaks),
                   adj = c(0, 0.5),
                   cex = cex.text)
  }
  else {
    graphics::text(x.title,
                   0.72,
                   labels = size.title,
                   adj = c(0, 0.5),
                   font = 2,
                   cex = cex.title)
    graphics::text(x.title,
                   0.64,
                   labels = "No positive values",
                   adj = c(0, 0.5),
                   cex = cex.text)
  }
  if (has.color) {
    if (is.null(color.title) || !nzchar(color.title)) {
      color.title <- "Importance"
    }
    x0 <- 0.18
    x1 <- 0.34
    y0 <- 0.16 + .275
    y1 <- 0.39 + .275
    #graphics::text(x.title,
    #               y1 + 0.035,
    #               labels = color.title,
    #               adj = c(0, 0.5),
    #               font = 2,
    #               cex = cex.title)
    n.bar <- 64L
    y.seq <- seq(y0, y1, length.out = n.bar + 1L)
    pal <- if (color_by == "value") {
      grDevices::colorRampPalette(value_colors)(n.bar)
    } else {
      grDevices::colorRampPalette(sign_colors)(n.bar)
    }
    for (i in seq_len(n.bar)) {
      graphics::rect(x0, y.seq[i], x1, y.seq[i + 1L],
                     col = pal[i], border = pal[i])
    }
    graphics::rect(x0, y0, x1, y1, border = "grey40")
    br <- if (color_by == "sign") {
      .rhf_pretty_breaks(color.range, n = 10L, symmetric = TRUE)
    } else {
      .rhf_pretty_breaks(color.range, n = 10L,
                         positive.only = FALSE,
                         symmetric = FALSE)
    }
    if (!length(br)) {
      br <- color.range
    }
    if (color_by == "sign") {
      lim <- max(abs(color.range), na.rm = TRUE)
      ypos <- if (is.finite(lim) && lim > 0) {
        y0 + (br + lim) / (2 * lim) * (y1 - y0)
      } else {
        rep(mean(c(y0, y1)), length(br))
      }
    } else {
      rng <- range(color.range, na.rm = TRUE)
      ypos <- if (is.finite(rng[1L]) && is.finite(rng[2L]) && rng[1L] != rng[2L]) {
        y0 + (br - rng[1L]) / diff(rng) * (y1 - y0)
      } else {
        rep(mean(c(y0, y1)), length(br))
      }
    }
    graphics::segments(x1, ypos, x1 + 0.035, ypos, col = "grey35")
    graphics::text(x1 + 0.055,
                   ypos,
                   labels = .rhf_format_legend_values(br),
                   adj = c(0, 0.5),
                   cex = cex.text)
  }
  invisible(TRUE)
}
.rhf_map_palette <- function(x, colors, symmetric = FALSE) {
  x <- as.numeric(x)
  pal <- grDevices::colorRampPalette(colors)(64L)
  idx <- rep(1L, length(x))
  ok <- is.finite(x)
  if (!any(ok)) {
    return(pal[idx])
  }
  if (symmetric) {
    lim <- max(abs(x[ok]), na.rm = TRUE)
    if (!is.finite(lim) || lim <= 0) {
      idx[ok] <- ceiling(length(pal) / 2)
    } else {
      idx[ok] <- 1L + floor((x[ok] + lim) / (2 * lim) * (length(pal) - 1L))
    }
  } else {
    rng <- range(x[ok], na.rm = TRUE)
    if (!is.finite(rng[1L]) || !is.finite(rng[2L]) || rng[1L] == rng[2L]) {
      idx[ok] <- length(pal)
    } else {
      idx[ok] <- 1L + floor((x[ok] - rng[1L]) / (rng[2L] - rng[1L]) * (length(pal) - 1L))
    }
  }
  idx <- pmax(1L, pmin(length(pal), idx))
  pal[idx]
}
.rhf_open_plot_device <- function(out.file, width, height) {
  ext <- tolower(sub("^.*\\.", "", out.file))
  if (!nzchar(ext) || identical(ext, out.file) || ext == "pdf") {
    grDevices::pdf(file = out.file, width = width, height = height)
  }
  else if (ext == "png") {
    grDevices::png(filename = out.file, width = width, height = height,
                   units = "in", res = 300)
  }
  else if (ext %in% c("jpg", "jpeg")) {
    grDevices::jpeg(filename = out.file, width = width, height = height,
                    units = "in", res = 300, quality = 100)
  }
  else if (ext %in% c("tif", "tiff")) {
    grDevices::tiff(filename = out.file, width = width, height = height,
                    units = "in", res = 300, compression = "lzw")
  }
  else {
    stop("Unsupported plot file extension in 'out.file': ", out.file)
  }
  invisible(TRUE)
}
########################################################################
## label helpers for plotting
########################################################################
.rhf_unique_in_order <- function(x) {
  x <- as.character(x)
  x[!duplicated(x)]
}
.rhf_clean_dummy_suffix <- function(s) {
  s <- as.character(s)
  if (!nzchar(s)) {
    return(s)
  }
  s <- gsub("\\.+", " ", s)
  s <- gsub("_+", " ", s)
  s <- trimws(gsub("\\s+", " ", s))
  if (nzchar(s) && identical(s, toupper(s)) && nchar(s) >= 4) {
    s <- tools::toTitleCase(tolower(s))
  }
  s <- sub(" ([A-Z]{2,6})$", " (\\1)", s)
  s
}
.rhf_infer_prefix_label <- function(code, map) {
  code <- as.character(code)
  if (is.null(map) || is.null(names(map))) {
    return(NA_character_)
  }
  keys <- names(map)
  if (!length(keys)) {
    return(NA_character_)
  }
  hit <- keys[startsWith(code, keys)]
  if (!length(hit)) {
    return(NA_character_)
  }
  hit <- hit[order(nchar(hit), decreasing = TRUE)]
  for (k in hit) {
    suf <- substr(code, nchar(k) + 1L, nchar(code))
    if (!nzchar(suf)) {
      next
    }
    first.chr <- substr(suf, 1L, 1L)
    if (first.chr %in% c("_")) {
      next
    }
    if (!grepl("^[A-Za-z0-9]", suf)) {
      next
    }
    base.lab <- as.character(map[[k]])
    if (!nzchar(base.lab) || is.na(base.lab)) {
      next
    }
    suf.clean <- .rhf_clean_dummy_suffix(suf)
    if (!nzchar(suf.clean)) {
      return(base.lab)
    }
    return(paste0(base.lab, ": ", suf.clean))
  }
  NA_character_
}
.rhf_label_lookup <- function(x, map, infer_prefix = TRUE) {
  x <- as.character(x)
  if (is.null(map)) {
    return(x)
  }
  if (is.data.frame(map)) {
    if (ncol(map) < 2L) {
      stop("label map data.frame must have at least 2 columns")
    }
    nms <- names(map)
    if (all(c("variable", "label") %in% nms)) {
      map <- stats::setNames(as.character(map$label), as.character(map$variable))
    }
    else if (all(c("outcome", "label") %in% nms)) {
      map <- stats::setNames(as.character(map$label), as.character(map$outcome))
    }
    else {
      map <- stats::setNames(as.character(map[[2L]]), as.character(map[[1L]]))
    }
  }
  if (is.null(names(map))) {
    return(x)
  }
  if (anyDuplicated(names(map))) {
    map <- map[!duplicated(names(map), fromLast = TRUE)]
  }
  idx <- match(x, names(map))
  out <- as.character(map[idx])
  miss <- is.na(out) | out == ""
  if (any(miss) && isTRUE(infer_prefix)) {
    out2 <- vapply(x[miss], .rhf_infer_prefix_label, character(1), map = map)
    fill.ok <- !is.na(out2) & nzchar(out2)
    out[which(miss)[fill.ok]] <- out2[fill.ok]
    miss <- is.na(out) | out == ""
  }
  out[miss] <- x[miss]
  out
}
.rhf_make_unique_labels <- function(labels, codes) {
  dup <- duplicated(labels) | duplicated(labels, fromLast = TRUE)
  labels[dup] <- paste0(labels[dup], " [", codes[dup], "]")
  labels
}

rhf.workhorse <- function(formula,
                          data,
                          ntree = 500,
                          nsplit = 10,
                          treesize = NULL,
                          nodesize = NULL,
                          block.size = 10,
                          bootstrap = c("by.root", "none", "by.user"),
                          samptype = c("swor", "swr"),
                          samp = NULL,
                          case.wt = NULL,
                          membership = FALSE,
                          sampsize = if (samptype == "swor") function(x){x * .632} else function(x){x},
                          xvar.wt = NULL,
                          ntime = 50,
                          min.events.per.gap = 10,
                          seed = NULL,
                          do.trace = FALSE,
                          ...
                          )
{
  ## hidden options (used later)
  dots <- list(...)
  ## initialize the seed
  seed <- get.seed(seed)
  ## formula, family, xvar.names, yvar.names
  formula.prelim <- parse.formula(as.formula(formula), data, NULL)
  formula.detail <- finalize.formula(formula.prelim, data)
  family <- formula.detail$family
  xvar.names <- formula.detail$xvar.names
  yvar.names <- formula.detail$yvar.names
  subj.names <- formula.detail$subj.names
  ## coherence check
  if (family != "surv-tdc") {
    stop("this function (currently) only works for TDC survival")
  }
  if (length(xvar.names) == 0) {
    stop("something seems wrong: your formula did not define any x-variables")
  }
  if (length(yvar.names) == 0) {
    stop("something seems wrong: your formula did not define any y-variables")
  }
  if (length(subj.names) == 0) {
    stop("something seems wrong: your formula did not identify an id variable")
  }
  ## process y, scale time to [0,1]
  data <- cleanup.counting(data, xvar.names, yvar.names, subj.names)
  max.time <- attr(data, "max.time")
  time.map <- attr(data, "time.map")
  subj <- data[, subj.names]
  subj.unique.count <- length(unique(subj))
  yvar <- data[, yvar.names]
  yfactor <- extract.factor(data, yvar.names)
  yfactor$types <- yvar.types <- get.yvar.type(family, yfactor$generic.types, yvar.names)
  yfactor$nlevels <- yvar.nlevels <- get.yvar.nlevels(family, yfactor$nlevels, yvar.names, yvar)
  yfactor$numeric.levels <- yvar.numeric.levels <- get.numeric.levels(family, yfactor$nlevels, yvar)
  ## process x
  xvar <- data[, xvar.names, drop = FALSE]
  n.xvar <- length(xvar.names)
  n <- nrow(xvar)
  ## We manually patch in the factor object.
  xfactor <- list(factor = character(0),
                  order = character(0),
                  levels = NULL,
                  order.levels = NULL,
                  nlevels = rep(0, n.xvar),
                  types = rep("R", n.xvar))
  xvar.types <- xfactor$types
  xvar.nlevels <- xfactor$nlevels
  xfactor$numeric.levels <- xvar.numeric.levels <- get.numeric.levels(family, xfactor$nlevels, xvar)
  xvar.wt  <- get.weight(xvar.wt, n.xvar)
  ## determine which x-variables are time-dependent
  ## determine which individuals have any time-varying information
  xvar.time <- get.tdc.cov(data, subj.names, yvar.names)
  subj.time <- get.tdc.subj.time(data, subj.names, yvar.names)
  ## don't need the data anymore
  remove(data)
  ## tree grow options
  mtry <- get.mtry(family, n.xvar, dots$mtry)
  nodesize <- get.nodesize(nodesize, any(xvar.time > 0))
  splitinfo <- get.splitinfo(formula.detail, dots$splitrule, nsplit)
  ## hard coded options 
  hcut <- 1
  tune <- FALSE
  augmentX <- augmentXlist <- NULL
  ## experimental options
  experimental.bits <- get.experimental.bits(dots$experimental.bits, FALSE)
  ## missingness not allowed in rhf
  data.pass <- TRUE
  data.pass.bits <- get.data.pass.bits(data.pass)
  ## convert treesize to integer
  lot <- get.lot(hcut,
                 treesize=treesize,
                 tune=tune,
                 lag=dots$lag,
                 strikeout=dots$strikeout,
                 max.two=dots$max.two,
                 max.three=dots$max.three,
                 max.four=dots$max.four)
  if (is.function(lot$treesize)) {
    lot$treesize <- as.integer(lot$treesize(n, any(xvar.time > 0)))
  }
  ## hard coded options (some are legacy)
  perf.type <- NULL
  ensemble <- NULL
  empirical.risk <- TRUE
  forest <- TRUE
  m.target.idx <- 1
  ## Get event information and dimensioning for families
  if (length(ntime) > 1) {
    ## user specified time points have to be mapped to [0,1]
    ntime <- .forward.time(ntime, time.map)
  }
  event.info <- get.grow.event.info(yvar,
                                    family,
                                    ntime = ntime,
                                    min.events.per.gap = min.events.per.gap)
  ## initialize samptype
  samptype <- match.arg(samptype, c("swor", "swr"))
  ## bootstrap specifics
  bootstrap <- match.arg(bootstrap, c("by.root", "none", "by.user"))
  if (bootstrap == "by.root") {
    ## Nominal bootstrap at the root node.
    if (!is.function(sampsize) && !is.numeric(sampsize)) {
      stop("sampsize must be a function or number specifying size of subsampled data")
    }
    if (is.function(sampsize)) {
      sampsize.function <- sampsize
    }
    else {
      ## convert user passed sample size to a function
      sampsize.function <- make.samplesize.function(sampsize / subj.unique.count)
    }
    sampsize <- round(sampsize.function(subj.unique.count))
    if (sampsize < 1) {
      stop("sampsize must be greater than zero")
    }
    ## for swor size is limited by the number of cases
    if (samptype == "swor" && (sampsize > subj.unique.count)) {
      sampsize <- subj.unique.count
      sampsize.function <- function(x){x}
    }
    samp <- NULL
    case.wt  <- get.weight(case.wt, subj.unique.count)
  }
  else if (bootstrap == "by.user") {    
    ## check for coherent sample: it will of be dim [.] x [ntree]
    ## where [.] is the number of rows in the data set or unique subjects in
    ## the case of time dependent covariates.
    if (is.null(samp)) {
      stop("samp must not be NULL when bootstrapping by user")
    }
    ## ntree should be coherent with the sample provided
    ntree <- ncol(samp)
    sampsize <- colSums(samp)
    if (sum(sampsize == sampsize[1]) != ntree) {
      stop("sampsize must be identical for each tree")
    }
    ## set the sample size and function
    sampsize <- sampsize[1]
    sampsize.function <- make.samplesize.function(sampsize[1] / subj.unique.count)
    case.wt  <- get.weight(NULL, subj.unique.count)
  }
  ## This is "none".
  else {
    sampsize <- subj.unique.count
    sampsize.function <- function(x){x}
    case.wt  <- get.weight(case.wt, sampsize)
  }
  ## Turn performance output off unless bootstrapping by root or user.
  if ((bootstrap != "by.root") && (bootstrap != "by.user")) {
    perf.type <- "none"
  }
  ## Set the ensemble option.
  if (!is.null(ensemble)) {
    ## leave the ensemble as is for testing purposes.
  }
  else {
    ensemble <- "all"
    if (bootstrap == "none") {
      ensemble <- "inbag"
    }
  }
  ## perf type
  perf.type <- get.perf(perf.type, family)
  ## Assign low bits for the native code
  bootstrap.bits <- get.bootstrap.bits(bootstrap)
  empirical.risk.bits <- get.empirical.risk.bits(empirical.risk)
  perf.bits <-  get.perf.bits(perf.type)
  ensemble.bits <- get.ensemble.bits(ensemble)
  forest.bits <- get.forest.bits(forest)
  ## Assign high bits for the native code
  membership.bits <-  get.membership.bits(membership)
  samptype.bits <- get.samptype.bits(samptype)
  ## block size
  block.size <- get.block.size.bits(block.size, ntree)
  ## integrated hazard calculation
  wmode <- is.hidden.wmode(dots)    
  ## values of wmode in [1, 2, 3] are converted to local bits.
  wmode.bits <- get.wmode.bits(wmode)
  ## Set the user defined trace.
  do.trace <- get.trace(do.trace)
  ## Start the C external timer.
  ctime.external.start  <- proc.time()
################################################ ##
  nativeOutput <- tryCatch({.Call("entryGrow",
                                  as.integer(do.trace),
                                  as.integer(seed),
                                  as.integer(bootstrap.bits +     ## low option word
                                             empirical.risk.bits +
                                             perf.bits +
                                             ensemble.bits +
                                             forest.bits),  
                                  as.integer(membership.bits +    ## high option word
                                             2^16 + 2^18 +        ## MEMB_OUTG and TERM_OUTG
                                             samptype.bits),
                                  as.integer(experimental.bits + wmode.bits),  ## rhf (local experimental) option word
                                  as.integer(ntree),
                                  as.integer(n),
                                  list(as.integer(subj.unique.count),
                                       if (is.null(case.wt)) NULL else as.double(case.wt),
                                       as.integer(sampsize),
                                       if (is.null(samp)) NULL else as.integer(samp)),
                                  as.integer(splitinfo$index),
                                  as.integer(splitinfo$nsplit),
                                  as.integer(mtry),
                                  as.integer(nodesize),
                                  lot,
                                  list(if (is.null(m.target.idx)) as.integer(0) else as.integer(length(m.target.idx)),
                                       if (is.null(m.target.idx)) NULL else as.integer(m.target.idx)),
                                  list(as.integer(length(yvar.types)),
                                       if (is.null(yvar.types)) NULL else as.character(yvar.types),
                                       ## No factors, so we massage pass zeros.
                                       if (is.null(yvar.types)) NULL else as.integer(yvar.nlevels),
                                       if (is.null(yvar.numeric.levels)) NULL else sapply(1:length(yvar.numeric.levels), function(nn) {as.integer(length(yvar.numeric.levels[[nn]]))}),
                                       if (is.null(subj)) NULL else as.integer(subj),
                                       if (is.null(event.info)) as.integer(0) else as.integer(length(event.info$event.type)),
                                       if (is.null(event.info)) NULL else as.integer(event.info$event.type)),
                                  if (is.null(yvar.nlevels)) {
                                    NULL
                                  }
                                  else {
                                    lapply(1:length(yvar.nlevels),
                                           function(nn) {as.integer(yvar.nlevels[[nn]])})
                                  },
                                  if (is.null(yvar.types)) NULL else as.double(as.vector(data.matrix(yvar))),
                                  list(if(is.null(event.info$time.interest)) as.integer(0) else as.integer(length(event.info$time.interest)),
                                       if(is.null(event.info$time.interest)) NULL else as.double(event.info$time.interest)),
                                  list(as.integer(n.xvar),
                                       as.character(xvar.types),
                                       if (is.null(xvar.types)) NULL else as.integer(xvar.nlevels),
                                       if (is.null(xvar.nlevels)) NULL else sapply(1:length(xvar.nlevels), function(nn) {as.integer(length(xvar.nlevels[[nn]]))}),
                                       if (is.null(xvar.time)) NULL else as.integer(xvar.time),
                                       if (is.null(subj.time)) NULL else as.integer(subj.time)),
                                  if (is.null(xvar.nlevels)) {
                                    NULL
                                  }
                                  else {
                                    lapply(1:length(xvar.nlevels),
                                           function(nn) {as.integer(xvar.nlevels[[nn]])})
                                  },
                                  as.double(as.vector(data.matrix(xvar))),
                                  as.double(xvar.wt),
                                  ## Here is the hack for the augmented x-vars.  The C-side needs to know how many there are,
                                  ## so we have two elements, the first is the dimension, the second is the array. You should
                                  ## replace these with the actual values.
                                  augmentXlist,
                                  as.integer(block.size),
                                  as.integer(get.rf.cores()))},
                           ## error = function(e) {NULL}
                           error = function(e) { print(e); NULL }
                           )
  ## Stop the C external timer.
  ctime.external.stop <- proc.time()
  ## check for error return condition in the native code
  if (is.null(nativeOutput)) {
    stop("An error has occurred on the C-side.  Please turn trace on for further analysis.")
  }
  nativeArraySize  <- 0
  mwcpCountSummary <- 0  
  nullO <- lapply(1:ntree, function(b) {
    if (nativeOutput$leafCount[b] > 0) {
      ## The tree was not rejected.  Count the number of internal
      ## and external (terminal) nodes in the forest.
      nativeArraySize <<- nativeArraySize + (2 * nativeOutput$leafCount[b]) - 1
      mwcpCountSummary <<- mwcpCountSummary + nativeOutput$mwcpCT[b]
    }
    else {
      ## The tree was rejected.  However, it acts as a
      ## placeholder, being a stump topologically and thus
      ## adds to the total node count.
      nativeArraySize <<- nativeArraySize + 1
    }
    NULL
  })
  rm(nullO)  ##memory saver
  ## save the native array
  nativeArray <- as.data.frame(cbind(nativeOutput$treeID[1:nativeArraySize],
                                     nativeOutput$nodeID[1:nativeArraySize],
                                     nativeOutput$nodeSZ[1:nativeArraySize],
                                     nativeOutput$brnodeID[1:nativeArraySize],
                                     nativeOutput$prnodeID[1:nativeArraySize],
                                     nativeOutput$parmID[1:nativeArraySize],
                                     nativeOutput$contPT[1:nativeArraySize],
                                     nativeOutput$mwcpSZ[1:nativeArraySize],
                                     nativeOutput$fsrecID[1:nativeArraySize]
                                     ))
  colnames(nativeArray) <- c("treeID", "nodeID", "nodeSZ", "brnodeID", "prnodeID", "parmID", "contPT", "mwcpSZ", "fsrecID")
  ## save the native factor array
  if (mwcpCountSummary > 0) {
    ## This can be NULL if there are no factor splits along this dimension.
    nativeFactorArray <- as.data.frame(cbind(nativeOutput$mwcpPT[1:mwcpCountSummary]))
    colnames(nativeFactorArray)  <- c("mwcpPT")
  }
  else {
    nativeFactorArray  <- NULL
  }
  ## save the forest
  lot <- c(lot, tune=tune)
  forest.out  <- list(forest = TRUE,
                      ntree = ntree,
                      n = n,
                      subj.names = subj.names,
                      id = subj,
                      yvar = .scale.yvar(yvar, time.map),
                      yvar.names = yvar.names,
                      yvar.factor = yfactor,
                      event.info = event.info,
                      xvar = xvar,
                      xvar.names = xvar.names,
                      xvar.factor = xfactor,
                      xvar.time = xvar.time,
                      seed = nativeOutput$seed,
                      nativeArray = nativeArray,
                      nativeFactorArray = nativeFactorArray,
                      leafCount = nativeOutput$leafCount,
                      totalNodeCount = dim(nativeArray)[1],
                      nodesize = nodesize,
                      perf.type = perf.type,
                      lot = lot,
                      family = family,
                      ##                        bootstrap = bootstrap,
                      ##                        samptype = samptype,
                      experimental.bits = experimental.bits,
                      data.pass = data.pass,
                      parms = list(splitrule = splitinfo$splitrule,
                                   ntree = ntree,
                                   mtry = mtry,
                                   nodesize = nodesize,
                                   ntime = if (length(ntime) > 1) .inverse.time(ntime, time.map) else ntime,
                                   nsplit = splitinfo$nsplit,
                                   experimental.bits = experimental.bits,
                                   bootstrap = bootstrap,
                                   case.wt = case.wt,
                                   sampsize = sampsize,             
                                   samp = samp, 
                                   samptype = samptype,
                                   xvar.wt = xvar.wt),
                      trmbrCaseCt = nativeOutput$trmbrCaseCt,
                      timbrCaseCt = nativeOutput$timbrCaseCt,
                      tombrCaseCt = nativeOutput$tombrCaseCt,
                      trmbrCaseId = nativeOutput$trmbrCaseId,
                      timbrCaseId = nativeOutput$timbrCaseId,
                      tombrCaseId = nativeOutput$tombrCaseId,
                      ibgSizeCase = nativeOutput$ibgSizeCase,
                      oobSizeCase = nativeOutput$oobSizeCase,
                      terminal.qualts = TRUE,
                      terminal.quants = TRUE,
                      ## add x,y values here as needed
                      xvar.time = xvar.time,
                      version = "1.0.1")
  empr.risk <- NULL
  oob.empr.risk <- NULL
  nodeStat <- NULL
  if (!is.null(nativeOutput$bsfOrder)) {
    bsf.order <- nativeOutput$bsfOrder
    nativeOutput$bsfOrder <- NULL
  }
  if (membership) {
    pseudo.membership <- matrix(nativeOutput$nodeMembership, c(n, ntree))
    inbag.out <- matrix(nativeOutput$bootstrapCount, c(subj.unique.count, ntree))
    nativeOutput$nodeMembership <- NULL
    nativeOutput$bootstrapCount <- NULL
  }
  else {
    pseudo.membership <- NULL
    inbag.out  <- NULL
  }
  if (!is.null(nativeOutput$ensembleID)) {
    ensemble.id <- nativeOutput$ensembleID
  }
  else {
    ensemble.id <- NULL
  }
  if (!is.null(nativeOutput$ibgEnsbHazard)) {
    hazard.ibg  <- array(nativeOutput$ibgEnsbHazard, c(subj.unique.count, length(event.info$time.interest)))
  } else {
    hazard.ibg <- NULL
  }
  if (!is.null(nativeOutput$oobEnsbHazard)) {
    hazard.oob  <- array(nativeOutput$oobEnsbHazard, c(subj.unique.count, length(event.info$time.interest)))
  } else {
    hazard.oob <- NULL
  }
  if (!is.null(nativeOutput$ibgEnsbNlsnAaln)) {
    chf.ibg  <- array(nativeOutput$ibgEnsbNlsnAaln, c(subj.unique.count, length(event.info$time.interest)))
  } else {
    chf.ibg <- NULL
  }
  if (!is.null(nativeOutput$oobEnsbNlsnAaln)) {
    chf.oob  <- array(nativeOutput$oobEnsbNlsnAaln, c(subj.unique.count, length(event.info$time.interest)))
  } else {
    chf.oob <- NULL
  }
  if (!is.null(nativeOutput$ibgRisk)) {
    risk.ibg <- nativeOutput$ibgRisk
  } else {
    risk.ibg <- NULL
  }
  if (!is.null(nativeOutput$oobRisk)) {
    risk.oob <- nativeOutput$oobRisk
  } else {
    risk.oob <- NULL
  }
  if (!is.null(nativeOutput$ibgWCase)) {
    int.haz.ibg <- nativeOutput$ibgWCase
  } else {
    int.haz.ibg <- NULL
  }
  if (!is.null(nativeOutput$oobWCase)) {
    int.haz.oob <- nativeOutput$oobWCase
  } else {
    int.haz.oob <- NULL
  }
  ## New left and right hand time points over which the case hazard was integrated.
  if (!is.null(nativeOutput$absWCaseTimeLeft)) {
    int.haz.left <- .inverse.time(nativeOutput$absWCaseTimeLeft, time.map)
  } else {
    int.haz.left <- NULL
  }
  if (!is.null(nativeOutput$absWCaseTimeRight)) {
    int.haz.right <- .inverse.time(nativeOutput$absWCaseTimeRight, time.map)
  } else {
    int.haz.right <- NULL
  }
  if (empirical.risk) {
    if (!is.null(nativeOutput$nodeRisk)) {
      node.risk <- array(nativeOutput$nodeRisk, c(nativeArraySize, 1))
      nativeOutput$nodeRisk <- NULL
    }
    if (!is.null(nativeOutput$ibgTreeRisk)) {
      tree.risk.ibg <- array(nativeOutput$ibgTreeRisk, c(lot$treesize, ntree))
      nativeOutput$ibgTreeRisk <- NULL
    }
    if (!is.null(nativeOutput$oobTreeRisk)) {
      tree.risk.oob <- array(nativeOutput$oobTreeRisk, c(lot$treesize, ntree))
      nativeOutput$oobTreeRisk <- NULL
    }
    ## Not populated for now.
    if (!is.null(nativeOutput$nodeStat)) {
      nodeStat <- array(nativeOutput$nodeStat[1:nativeArraySize])
      nativeOutput$nodeStat <- NULL
    }
  }
  if (!is.null(nativeOutput$tNelsonAalen)) {
    t.chf <- vector("list", ntree)
    offset = 0
    for (i in 1:ntree) {
      t.chf[[i]] <- matrix(nativeOutput$tNelsonAalen[(offset+1):(offset + (nativeOutput$leafCount[i] * length(event.info$time.interest)))],
                           c(nativeOutput$leafCount[i], length(event.info$time.interest)), byrow=TRUE)
      offset = offset + (nativeOutput$leafCount[i] * length(event.info$time.interest))
    }
  } else {
    t.chf <- NULL
  }
  if (!is.null(nativeOutput$tHazard)) {
    t.hazard <- vector("list", ntree)
    offset = 0
    for (i in 1:ntree) {
      t.hazard[[i]] <- matrix(nativeOutput$tHazard[(offset+1):(offset + (nativeOutput$leafCount[i] * length(event.info$time.interest)))],
                              c(nativeOutput$leafCount[i], length(event.info$time.interest)), byrow=TRUE)
      offset = offset + (nativeOutput$leafCount[i] * length(event.info$time.interest))
    }
  } else {
    t.hazard <- NULL
  }
  if (!is.null(nativeOutput$tHazardTimeCnt)) {
    t.haz.time.cnt  <- nativeOutput$tHazardTimeCnt
  }
  else {
    t.haz.time.cnt  <- NULL
  }
  if (!is.null(nativeOutput$tHazardTimeIdx)) {
    t.haz.time.idx <- vector("list", n)
    offset = 0
    for (i in 1:n) {
      if (t.haz.time.cnt[i] != 0) {
        t.haz.time.idx[[i]]  <- nativeOutput$tHazardTimeIdx[(offset+1):(offset + t.haz.time.cnt[i])]
        offset = offset + t.haz.time.cnt[i]
      }
      else {
        t.haz.time.idx[[i]] <- NA_integer_
      }
    }
  }
  else {
    t.haz.time.idx  <- NULL
  }
  ## Add the terminal quantitative information to the forest object for now.
  hz.scale <- .hazard.scale(event.info$time.interest, time.map)
  if (!is.null(hazard.ibg)) {
    hazard.ibg <- sweep(hazard.ibg, 2L, hz.scale, "*")
  }
  if (!is.null(hazard.oob)) {
    hazard.oob <- sweep(hazard.oob, 2L, hz.scale, "*")
  }
  if (!is.null(t.hazard)) {
    t.hazard <- lapply(t.hazard, function(h) sweep(h, 2L, hz.scale, "*"))
  }
  forest.out  <- append(forest.out,
                        list(time.map = time.map,
                             t.chf = t.chf,
                             t.hazard = t.hazard,
                             t.haz.time.cnt = t.haz.time.cnt,
                             t.haz.time.idx = t.haz.time.idx))
  ## Initialize the default class of the forest.
  class(forest.out) <- c("rhf", "forest", family)
  ## set this to NULL 
  err.rate <- NULL
  ## make the output object    
  rhfOutput  <- list(
    family = family,
    n = n,
    ntree = ntree,
    hcut = hcut,
    max.time = max.time,
    time.map = time.map,
    time.interest = .inverse.time(event.info$time.interest, time.map),
    hazard.inbag = hazard.ibg,
    hazard.oob = hazard.oob,
    chf.inbag = chf.ibg,
    chf.oob   = chf.oob,
    risk.inbag = risk.ibg,
    risk.oob   = risk.oob,
    int.haz.inbag = int.haz.ibg,
    int.haz.oob = int.haz.oob,
    int.haz.left = int.haz.left,
    int.haz.right = int.haz.right,
    event.info = event.info,
    ensemble.id = ensemble.id,
    splitrule = splitinfo$splitrule,
    splitinfo = splitinfo,
    subj.names = subj.names,
    id = subj,
    yvar = as.data.frame(.scale.yvar(yvar, time.map)),
    yvar.names = yvar.names,
    yvar.factor = yfactor,
    yvar.types = yvar.types,
    xvar = xvar,
    xvar.names = xvar.names,
    xvar.types = xvar.types,
    xvar.time = xvar.time,
    forest = forest.out,
    node.risk = node.risk,
    tree.risk.inbag = tree.risk.ibg,
    tree.risk.oob   = tree.risk.oob,
    nodeStat = nodeStat,
    bsf.order = bsf.order,
    pseudo.membership = pseudo.membership,
    inbag = inbag.out,
    err.rate = err.rate,
    ctime.internal = nativeOutput$cTimeInternal,
    ctime.external = ctime.external.stop - ctime.external.start
  )
  ## save the call/formula for the return object
  my.call <- match.call()
  my.call$formula <- eval(formula)
  rhfOutput$call <- my.call
  ## return the object
  class(rhfOutput) <- c("rhf", "grow", family)
  return(rhfOutput)
}

.rhf_delta <- function(time) {
  time <- as.numeric(time)
  if (length(time) < 1L) stop("time must have positive length")
  if (anyNA(time)) stop("time contains NA")
  if (any(diff(time) <= 0)) stop("time must be strictly increasing")
  c(time[1L], diff(time))
}
.rhf_chf_from_hazard <- function(hazard, time) {
  # hazard: n x q matrix
  hazard <- as.matrix(hazard)
  time <- as.numeric(time)
  q <- length(time)
  if (ncol(hazard) != q) stop("hazard ncol must match length(time)")
  dt <- .rhf_delta(time)
  hd <- sweep(hazard, 2L, dt, `*`)
  t(apply(hd, 1L, cumsum))
}
.rhf_loess_smooth_vec <- function(z, x, span, degree, family, suppress.warnings = TRUE) {
  # Smooth a single curve z(x). Returns vector length(x).
  z <- as.numeric(z)
  x <- as.numeric(x)
  ok <- is.finite(z) & is.finite(x)
  if (sum(ok) < max(6L, degree + 2L)) return(z)
  df <- data.frame(x = x[ok], z = z[ok])
  .muffle <- function(expr) {
    if (!suppress.warnings) return(expr)
    withCallingHandlers(expr, warning = function(w) invokeRestart("muffleWarning"))
  }
  fit <- try(
    .muffle(
      stats::loess(
        z ~ x,
        data = df,
        span = span,
        degree = degree,
        family = family,
        control = stats::loess.control(surface = "direct")
      )
    ),
    silent = TRUE
  )
  if (inherits(fit, "try-error")) return(z)
  pred <- try(
    .muffle(stats::predict(fit, newdata = data.frame(x = x))),
    silent = TRUE
  )
  if (inherits(pred, "try-error")) return(z)
  pred <- as.numeric(pred)
  if (all(!is.finite(pred))) return(z)
  # Fill any NA/Inf predictions with original values
  bad <- !is.finite(pred)
  if (any(bad)) pred[bad] <- z[bad]
  pred
}
.rhf_smooth_loghazard_matrix <- function(Z, time, span, degree, family, suppress.warnings = TRUE) {
  Z <- as.matrix(Z)
  time <- as.numeric(time)
  if (ncol(Z) != length(time)) stop("Z ncol must match length(time)")
  t(apply(
    Z, 1L, .rhf_loess_smooth_vec,
    x = time, span = span, degree = degree, family = family,
    suppress.warnings = suppress.warnings
  ))
}
.rhf_map_record_ids <- function(o) {
  id_raw <- o$id
  if (is.null(id_raw)) stop("RHF object is missing $id")
  if (!is.null(o$ensemble.id)) {
    id <- match(id_raw, o$ensemble.id)
    if (anyNA(id)) stop("Unable to map o$id onto o$ensemble.id")
  } else {
    id <- id_raw
  }
  as.integer(id)
}
.rhf_get_hazard_matrix <- function(o, target = c("oob", "inbag", "test")) {
  target <- match.arg(target)
  nm <- paste0("hazard.", target)
  if (!is.null(o[[nm]])) return(as.matrix(o[[nm]]))
  stop("Object is missing ", nm)
}
.rhf_loghazard_median_tree <- function(o, target = c("oob", "inbag", "both", "test"), eps = 1e-12) {
  target <- match.arg(target)
  if (is.null(o$forest) || is.null(o$forest$t.hazard)) {
    stop("RHF object must contain forest$t.hazard. Fit with forest=TRUE and keep.forest=TRUE.")
  }
  time <- as.numeric(o$time.interest)
  if (length(time) < 1L) stop("Missing/empty o$time.interest")
  y <- o$yvar
  if (is.null(y) || ncol(y) < 2L) stop("Missing o$yvar with start/stop")
  start <- as.numeric(y[, 1L])
  stop  <- as.numeric(y[, 2L])
  id <- .rhf_map_record_ids(o)
  leafCount <- as.integer(o$forest$leafCount)
  B <- length(leafCount)
  totalLeaves <- sum(leafCount)
  if (target == "test") {
    ## pull test membership from prediction object
    tCt <- o$ttmbrCaseCt
    tId <- o$ttmbrCaseId
    if (is.null(tCt) || is.null(tId)) {
      stop("target='test' requires o$ttmbrCaseCt and o$ttmbrCaseId (returned by predict()).")
    }
    tCt <- as.integer(tCt)
    tId <- as.integer(tId)
    if (length(tCt) != totalLeaves) stop("length(ttmbrCaseCt) must equal sum(leafCount).")
    if (sum(tCt) != length(tId)) stop("sum(ttmbrCaseCt) must equal length(ttmbrCaseId).")
    ## dummy inbag: all zeros => all trees treated as oob => "use all trees"
    nSubj <- if (!is.null(o$ensemble.id)) length(o$ensemble.id) else max(id, na.rm = TRUE)
    inbag <- matrix(0L, nSubj, B)
    ## empty inbag membership lists
    timbrCaseCt <- integer(totalLeaves)
    timbrCaseId <- integer(0)
    ## put test membership into "oob membership"
    tombrCaseCt <- tCt
    tombrCaseId <- tId
    out <- .Call(
      "medianLogStitchedHazard",
      as.integer(id),
      as.numeric(start),
      as.numeric(stop),
      as.numeric(time),
      inbag,
      leafCount,
      timbrCaseCt,
      timbrCaseId,
      tombrCaseCt,
      tombrCaseId,
      o$forest$t.hazard,
      as.integer(0L),        # targetCode = oob
      as.numeric(eps)
    )
    return(list(loghazard.test = out$loghazard.oob))
  }
  ## existing training path (unchanged)
  req <- c("timbrCaseCt","timbrCaseId","tombrCaseCt","tombrCaseId")
  miss <- req[!req %in% names(o$forest)]
  if (length(miss) > 0L) stop("RHF object forest is missing: ", paste(miss, collapse = ", "))
  inbag <- o$inbag
  if (!is.matrix(inbag)) inbag <- as.matrix(inbag)
  storage.mode(inbag) <- "integer"
  timbrCaseCt <- as.integer(o$forest$timbrCaseCt)
  timbrCaseId <- as.integer(o$forest$timbrCaseId)
  tombrCaseCt <- as.integer(o$forest$tombrCaseCt)
  tombrCaseId <- as.integer(o$forest$tombrCaseId)
  targetCode <- switch(target, oob = 0L, inbag = 1L, both = 2L)
  .Call(
    "medianLogStitchedHazard",
    as.integer(id),
    as.numeric(start),
    as.numeric(stop),
    as.numeric(time),
    inbag,
    leafCount,
    timbrCaseCt,
    timbrCaseId,
    tombrCaseCt,
    tombrCaseId,
    o$forest$t.hazard,
    as.integer(targetCode),
    as.numeric(eps)
  )
}
smoothed.hazard.rhf <- function(
    o,
    method = c("median.loess", "loess"),
    oob = TRUE,
    span = 0.35,
    degree = 1,
    family = NULL,
    eps = 1e-12,
    suppress.warnings = TRUE,
    trace = FALSE )
{
  method <- match.arg(method)
  if (is.null(family)) {
    family <- if (method == "median.loess") "gaussian" else "symmetric"
  }
  family <- match.arg(family, choices = c("gaussian", "symmetric"))
  time <- as.numeric(o$time.interest)
  if (length(time) < 1L) stop("Missing/empty o$time.interest")
  # output object: keep only what downstream tools typically need
  out <- list(
    id = o$id,
    ensemble.id = o$ensemble.id,
    yvar = o$yvar,
    time.interest = time
  )
  ## set the target: oob,inbag, test
  if (inherits(o, "predict")) {
    target <- "test"
  }
  else {
    target <- ifelse(oob, "oob", "inbag")
  }
  do_oob   <- target %in% c("oob", "both")
  do_inbag <- target %in% c("inbag", "both")
  do_test  <- identical(target, "test")  
  if (method == "loess") {
    if (do_oob) {
      H <- .rhf_get_hazard_matrix(o, "oob")
      Z <- log(pmax(H, 0) + eps)
      Zs <- .rhf_smooth_loghazard_matrix(
        Z, time,
        span = span, degree = degree, family = family,
        suppress.warnings = suppress.warnings
      )
      Hs <- pmax(exp(Zs) - eps, 0)
      out$hazard.oob <- Hs
      out$chf.oob <- .rhf_chf_from_hazard(Hs, time)
      if (trace) message("smoothed oob hazard via loess")
    }
    if (do_inbag) {
      H <- .rhf_get_hazard_matrix(o, "inbag")
      Z <- log(pmax(H, 0) + eps)
      Zs <- .rhf_smooth_loghazard_matrix(
        Z, time,
        span = span, degree = degree, family = family,
        suppress.warnings = suppress.warnings
      )
      Hs <- pmax(exp(Zs) - eps, 0)
      out$hazard.inbag <- Hs
      out$chf.inbag <- .rhf_chf_from_hazard(Hs, time)
      if (trace) message("smoothed inbag hazard via loess")
    }
    if (do_test) {
      H <- .rhf_get_hazard_matrix(o, "test")
      Z <- log(pmax(H, 0) + eps)
      Zs <- .rhf_smooth_loghazard_matrix(
        Z, time,
        span = span, degree = degree, family = family,
        suppress.warnings = suppress.warnings
      )
      Hs <- pmax(exp(Zs) - eps, 0)
      out$hazard.test <- Hs
      out$chf.test <- .rhf_chf_from_hazard(Hs, time)
      if (trace) message("smoothed test hazard via loess")
    }
    class(out) <- c("smoothed.hazard.rhf", "rhf")
    return(out)
  }
  # method == 'median.loess'
  if (do_test) {
    if (trace) message("computing TEST median log-hazards across trees via C")
    med <- .rhf_loghazard_median_tree(o, target = "test", eps = eps)  # you add this mode
    Z <- med$loghazard.test
    Zs <- .rhf_smooth_loghazard_matrix(Z, time, span, degree, family,
                                       suppress.warnings = suppress.warnings)
    Hs <- pmax(exp(Zs) - eps, 0)
    out$hazard.test <- Hs
    out$chf.test    <- .rhf_chf_from_hazard(Hs, time)
    if (trace) message("smoothed test hazard via median(loghazard) + loess")
    return(out)  # optional, but keeps logic simple
  }
  else {
    # existing training path unchanged:
    if (trace) message("computing train median log-hazards across trees via C")
      # Compute only the requested target (oob or inbag) for speed
      medTarget <- if (target == "both") "both" else if (do_oob) "oob" else "inbag"
      med <- .rhf_loghazard_median_tree(o, target = medTarget, eps = eps)
      if (do_oob) {
        Z <- med$loghazard.oob
        Zs <- .rhf_smooth_loghazard_matrix(
            Z, time,
            span = span, degree = degree, family = family,
            suppress.warnings = suppress.warnings
        )
        Hs <- pmax(exp(Zs) - eps, 0)
        out$hazard.oob <- Hs
        out$chf.oob <- .rhf_chf_from_hazard(Hs, time)
        if (trace) message("smoothed oob hazard via median(loghazard) + loess")
    }
    if (do_inbag) {
        Z <- med$loghazard.inbag
        Zs <- .rhf_smooth_loghazard_matrix(
            Z, time,
            span = span, degree = degree, family = family,
            suppress.warnings = suppress.warnings
        )
        Hs <- pmax(exp(Zs) - eps, 0)
        out$hazard.inbag <- Hs
        out$chf.inbag <- .rhf_chf_from_hazard(Hs, time)
        if (trace) message("smoothed inbag hazard via median(loghazard) + loess")
    }
    class(out) <- c("smoothed.hazard.rhf", "rhf")
    return(out)
  }
}
smoothed.hazard <- smoothed.hazard.rhf

print.rhf <- function(x, digits = 3, label.width = 34, ...) {
  ## ---- Class / mode -------------------------------------------------------
  is_rhf     <- inherits(x, "rhf")
  is_grow    <- "grow" %in% class(x)
  is_predict <- any(c("predict", "predict.rhf") %in% class(x))
  if (!(is_rhf && (is_grow || is_predict))) {
    stop("print.rhf() supports only objects of class c('rhf','grow') or c('rhf','predict').")
  }
  has_test  <- is_predict && !is.null(x$risk.test)
  is_restore <- is_predict && !has_test
  ## ---- Helpers ------------------------------------------------------------
  .lab <- function(label) sprintf("%*s: ", label.width, label)
  .fmt <- function(label, value) paste0(.lab(label), value)
  .is_num1 <- function(v) is.numeric(v) && length(v) == 1 && is.finite(v)
  fnum <- function(v) if (.is_num1(v)) format(round(v, digits), scientific = FALSE, trim = TRUE) else "NA"
  fint <- function(v) if (.is_num1(v)) format(as.integer(round(v)), big.mark = ",", trim = TRUE) else "NA"
  ## ---- Identify yvar indices (id/status if available) ---------------------
  status_idx <- NA_integer_
  id_idx     <- NA_integer_
  if (!is.null(x$yvar.names)) {
    id_idx     <- match("id",     x$yvar.names, nomatch = NA_integer_)
    status_idx <- match("status", x$yvar.names, nomatch = NA_integer_)
  } else if (!is.null(colnames(x$yvar))) {
    id_idx     <- match("id",     colnames(x$yvar), nomatch = NA_integer_)
    status_idx <- match("status", colnames(x$yvar), nomatch = NA_integer_)
  }
  if (is.na(status_idx) && !is.null(x$yvar) && NCOL(x$yvar) >= 3) status_idx <- 3
  ## ---- Counts: grow/restore vs predict-with-test --------------------------
  n_records <- NA_real_
  n_ids     <- NA_real_
  rec_per   <- NA_real_
  if (is_grow || is_restore) {
    ## Training domain
    n_records <- if (!is.null(x$n)) as.integer(x$n) else if (!is.null(x$yvar)) NROW(x$yvar) else NA_integer_
    n_ids <- if (!is.null(x$id)) length(unique(x$id))
             else if (!is.null(x$yvar) && !is.na(id_idx) && id_idx <= NCOL(x$yvar))
               length(unique(x$yvar[, id_idx])) else NA_integer_
    rec_per <- if (.is_num1(n_records) && .is_num1(n_ids) && n_ids > 0) n_records / n_ids else NA_real_
  } else {
    ## Predict with test data
    ## Cases (prefer xvar rows; yvar rows may be CP or condensed; n may be training n)
    n_cases <- if (!is.null(x$xvar)) NROW(x$xvar)
               else if (!is.null(x$id)) length(unique(x$id))
               else if (!is.null(x$yvar)) NROW(x$yvar)
               else if (!is.null(x$n)) as.integer(x$n) else NA_integer_
    ## Unique IDs for test
    n_ids <- if (!is.null(x$id) && length(x$id) == n_cases) length(unique(x$id))
             else if (!is.null(x$yvar) && !is.na(id_idx) && id_idx <= NCOL(x$yvar))
               length(unique(x$yvar[, id_idx]))
             else n_cases
    ## Records: if yvar carries CP rows and is >= ids, use it; otherwise fall back to cases
    if (!is.null(x$yvar)) {
      nr <- NROW(x$yvar)
      n_records <- if (.is_num1(nr) && .is_num1(n_ids) && nr >= n_ids) nr else n_cases
    } else {
      n_records <- n_cases
    }
    ## Note: rec_per from test rows can be misleading (many predict objects don't carry CP rows),
    ## so we intentionally do NOT print average records/case for predict-with-test.
  }
  ## ---- Deaths/events (available if yvar has status) -----------------------
  n_deaths <- if (!is.null(x$yvar) && !is.na(status_idx)) sum(x$yvar[, status_idx], na.rm = TRUE) else NA_integer_
  ## ---- Forest stats (always present) --------------------------------------
  avg_tree_size <- if (!is.null(x$forest$leafCount)) mean(as.numeric(x$forest$leafCount), na.rm = TRUE) else NA_real_
  nodeSZ <- NA_real_
  na_arr <- x$forest$nativeArray
  if (!is.null(na_arr) && all(c("brnodeID", "nodeSZ", "treeID") %in% names(na_arr))) {
    arr0 <- na_arr[!is.na(na_arr$brnodeID) & na_arr$brnodeID == 0, , drop = FALSE]
    if (NROW(arr0)) nodeSZ <- mean(tapply(arr0$nodeSZ, arr0$treeID, mean, na.rm = TRUE), na.rm = TRUE)
  }
  ## ---- Time-varying features & TDC line ----------------------------------
  n_tdc <- if (!is.null(x$xvar.time)) sum(x$xvar.time > 0, na.rm = TRUE)
           else if (!is.null(x$forest$xvar.time)) sum(x$forest$xvar.time > 0, na.rm = TRUE)
           else NA_integer_
  fam <- if (!is.null(x$family)) tolower(as.character(x$family)) else ""
  is_tdc_family <- grepl("tdc", fam)
  if (is_grow || is_restore) {
    tdc_note <- if (!isTRUE(rec_per > 1)) {
      "NO (records per ID = 1)"
    } else if (!isTRUE(n_tdc > 0)) {
      "NO (0 time-varying features)"
    } else "YES"
  } else {
    ## predict-with-test: CP rows may not be present; base on model + features
    if (!isTRUE(is_tdc_family)) {
      tdc_note <- "NO (family is not TDC)"
    } else if (!isTRUE(n_tdc > 0)) {
      tdc_note <- "NO (0 time-varying features)"
    } else {
      tdc_note <- "YES"
    }
  }
  ## ---- Risk vectors / ordering -------------------------------------------
  risks <- list(
    "In-bag" = if (!is.null(x$risk.inbag)) x$risk.inbag[is.finite(x$risk.inbag)] else NULL,
    "OOB"    = if (!is.null(x$risk.oob))   x$risk.oob[is.finite(x$risk.oob)]     else NULL,
    "Test"   = if (!is.null(x$risk.test))  x$risk.test[is.finite(x$risk.test)]   else NULL
  )
  rmean <- function(v) if (!is.null(v) && length(v)) mean(v, na.rm = TRUE) else NA_real_
  ## ---- Print --------------------------------------------------------------
  out <- c(
    .fmt("Number of records", fint(n_records)),
    .fmt("Sample size (unique IDs)", fint(n_ids))
  )
  if (is_grow || is_restore) {
    out <- c(out, .fmt("Average records per ID",
                       fnum(if (.is_num1(n_records) && .is_num1(n_ids) && n_ids > 0) n_records / n_ids else NA_real_)))
  }
  if (!is.na(n_deaths)) out <- c(out, .fmt("Number of deaths/events", fint(n_deaths)))
  if (!is.null(x$sampsize))  out <- c(out, .fmt("Tree sample size", paste(x$sampsize, collapse = ", ")))
  if (!is.null(x$ntree))     out <- c(out, .fmt("Number of trees",  fint(x$ntree)))
  out <- c(out,
           .fmt("Average tree size (leaves)", fnum(avg_tree_size)),
           .fmt("Average node size",          fnum(nodeSZ)))
  if (!is.null(x$forest$parms) && !is.null(x$forest$parms$mtry)) {
    out <- c(out, .fmt("No. features tried at each split", fint(x$forest$parms$mtry)))
  }
  out <- c(out, .fmt("Total no. time-varying features", if (!is.na(n_tdc)) fint(n_tdc) else "NA"))
  if (!is.null(x$forest$parms) && !is.null(x$forest$parms$xvar.wt)) {
    out <- c(out, .fmt("Total no. features", fint(length(x$forest$parms$xvar.wt))))
  }
  if (!is.null(x$family))    out <- c(out, .fmt("Family",    as.character(x$family)))
  if (!is.null(x$splitrule)) out <- c(out, .fmt("Splitrule", as.character(x$splitrule)))
  if (!is.null(x$forest$parms) && !is.null(x$forest$parms$nsplit)) {
    out <- c(out, .fmt("No. of random splits", fint(x$forest$parms$nsplit)))
  }
  ## Risk lines
  if (is_grow || is_restore) {
    oob_lbl <- if (is_restore) "OOB (restore)" else "OOB"
    if (!is.null(risks[["OOB"]])) {
      out <- c(out, .fmt(paste0(oob_lbl, " risk"), fnum(rmean(risks[["OOB"]]))))
    }
    if (is_restore) out <- c(out, .fmt("Predict restore mode", "YES (using original forest)"))
  } else {
    if (!is.null(risks[["Test"]])) {
      out <- c(out, .fmt("Test risk", fnum(rmean(risks[["Test"]]))))
    }
    if (!is.null(risks[["OOB"]])) {
      out <- c(out, .fmt("OOB risk",  fnum(rmean(risks[["OOB"]]))))
    }
  }
  out <- c(out, .fmt("TDC analysis", tdc_note))
  message(paste(out, collapse = "\n"))
  invisible(x)
}

########################################################################
## display helpers: robust summaries, formatting, and display capping
########################################################################
.rhf_nonmissing_median <- function(x) {
  x <- x[is.finite(x)]
  if (!length(x)) {
    NA_real_
  } else {
    stats::median(x)
  }
}
.rhf_nonmissing_sum <- function(x) {
  x <- x[is.finite(x)]
  if (!length(x)) {
    NA_real_
  } else {
    sum(x)
  }
}
.rhf_nonmissing_quantile <- function(x,
                                     prob = 0.90) {
  x <- x[is.finite(x)]
  if (!length(x)) {
    NA_real_
  } else {
    as.numeric(stats::quantile(x,
                               probs = prob,
                               na.rm = TRUE,
                               names = FALSE))
  }
}
.rhf_row_summary <- function(mat,
                             rank.by = c("q90", "median", "mean", "max", "sum"),
                             abs = FALSE) {
  rank.by <- match.arg(rank.by)
  if (!is.matrix(mat)) {
    mat <- as.matrix(mat)
  }
  if (isTRUE(abs)) {
    mat <- abs(mat)
  }
  switch(
    rank.by,
    q90 = apply(mat, 1L, .rhf_nonmissing_quantile, prob = 0.90),
    median = apply(mat, 1L, .rhf_nonmissing_median),
    mean = apply(mat, 1L, .rhf_nonmissing_mean),
    max = apply(mat, 1L, .rhf_nonmissing_max),
    sum = apply(mat, 1L, .rhf_nonmissing_sum)
  )
}
.rhf_format_importance_numbers <- function(x,
                                           digits = 4L,
                                           scientific.threshold = 1e4) {
  x <- as.numeric(x)
  digits <- max(1L, as.integer(digits)[1L])
  scientific.threshold <- as.numeric(scientific.threshold)[1L]
  out <- rep(NA_character_, length(x))
  out[is.nan(x)] <- "NaN"
  out[is.infinite(x) & x > 0] <- "Inf"
  out[is.infinite(x) & x < 0] <- "-Inf"
  ok <- is.finite(x)
  if (!any(ok)) {
    return(out)
  }
  xx <- signif(x[ok], digits = digits)
  sci <- abs(xx) >= scientific.threshold |
    (abs(xx) > 0 & abs(xx) < 10^(-digits))
  yy <- rep("", length(xx))
  if (any(sci)) {
    yy[sci] <- format(xx[sci], trim = TRUE, scientific = TRUE)
  }
  if (any(!sci)) {
    yy[!sci] <- format(xx[!sci], trim = TRUE, scientific = FALSE)
  }
  out[ok] <- yy
  out
}
.rhf_cap_label <- function(prob) {
  pct <- 100 * as.numeric(prob)[1L]
  out <- formatC(pct, format = "f", digits = 1)
  out <- sub("\\.?0+$", "", out)
  paste0("q", out)
}
.rhf_parse_display_cap <- function(cap,
                                   arg = "display.cap") {
  if (is.null(cap)) {
    return(list(enabled = FALSE,
                prob = NA_real_,
                label = "none"))
  }
  if (is.character(cap)) {
    if (length(cap) != 1L) {
      stop("'", arg, "' must have length 1.")
    }
    cap.chr <- tolower(trimws(cap))
    if (cap.chr %in% c("none", "off", "false", "no")) {
      return(list(enabled = FALSE,
                  prob = NA_real_,
                  label = "none"))
    }
    cap.chr <- sub("^q", "", cap.chr)
    prob <- suppressWarnings(as.numeric(cap.chr))
  }
  else {
    prob <- as.numeric(cap)[1L]
  }
  if (!is.finite(prob)) {
    stop("'", arg, "' must be NULL, 'none', a quantile label like 'q99', or a number in (0, 1].")
  }
  if (prob > 1) {
    prob <- prob / 100
  }
  if (!is.finite(prob) || prob <= 0 || prob > 1) {
    stop("'", arg, "' must be NULL, 'none', a quantile label like 'q99', or a number in (0, 1].")
  }
  if (prob >= 1 - sqrt(.Machine$double.eps)) {
    return(list(enabled = FALSE,
                prob = prob,
                label = .rhf_cap_label(prob)))
  }
  list(enabled = TRUE,
       prob = prob,
       label = .rhf_cap_label(prob))
}
.rhf_cap_values <- function(x,
                            cap = 0.99,
                            symmetric = FALSE,
                            positive.only = FALSE,
                            arg = "display.cap") {
  info <- .rhf_parse_display_cap(cap = cap, arg = arg)
  x.vec <- as.numeric(x)
  out <- x.vec
  base.ok <- is.finite(out)
  if (isTRUE(positive.only)) {
    base.ok <- base.ok & out > 0
  }
  rng <- if (any(is.finite(out))) range(out[is.finite(out)], na.rm = TRUE) else c(NA_real_, NA_real_)
  result <- list(values = out,
                 applied = FALSE,
                 cap = NA_real_,
                 prob = info$prob,
                 label = info$label,
                 symmetric = isTRUE(symmetric),
                 positive.only = isTRUE(positive.only),
                 range = rng)
  if (!isTRUE(info$enabled) || !any(base.ok)) {
    return(result)
  }
  if (isTRUE(symmetric)) {
    cap.value <- as.numeric(stats::quantile(abs(out[base.ok]),
                                            probs = info$prob,
                                            na.rm = TRUE,
                                            names = FALSE))
    if (!is.finite(cap.value) || cap.value <= 0) {
      return(result)
    }
    ok <- is.finite(out)
    out[ok] <- pmax(pmin(out[ok], cap.value), -cap.value)
    rng <- c(-cap.value, cap.value)
  }
  else {
    cap.value <- as.numeric(stats::quantile(out[base.ok],
                                            probs = info$prob,
                                            na.rm = TRUE,
                                            names = FALSE))
    if (!is.finite(cap.value)) {
      return(result)
    }
    ok <- is.finite(out)
    out[ok] <- pmin(out[ok], cap.value)
    rng <- if (any(is.finite(out))) range(out[is.finite(out)], na.rm = TRUE) else c(NA_real_, NA_real_)
  }
  result$values <- out
  result$applied <- any(is.finite(x.vec) & is.finite(out) & (out != x.vec))
  result$cap <- cap.value
  result$range <- rng
  result
}

xvar.wt.rhf <- function(f, d, scale = 4, parallel = TRUE) {
  o.stump <- rhf(f, d, ntree = 1, treesize = 1)
  xvar.names <- o.stump$xvar.names
  vp  <- varpro(Surv(time, event) ~ .,
                convert.standard.counting(f, d),
                parallel = parallel)
  imp <- get.orgvimp(vp, pretty = FALSE)
  weights <- rep(0, length(xvar.names))
  names(weights) <- xvar.names
  weights[names(imp)] <- imp ^ scale
  weights
}
########################################################################
## small internal helpers
########################################################################
.rhf_get_int_haz <- function(o) {
  if (!inherits(o, "rhf")) {
    stop("This function only works for objects inheriting from class 'rhf'.")
  }
  ## Native RHF now supplies the integrated hazard exposure directly.
  ## In grow mode this is int.haz.oob; in prediction / test mode it is
  ## int.haz.test.  The importance code always prioritizes int.haz.oob.
  if (!is.null(o$int.haz.oob)) {
    return(list(values = as.numeric(o$int.haz.oob), source = "int.haz.oob"))
  }
  if (!is.null(o$int.haz.test)) {
    return(list(values = as.numeric(o$int.haz.test), source = "int.haz.test"))
  }
  stop("RHF object is missing integrated hazard exposure values. ",
       "Expected 'int.haz.oob' in grow mode or 'int.haz.test' in test mode.")
}
.rhf_get_working_response <- function(o,
                                      y.external = NULL,
                                      eps = 1e-6) {
  eps <- as.numeric(eps)[1L]
  if (!is.finite(eps) || eps < 0) {
    stop("Argument 'eps' must be a finite nonnegative number.")
  }
  if (!is.null(y.external)) {
    return(list(
      y = as.numeric(y.external),
      y.source = "y.external"
    ))
  }
  int.haz.info <- .rhf_get_int_haz(o)
  if (any(int.haz.info$values < 0, na.rm = TRUE)) {
    stop("Integrated hazard exposure values must be nonnegative before log transformation.")
  }
  list(
    y = log(as.numeric(int.haz.info$values) + eps),
    y.source = int.haz.info$source
  )
}
.rhf_get_start_stop <- function(o) {
  if (!inherits(o, "rhf")) {
    stop("This function only works for objects inheriting from class 'rhf'.")
  }
  if (is.null(o$yvar) || !is.data.frame(o$yvar)) {
    stop("RHF object does not contain a valid 'yvar' data.frame.")
  }
  if (ncol(o$yvar) < 2L) {
    stop("RHF object does not appear to contain counting-process start/stop times.")
  }
  start <- as.numeric(o$yvar[[1L]])
  stopv <- as.numeric(o$yvar[[2L]])
  if (length(start) != length(stopv)) {
    stop("Internal error: start/stop vectors have different lengths.")
  }
  if (any(!is.finite(start)) || any(!is.finite(stopv))) {
    stop("Non-finite start/stop values found in the RHF object.")
  }
  if (any(stopv <= start)) {
    stop("Found non-positive length start/stop intervals in the RHF object.")
  }
  list(start = start, stop = stopv)
}
.rhf_get_grid <- function(o) {
  grid <- as.numeric(o$time.interest)
  if (!length(grid)) {
    stop("RHF object has an empty 'time.interest' grid.")
  }
  if (any(!is.finite(grid))) {
    stop("Non-finite values found in 'time.interest'.")
  }
  if (is.unsorted(grid, strictly = TRUE)) {
    stop("'time.interest' must be strictly increasing.")
  }
  grid
}
.rhf_format_time_names <- function(x) {
  out <- format(as.numeric(x), trim = TRUE, scientific = FALSE)
  bad <- !nzchar(out) | is.na(out)
  if (any(bad)) {
    out[bad] <- as.character(x[bad])
  }
  out
}
.rhf_format_time_labels <- function(x, digits = 2) {
  x <- suppressWarnings(as.numeric(x))
  out <- as.character(x)
  ok <- is.finite(x)
  if (any(ok)) {
    tmp <- formatC(x[ok], format = "f", digits = digits)
    tmp <- sub("\\.?0+$", "", tmp)
    out[ok] <- paste0("t=", tmp)
  }
  out
}
.rhf_format_elapsed <- function(seconds) {
  seconds <- as.numeric(seconds)[1L]
  if (!is.finite(seconds) || seconds < 0) {
    return("NA")
  }
  if (seconds < 60) {
    return(sprintf("%.1fs", seconds))
  }
  minutes <- floor(seconds / 60)
  secs <- round(seconds - 60 * minutes)
  if (minutes < 60) {
    return(sprintf("%dm %02ds", minutes, secs))
  }
  hours <- floor(minutes / 60)
  minutes <- minutes - 60 * hours
  sprintf("%dh %02dm %02ds", hours, minutes, secs)
}
.rhf_validate_cache <- function(o, cache) {
  if (!inherits(cache, "varpro.cache.rhf")) {
    stop("Argument 'cache' must be NULL or an object returned by varpro.cache.rhf().")
  }
  ss <- .rhf_get_start_stop(o)
  grid <- .rhf_get_grid(o)
  if (length(ss$start) != cache$n.pseudo) {
    stop("The supplied cache does not match the RHF object: pseudo-individual count differs.")
  }
  if (length(o$xvar.names) != length(cache$xvar.names) ||
      any(as.character(o$xvar.names) != as.character(cache$xvar.names))) {
    stop("The supplied cache does not match the RHF object: xvar.names differ.")
  }
  if (length(grid) != length(cache$grid) || any(grid != cache$grid)) {
    stop("The supplied cache does not match the RHF object: time.interest differs.")
  }
  invisible(TRUE)
}
.rhf_nonmissing_mean <- function(x) {
  x <- x[is.finite(x)]
  if (!length(x)) {
    NA_real_
  } else {
    mean(x)
  }
}
.rhf_nonmissing_max <- function(x) {
  x <- x[is.finite(x)]
  if (!length(x)) {
    NA_real_
  } else {
    max(x)
  }
}

## tune.treesize: choose treesize to optimize either OOB risk or OOB iAUC.uno
tune.treesize.rhf <- function(formula, data,
                              ntree = 500,
                              nsplit = 10,
                              nodesize = NULL,
                              ## performance measure
                              perf = c("risk", "iAUC"),
                              ## extra arguments passed to auct.rhf when perf = "iAUC"
                              auct.args = NULL,
                              ## search bounds
                              lower = 2L,
                              upper = NULL,
                              C = 3,
                              ## search control
                              method = c("golden", "bisect"),
                              max.evals = 20L,
                              bracket.tol = 2L,   ## stop when width <= bracket.tol
                              ## reproducibility and output
                              seed = NULL,
                              verbose = TRUE,
                              forest = TRUE,
                              ...) {
  method <- match.arg(method)
  perf   <- match.arg(perf)
  ## ---- robust subject count (prefer explicit id, else 'id' column, else rows)
  get_id_from_formula <- function(formula, data) {
    f   <- as.formula(formula)
    lhs <- f[[2]]
    if (is.call(lhs) && identical(as.character(lhs[[1]]), "Surv")) {
      L <- as.list(lhs)[-1]
      if (length(L) == 4L) {                # Surv(id, start, stop, event)
        nm <- paste(deparse(L[[1]]), collapse = "")
        if (nm %in% names(data)) return(nm)
      }
    }
    if ("id" %in% names(data)) return("id")
    NULL
  }
  id.var <- get_id_from_formula(formula, data)
  n.subj <- if (!is.null(id.var)) length(unique(data[[id.var]])) else nrow(data)
  ## ---- bounds
  if (is.null(upper)) upper <- ceiling(C * min(30, n.subj / 5))
  lower <- as.integer(max(2L, lower))
  upper <- as.integer(max(lower + 1L, upper))
  if (!is.null(seed)) set.seed(seed)
  ## ---- caches (keyed by treesize)
  risk.cache <- new.env(parent = emptyenv())
  obj.cache  <- if (isTRUE(forest)) new.env(parent = emptyenv()) else NULL
  ## Only used when perf = "iAUC"
  iauc.cache    <- if (perf == "iAUC") new.env(parent = emptyenv()) else NULL
  iauc.se.cache <- if (perf == "iAUC") new.env(parent = emptyenv()) else NULL
  ## ---- helper to evaluate a single treesize
  eval.one <- function(ts) {
    ts  <- as.integer(ts)
    key <- as.character(ts)
    ## cached result?
    if (exists(key, envir = risk.cache, inherits = FALSE)) {
      r <- get(key, envir = risk.cache, inherits = FALSE)
      if (isTRUE(verbose)) {
        if (perf == "risk") {
          message(sprintf("treesize = %d   OOB risk (criterion) = %.6f (cached)", ts, r))
        } else {
          message(sprintf("treesize = %d   1 - iAUC.uno (criterion) = %.6f (cached)", ts, r))
        }
      }
      return(r)
    }
    ## refit RHF at this treesize
    if (!is.null(seed)) set.seed(seed)  ## make cross-size comparisons comparable
    fit <- rhf(formula = formula,
               data    = data,
               ntree   = ntree,
               nsplit  = nsplit,
               treesize = ts,
               nodesize = nodesize,
               ...)
    ## ---- performance measure
    if (perf == "risk") {
      rsk <- fit$risk.oob
      if (is.null(rsk))
        stop("OOB risk values (fit$risk.oob) are missing; cannot tune by 'risk'.")
      r <- mean(rsk[is.finite(rsk)], na.rm = TRUE)
      if (!is.finite(r)) r <- Inf
      if (isTRUE(verbose))
        message(sprintf("treesize = %d   OOB risk (criterion) = %.6f", ts, r))
    } else { ## perf == "iAUC"
      args.auct <- auct.args
      if (is.null(args.auct)) args.auct <- list()
      args.auct$object <- fit  ## enforce
      auct.obj <- try(do.call(auct.rhf, args.auct), silent = TRUE)
      if (inherits(auct.obj, "try-error")) {
        if (isTRUE(verbose)) {
          message(sprintf("treesize = %d   auct.rhf() failed; treating criterion as +Inf", ts))
        }
        r <- Inf
        if (!is.null(iauc.cache))    assign(key, NA_real_, envir = iauc.cache)
        if (!is.null(iauc.se.cache)) assign(key, NA_real_, envir = iauc.se.cache)
      } else {
        iauc <- auct.obj$iAUC.uno
        if (!is.finite(iauc)) {
          if (isTRUE(verbose)) {
            message(sprintf("treesize = %d   iAUC.uno is non-finite; treating criterion as +Inf", ts))
          }
          r <- Inf
          se <- NA_real_
        } else {
          r <- 1 - iauc
          if (isTRUE(verbose)) {
            message(sprintf("treesize = %d   iAUC.uno = %.6f   criterion (1 - iAUC.uno) = %.6f",
                            ts, iauc, r))
          }
          ## try to grab bootstrap SE if available
          se <- NA_real_
          if (!is.null(auct.obj$boot) &&
              !is.null(auct.obj$boot$iAUC.uno.se) &&
              is.finite(auct.obj$boot$iAUC.uno.se)) {
            se <- auct.obj$boot$iAUC.uno.se
          }
        }
        if (!is.null(iauc.cache))    assign(key, iauc, envir = iauc.cache)
        if (!is.null(iauc.se.cache)) assign(key, se,   envir = iauc.se.cache)
      }
    }
    ## cache results
    assign(key, r, envir = risk.cache)
    if (isTRUE(forest)) assign(key, fit, envir = obj.cache)
    r
  }
  ## collect the path
  collect.path <- function() {
    sizes <- as.integer(sort(as.integer(ls(risk.cache))))
    crit  <- vapply(
      sizes,
      function(k) get(as.character(k), envir = risk.cache, inherits = FALSE),
      numeric(1)
    )
    path <- data.frame(treesize = sizes, risk = crit)
    if (perf == "iAUC" && !is.null(iauc.cache)) {
      iauc <- vapply(
        sizes,
        function(k) {
          key <- as.character(k)
          if (exists(key, envir = iauc.cache, inherits = FALSE)) {
            get(key, envir = iauc.cache, inherits = FALSE)
          } else {
            NA_real_
          }
        },
        numeric(1)
      )
      se <- vapply(
        sizes,
        function(k) {
          key <- as.character(k)
          if (!is.null(iauc.se.cache) &&
              exists(key, envir = iauc.se.cache, inherits = FALSE)) {
            get(key, envir = iauc.se.cache, inherits = FALSE)
          } else {
            NA_real_
          }
        },
        numeric(1)
      )
      path$iAUC    <- iauc
      path$iAUC.se <- se
    }
    path
  }
  ## ---- initialize
  a <- lower; b <- upper
  evals <- 0L
  ## Always evaluate the smallest treesize at least once
  eval.one(lower); evals <- evals + 1L
  if (method == "golden") {
    ## classic discrete golden-section on integers
    phi <- (1 + sqrt(5)) / 2
    x1 <- as.integer(round(b - (b - a) / phi))
    x2 <- as.integer(round(a + (b - a) / phi))
    if (x1 == x2) x2 <- min(b, x1 + 1L)
    f1 <- eval.one(x1); evals <- evals + 1L
    f2 <- eval.one(x2); evals <- evals + 1L
    while ((b - a > bracket.tol) && evals < max.evals) {
      if (f1 > f2) {
        a <- x1
        x1 <- x2
        f1 <- f2
        x2 <- as.integer(round(a + (b - a) / phi))
        if (x2 <= x1) x2 <- min(b, x1 + 1L)
        f2 <- eval.one(x2); evals <- evals + 1L
      } else {
        b <- x2
        x2 <- x1
        f2 <- f1
        x1 <- as.integer(round(b - (b - a) / phi))
        if (x1 >= x2) x1 <- max(a, x2 - 1L)
        f1 <- eval.one(x1); evals <- evals + 1L
      }
    }
  } else { ## method == "bisect"
    while ((b - a > bracket.tol) && evals < max.evals) {
      m  <- as.integer(floor((a + b) / 2))
      fm <- eval.one(m);       evals <- evals + 1L
      fl <- eval.one(m - 1L);  evals <- evals + 1L
      fr <- eval.one(m + 1L);  evals <- evals + 1L
      if (fl >= fm && fr >= fm) { a <- max(a, m - 1L); b <- min(b, m + 1L); break }
      if (fl <  fm) { b <- m - 1L } else
      if (fr <  fm) { a <- m + 1L } else {
        a <- max(a, m - 1L); b <- min(b, m + 1L)
      }
    }
  }
  ## Evaluate remaining integers in [a, b], but only those not already cached
  left  <- max(lower, a)
  right <- min(upper, b)
  if (right >= left) {
    cand <- seq.int(left, right)
    not.eval <- cand[!(as.character(cand) %in% ls(risk.cache))]
    if (length(not.eval)) {
      vapply(not.eval, eval.one, numeric(1))
      evals <- evals + length(not.eval)
    }
  }
  ## ---- GLOBAL best over *all* evaluated sizes
  path <- collect.path()
  idx  <- which.min(path$risk)
  best.size <- path$treesize[idx]
  best.err  <- path$risk[idx]
  ## Optionally return the forest at best.size
  best.fit <- NULL
  if (isTRUE(forest)) {
    key <- as.character(best.size)
    if (!exists(key, envir = obj.cache, inherits = FALSE)) {
      if (!is.null(seed)) set.seed(seed)
      best.fit <- rhf(formula = formula, data = data,
                      ntree = ntree, nsplit = nsplit,
                      treesize = best.size, nodesize = nodesize, ...)
    } else {
      best.fit <- get(key, envir = obj.cache, inherits = FALSE)
    }
  }
  out <- list(
    best.size   = best.size,
    best.err    = best.err,
    bounds      = c(lower = lower, upper = upper),
    n.subjects  = n.subj,
    C           = C,
    method      = method,
    perf        = perf,
    path        = path
  )
  if (!is.null(best.fit)) out$forest <- best.fit
  class(out) <- "tune.treesize.rhf"
  out
}
tune.rhf <- tune.treesize.rhf
## Convenience wrapper: tune treesize by iAUC.uno from auct.rhf
tune.iAUC.rhf <- function(formula, data, auct.args = NULL, ...) {
  tune.treesize.rhf(
    formula  = formula,
    data     = data,
    perf     = "iAUC",
    auct.args = auct.args,
    ...
  )
}
tune.iAUC <- tune.iAUC.rhf
## plot results (metric vs treesize, also has bootstrap now)
plot.tune.treesize.rhf <- function(x,
                                   ylab   = NULL,
                                   main   = NULL,
                                   se.band = TRUE,
                                   se.mult = 1,
                                   ylim   = NULL,
                                   ...) {
  stopifnot(inherits(x, "tune.treesize.rhf"))
  path <- x$path
  perf <- if (!is.null(x$perf)) x$perf else "risk"
  if (perf == "iAUC" && "iAUC" %in% names(path)) {
    ## ---------- iAUC tuning plot ----------
    xx <- path$treesize
    yy <- path$iAUC
    if (is.null(ylab)) ylab <- "OOB iAUC.uno"
    if (is.null(main)) main <- "Tuning treesize by OOB iAUC"
    ## compute ylim to include band if needed
    if (is.null(ylim)) {
      y.all <- yy
      if (isTRUE(se.band) && "iAUC.se" %in% names(path)) {
        se   <- path$iAUC.se
        mult <- if (is.finite(se.mult) && se.mult > 0) se.mult else 1
        yl   <- yy - mult * se
        yu   <- yy + mult * se
        y.all <- c(y.all, yl, yu)
      }
      y.all <- y.all[is.finite(y.all)]
      if (length(y.all)) {
        ymin <- min(y.all)
        ymax <- max(y.all)
        pad  <- 0.05 * (ymax - ymin)
        ylim <- c(ymin - pad, ymax + pad)
      }
    }
    plot(xx, yy, type = "p", pch = 16,
         xlab = "treesize", ylab = ylab, main = main,
         ylim = ylim, ...)
    ok <- is.finite(xx) & is.finite(yy)
    if (sum(ok) >= 3L) {
      sm <- stats::lowess(xx[ok], yy[ok])
      lines(sm, lwd = 2)
    }
    ## gray SE ribbon, if available
    if (isTRUE(se.band) &&
        "iAUC.se" %in% names(path) &&
        any(is.finite(path$iAUC.se))) {
      se   <- path$iAUC.se
      mult <- if (is.finite(se.mult) && se.mult > 0) se.mult else 1
      yl   <- yy - mult * se
      yu   <- yy + mult * se
      okb <- is.finite(xx) & is.finite(yl) & is.finite(yu)
      if (sum(okb) >= 2L) {
        ord  <- order(xx[okb])
        xt   <- xx[okb][ord]
        yl.t <- yl[okb][ord]
        yu.t <- yu[okb][ord]
        polygon(c(xt, rev(xt)),
                c(yl.t, rev(yu.t)),
                border = NA,
                col = grDevices::adjustcolor("gray", alpha.f = 0.25))
        ## redraw points on top
        points(xx, yy, pch = 16)
      }
    }
    ## vertical line + annotation at best treesize
    abline(v = x$best.size, lty = 2)
    idx <- which(path$treesize == x$best.size)
    best.iAUC <- if (length(idx) >= 1L && is.finite(path$iAUC[idx[1L]])) {
      path$iAUC[idx[1L]]
    } else {
      1 - x$best.err
    }
    mtext(sprintf("best treesize = %d, iAUC = %.4f",
                  x$best.size, best.iAUC),
          line = 0.5)
  } else {
    ## ---------- OOB risk plot (no band) ----------
    xx <- path$treesize
    yy <- path$risk
    if (is.null(ylab)) ylab <- "OOB empirical risk"
    if (is.null(main)) main <- "Tuning treesize by OOB risk"
    if (is.null(ylim)) {
      y.all <- yy[is.finite(yy)]
      if (length(y.all)) {
        ymin <- min(y.all); ymax <- max(y.all)
        pad  <- 0.05 * (ymax - ymin)
        ylim <- c(ymin - pad, ymax + pad)
      }
    }
    plot(xx, yy, type = "p", pch = 16,
         xlab = "treesize", ylab = ylab, main = main,
         ylim = ylim, ...)
    ok <- is.finite(xx) & is.finite(yy)
    if (sum(ok) >= 3L) {
      sm <- stats::lowess(xx[ok], yy[ok])
      lines(sm, lwd = 2)
    }
    abline(v = x$best.size, lty = 2)
    mtext(sprintf("best treesize = %d, risk = %.4f",
                  x$best.size, x$best.err),
          line = 0.5)
  }
}

########################################################################
## print and coercion helpers
########################################################################
print.importance.rhf <- function(x,
                                 top = 10L,
                                 rank.by = c("q90", "median", "mean", "max"),
                                 digits = 4L,
                                 scientific.threshold = 1e4,
                                 ...) {
  rank.by <- match.arg(rank.by)
  top <- as.integer(top)[1L]
  if (!is.finite(top) || top < 1L) {
    stop("'top' must be a positive integer.")
  }
  header <- c(
    "",
    "RHF time-localized VarPro importance",
    paste0("  windows:   ", ncol(x$importance.matrix)),
    paste0("  variables: ", nrow(x$importance.matrix)),
    paste0("  ranking:   ", rank.by)
  )
  if (!is.null(x$y.source) && nzchar(x$y.source)) {
    header <- c(header, paste0("  y.source:  ", x$y.source))
  }
  header <- c(
    header,
    "  matrix:    imp.t$importance.matrix (variables x time)",
    "  long:      imp.t$importance.long",
    ""
  )
  message(paste(header, collapse = "\n"))
  score.q90 <- .rhf_row_summary(x$importance.matrix, rank.by = "q90")
  score.median <- .rhf_row_summary(x$importance.matrix, rank.by = "median")
  score.mean <- .rhf_row_summary(x$importance.matrix, rank.by = "mean")
  score.max <- .rhf_row_summary(x$importance.matrix, rank.by = "max")
  score.rank <- switch(rank.by,
                       q90 = score.q90,
                       median = score.median,
                       mean = score.mean,
                       max = score.max)
  ord <- order(score.rank, decreasing = TRUE, na.last = TRUE)
  top.show <- head(ord, min(top, length(ord)))
  if (length(top.show)) {
    tab <- data.frame(
      variable = rownames(x$importance.matrix)[top.show],
      q90.importance = .rhf_format_importance_numbers(score.q90[top.show],
                                                      digits = digits,
                                                      scientific.threshold = scientific.threshold),
      median.importance = .rhf_format_importance_numbers(score.median[top.show],
                                                         digits = digits,
                                                         scientific.threshold = scientific.threshold),
      mean.importance = .rhf_format_importance_numbers(score.mean[top.show],
                                                       digits = digits,
                                                       scientific.threshold = scientific.threshold),
      max.importance = .rhf_format_importance_numbers(score.max[top.show],
                                                      digits = digits,
                                                      scientific.threshold = scientific.threshold),
      stringsAsFactors = FALSE,
      row.names = NULL,
      check.names = FALSE
    )
    print(tab, row.names = FALSE, right = TRUE)
  }
  invisible(x)
}
as.data.frame.importance.rhf <- function(x,
                                         row.names = NULL,
                                         optional = FALSE,
                                         format = c("long", "variable_by_time", "time_by_variable"),
                                         ...) {
  format <- match.arg(format)
  if (format == "long") {
    return(x$importance.long)
  }
  if (format == "variable_by_time") {
    out <- as.data.frame(x$importance.matrix,
                         stringsAsFactors = FALSE,
                         check.names = FALSE)
    out$variable <- rownames(x$importance.matrix)
    out <- out[, c("variable", colnames(x$importance.matrix)), drop = FALSE]
    rownames(out) <- NULL
    return(out)
  }
  out <- as.data.frame(t(x$importance.matrix),
                       stringsAsFactors = FALSE,
                       check.names = FALSE)
  out$time <- x$window.info$time
  out$window <- x$window.info$label
  out <- out[, c("time", "window", rownames(x$importance.matrix)), drop = FALSE]
  rownames(out) <- NULL
  out
}

########################################################################
## cache builder
########################################################################
varpro.cache.rhf <- function(o,
                             max.rules.tree = 150L,
                             max.tree = 150L,
                             y.external = NULL,
                             eps = 1e-6,
                             verbose = FALSE) {
  if (!inherits(o, "rhf")) {
    stop("This function only works for objects inheriting from class 'rhf'.")
  }
  t.start <- proc.time()[3L]
  if (isTRUE(verbose)) {
    message("[varpro.cache.rhf] building rule-membership cache")
  }
  varpro.strength.fun <- varpro.strength
  get.strengthArray.fun <- get.varpro.strengthArray
  oo <- varpro.strength.fun(object = o,
                            max.rules.tree = max.rules.tree,
                            max.tree = max.tree,
                            membership = TRUE)
  if (is.null(oo$strengthArray)) {
    stop("varpro.strength() did not return a valid 'strengthArray'.")
  }
  if (is.null(oo$oobMembership) || is.null(oo$compMembership)) {
    stop("varpro.strength() did not return membership lists.")
  }
  y.info <- .rhf_get_working_response(o,
                                      y.external = y.external,
                                      eps = eps)
  y <- y.info$y
  y.source <- y.info$y.source
  ss <- .rhf_get_start_stop(o)
  grid <- .rhf_get_grid(o)
  n.pseudo <- length(ss$start)
  if (length(y) != n.pseudo) {
    stop("Length of working response does not match the number of pseudo-individuals.")
  }
  attr(y, "family") <- "regr"
  attr(y, "y.org") <- if (!is.null(y.external)) y.external else if (!is.null(o$y.org)) o$y.org else y
  ## Mirror the RHF branch in importance.varpro(): add an importance slot and
  ## build the regression-style strength array used downstream.
  oo$strengthArray$importance <- NA_real_
  results.template <- get.strengthArray.fun(oo$strengthArray, "regr", y)
  imp.col <- colnames(results.template) == "imp"
  if (sum(imp.col) != 1L) {
    imp.col <- grepl("^imp($|[.])", colnames(results.template))
  }
  if (sum(imp.col) != 1L) {
    stop("RHF time-localized importance expects a single regression-style 'imp' column.")
  }
  if (!("n.oob" %in% colnames(results.template))) {
    stop("Expected column 'n.oob' in the regression-style strength array.")
  }
  if (!all(c("tree", "variable") %in% colnames(results.template))) {
    stop("Expected columns 'tree' and 'variable' in the regression-style strength array.")
  }
  base.rules <- which(oo$strengthArray$oobCT > 0 & oo$strengthArray$compCT > 0)
  results.base <- results.template[base.rules, , drop = FALSE]
  results.base[, imp.col] <- NA_real_
  plan <- .rhf_build_window_plan(start = ss$start,
                                 stopv = ss$stop,
                                 grid = grid)
  precomp <- .rhf_precompute_rule_windows(
    strength = oo$strengthArray,
    base.rules = base.rules,
    oobMembership = oo$oobMembership,
    compMembership = oo$compMembership,
    y = y,
    left.idx = plan$left.idx,
    right.idx = plan$right.idx,
    K = plan$K,
    verbose = verbose
  )
  out <- list(
    call = match.call(),
    results.base = results.base,
    imp.col = imp.col,
    local.imp.window = precomp$local.imp,
    keep.window = precomp$keep,
    n.oob.window = precomp$n.oob,
    n.rules.window = precomp$n.rules,
    xvar.names = o$xvar.names,
    y = y,
    y.source = y.source,
    start = ss$start,
    stop = ss$stop,
    grid = grid,
    K = plan$K,
    window.left = plan$window.left,
    window.right = plan$window.right,
    window.mid = plan$window.mid,
    window.label = plan$window.label,
    n.risk.window = plan$n.risk,
    max.rules.tree = max.rules.tree,
    max.tree = max.tree,
    n.pseudo = n.pseudo
  )
  class(out) <- "varpro.cache.rhf"
  if (isTRUE(verbose)) {
    message("[varpro.cache.rhf] done: ",
            nrow(results.base), " usable rules, ",
            n.pseudo, " pseudo-individuals",
            if (!is.null(y.source) && nzchar(y.source)) paste0(", y.source=", y.source) else "",
            ", elapsed ",
            .rhf_format_elapsed(proc.time()[3L] - t.start))
  }
  out
}
varpro.cache <- varpro.cache.rhf
########################################################################
## localized rule calculations
########################################################################
.rhf_local_results_from_cache <- function(cache,
                                          window.index,
                                          trim,
                                          sort,
                                          ...) {
  workhorse.fun <- function(...) {
    args <- list(...)
    ## hard-coded for compatibility
    args$cutoff <- 0.76
    args$plot.it <- FALSE
    args$conf <- FALSE
    args$ylab <- "Importance"
    do.call(importance.varpro.workhorse, args)
  }
  take <- which(cache$keep.window[, window.index])
  if (!length(take)) {
    return(list(table = data.frame(), n.rules = 0L))
  }
  results <- cache$results.base[take, , drop = FALSE]
  results[, cache$imp.col] <- cache$local.imp.window[take, window.index]
  results$n.oob <- cache$n.oob.window[take, window.index]
  o.tmp <- list(
    results = results,
    xvar.names = cache$xvar.names,
    family = "regr"
  )
  tab <- workhorse.fun(o = o.tmp,
                       trim = trim,
                       sort = sort,
                       local.std = TRUE,
                       ...)
  list(table = tab, n.rules = length(take))
}
.rhf_importance_long <- function(importance.matrix, window.info) {
  if (!is.matrix(importance.matrix) || !length(importance.matrix)) {
    return(data.frame())
  }
  if (ncol(importance.matrix) != nrow(window.info)) {
    stop("Internal error: window metadata do not align with the importance matrix.")
  }
  imp.vec <- as.vector(importance.matrix)
  out <- data.frame(
    variable = rep(rownames(importance.matrix), times = ncol(importance.matrix)),
    time = rep(window.info$time, each = nrow(importance.matrix)),
    time.index = rep(window.info$index, each = nrow(importance.matrix)),
    window = rep(window.info$label, each = nrow(importance.matrix)),
    start = rep(window.info$start, each = nrow(importance.matrix)),
    stop = rep(window.info$stop, each = nrow(importance.matrix)),
    midpoint = rep(window.info$midpoint, each = nrow(importance.matrix)),
    n.risk = rep(window.info$n.risk, each = nrow(importance.matrix)),
    n.rules = rep(window.info$n.rules, each = nrow(importance.matrix)),
    importance = imp.vec,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
  ord.imp <- out$importance
  ord.imp[is.na(ord.imp)] <- -Inf
  ord <- order(out$time.index, -ord.imp, out$variable)
  out[ord, , drop = FALSE]
}
########################################################################
## main user routine
########################################################################
importance.rhf <- function(o,
                           cache = NULL,
                           time.index = NULL,
                           trim = 0.1,
                           sort = TRUE,
                           max.rules.tree,
                           max.tree,
                           eps = 1e-6,
                           y.external = NULL,
                           verbose = FALSE,
                           ...) {
  if (!inherits(o, "rhf")) {
    stop("This function only works for objects inheriting from class 'rhf'.")
  }
  if (is.null(cache)) {
    cache.args <- list(o = o,
                       eps = eps,
                       y.external = y.external,
                       verbose = verbose)
    if (!missing(max.rules.tree)) {
      cache.args$max.rules.tree <- max.rules.tree
    }
    if (!missing(max.tree)) {
      cache.args$max.tree <- max.tree
    }
    cache <- do.call(varpro.cache.rhf, cache.args)
  }
  else {
    .rhf_validate_cache(o, cache)
    if (isTRUE(verbose)) {
      message("[importance.rhf] using supplied cache",
              if (!is.null(cache$y.source) && nzchar(cache$y.source)) {
                paste0(" (y.source=", cache$y.source, ")")
              } else "")
    }
  }
  if (is.null(time.index)) {
    time.index <- seq_len(cache$K)
  }
  if (is.logical(time.index)) {
    if (length(time.index) != cache$K) {
      stop("Logical 'time.index' must have length equal to length(o$time.interest).")
    }
    time.index <- which(time.index)
  }
  time.index <- sort(unique(as.integer(time.index)))
  if (!length(time.index)) {
    stop("No valid time indices supplied.")
  }
  if (any(!is.finite(time.index)) || any(time.index < 1L) || any(time.index > cache$K)) {
    stop("Argument 'time.index' contains invalid window indices.")
  }
  p <- length(cache$xvar.names)
  m <- length(time.index)
  importance.matrix <- matrix(
    NA_real_,
    nrow = p,
    ncol = m,
    dimnames = list(
      cache$xvar.names,
      .rhf_format_time_names(cache$window.right[time.index])
    )
  )
  n.risk <- as.integer(cache$n.risk.window[time.index])
  n.rules <- as.integer(cache$n.rules.window[time.index])
  t.start <- proc.time()[3L]
  for (jj in seq_len(m)) {
    target.window <- time.index[jj]
    local.out <- .rhf_local_results_from_cache(
      cache = cache,
      window.index = target.window,
      trim = trim,
      sort = sort,
      ...
    )
    if (is.data.frame(local.out$table) && nrow(local.out$table)) {
      importance.matrix[rownames(local.out$table), jj] <- local.out$table$z
    }
    if (isTRUE(verbose)) {
      elapsed <- proc.time()[3L] - t.start
      eta <- if (jj < m) elapsed / jj * (m - jj) else 0
      message(
        "[importance.rhf] window ", jj, "/", m,
        " (grid index ", target.window,
        ", t=", .rhf_format_time_names(cache$window.right[target.window]), ")",
        " | remaining ", m - jj,
        " | n.risk ", n.risk[jj],
        " | n.rules ", n.rules[jj],
        " | elapsed ", .rhf_format_elapsed(elapsed),
        " | ETA ", .rhf_format_elapsed(eta)
      )
    }
  }
  window.info <- data.frame(
    index = time.index,
    time = cache$window.right[time.index],
    start = cache$window.left[time.index],
    stop = cache$window.right[time.index],
    midpoint = cache$window.mid[time.index],
    n.risk = n.risk,
    n.rules = n.rules,
    label = cache$window.label[time.index],
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
  out <- list(
    call = match.call(),
    xvar.names = cache$xvar.names,
    importance.matrix = importance.matrix,
    importance.long = .rhf_importance_long(
      importance.matrix = importance.matrix,
      window.info = window.info
    ),
    window.info = window.info,
    y.source = cache$y.source,
    trim = trim
  )
  class(out) <- "importance.rhf"
  out
}

\name{print.rhf}
\alias{print.rhf}
\alias{print.vimp.rhf}
\title{Print Output from a Random Hazard Forest Analysis}
\description{
  Print a compact summary of a Random Hazard Forest analysis. This is the
  default print method for objects produced by \code{grow} or \code{predict}.
}
\usage{
\method{print}{rhf}(x, digits = 3, label.width = 34, ...)
}
\arguments{
  \item{x}{An object of class \code{rhf} created by \code{grow} or \code{predict}.
           When \code{predict} is called without new data (``restore mode''), the
           original forest and its OOB risk are returned.}
  \item{digits}{Number of digits used for printed numeric summaries.}
  \item{label.width}{Width used to right-justify labels in the printed output.}
  \item{...}{Additional arguments (currently unused).}
}
\details{
  The printed summary includes basic sample information, forest size
  characteristics, event counts when available, and risk metrics
  (in-bag, OOB, or test risk depending on how the object was created).
}
\value{
  Returns the object \code{x} invisibly.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\keyword{print}

\name{predict.rhf}
\alias{predict.rhf}
\title{Prediction on Test Data for Random Hazard Forests}

\description{
  Obtain predicted values on test data using a trained random hazard
  forests.
}

\usage{\method{predict}{rhf}(object, newdata,  get.tree = NULL,
     block.size = 10, membership = TRUE, seed = NULL, do.trace = FALSE,...)}

\arguments{

  \item{object}{An \code{rhf} object returned from a previous training call to \code{rhf}.}

  \item{newdata}{Test data frame. If omitted, the original training data is used and the full training forest is restored.}

  \item{get.tree}{Optional vector of integer indices specifying which trees to use for ensemble predictions. Defaults to using all trees in the forest.}

  \item{block.size}{Controls how cumulative error rate is reported. To obtain cumulative error every \code{n} trees, set this to an integer between 1 and \code{ntree}.}

  \item{membership}{Logical flag indicating whether terminal node membership and inbag information should be returned.}

  \item{seed}{Negative integer specifying the random seed for reproducibility.}

  \item{do.trace}{Number of seconds between progress updates printed to the console.}

  \item{...}{Additional optional arguments passed to internal methods.}

}

\details{

  Returns the predicted values for a random hazard forests.
  
}

\value{
An object of class \code{c("rhf", "predict", family)}. The returned list
contains the fitted forest together with prediction summaries on the
evaluation grid \code{time.interest}. Important components include:

\itemize{
  \item \code{hazard.test}, \code{chf.test}, \code{risk.test}, and
    \code{int.haz.test}: test-set hazard, cumulative hazard, risk, and
    integrated-hazard summaries when \code{newdata} is supplied.
  \item \code{hazard.oob}, \code{chf.oob}, \code{risk.oob}, and
    \code{int.haz.oob}: out-of-bag summaries for the training data.
  \item \code{hazard.inbag}, \code{chf.inbag}, \code{risk.inbag}, and
    \code{int.haz.inbag}: in-bag summaries when available.
  \item \code{id}, \code{yvar}, and \code{xvar}: identifiers and processed
    outcome/predictor data used by the returned prediction object.
  \item \code{pseudo.membership} and \code{inbag}: terminal-node membership
    and inbag information when \code{membership = TRUE}.
  \item \code{forest}: the fitted forest object used to generate the
    predictions.
}

If \code{newdata} is omitted, the function restores predictions for the
original training data using the stored forest and returns the same class of
object.
}

\author{

  Hemant Ishwaran and Udaya B. Kogalur

}

\references{

  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
  
  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.

  Lee, D.K. and Chen N. and Ishwaran H (2021). Boosted nonparametric
  hazards with time-dependent covariates. \emph{Annals of Statistics},
  49: 2101-2128.



}

\seealso{
  \command{\link{rhf}}
}

\examples{

## ------------------------------------------------------------
## canonical train/test example (synthetic data)
## ------------------------------------------------------------

simID <- 1
trn <- hazard.simulation(simID)$dta
tst <- hazard.simulation(simID)$dta
f <- "Surv(id, start, stop, event) ~ ."

## training
o <- rhf(f, trn, ntree = 3)
print(o)

## testing
p <- predict(o, tst)
print(p)

\donttest{

## ------------------------------------------------------------
##  pbc: train/test example
## ------------------------------------------------------------

library("randomForestSRC")
data(pbc, package = "randomForestSRC")
pbc.raw <- na.omit(pbc)
trn <- sample(1:nrow(pbc.raw), size=nrow(pbc.raw) * .75, replace = FALSE)

d.trn <- convert.counting(Surv(days, status) ~ ., pbc.raw[trn,])
d.tst <- convert.counting(Surv(days, status) ~ ., pbc.raw[-trn,])
f <- "Surv(id, start, stop, event) ~ ."

## train/predict
o <- rhf(f, d.trn)
print(predict(o, d.tst))

## restore the forest
print(predict(o))

}
}

\keyword{predict rhf}

\name{internal.utils}
\alias{convert.counting}
\alias{convert.standard.counting}
\alias{hazard.simulation}
\alias{xvar.wt.rhf}

\title{Advanced Utility Functions for RHF Workflows}

\description{
These helper functions support advanced random-hazard-forest workflows,
including data conversion, simulation, and predictor weighting.
}

\usage{
convert.counting(f, dta, scale = FALSE)

convert.standard.counting(formula, data,
  scale = FALSE,
  rescale.from.attr = FALSE,
  keep.id = FALSE,
  keep.row_index = FALSE,
  sorted = FALSE,
  id.default = "id",
  eps = 1e-8,
  landmark.time = NULL,
  landmark.use.tminus = TRUE,
  return.type = c("survival", "x"),
  keep.landmark.cols = FALSE)

hazard.simulation(type = 1,
  n = 500, p = 10, nrecords = 7,
  scale = FALSE, ngrid = 1e5, ...)

xvar.wt.rhf(f, d, scale = 4, parallel = TRUE)
}

\arguments{
  \item{f, formula}{A survival formula. \code{convert.counting()} expects a
  standard survival formula such as \code{Surv(time, event) ~ .}. 
  \code{convert.standard.counting()} expects counting-process style input such
  as \code{Surv(id, start, stop, event) ~ .} or
  \code{Surv(start, stop, event) ~ .}.}

  \item{dta, data, d}{Input data frame.}

  \item{scale}{Logical or numeric time-scaling control, depending on the
  helper being used.}

  \item{rescale.from.attr}{Logical; if \code{TRUE},
  \code{convert.standard.counting()} rescales times using attributes stored on
  the counting-process data.}

  \item{keep.id}{Logical; if \code{TRUE}, retain the subject identifier in the
  returned data frame from \code{convert.standard.counting()}.}

  \item{keep.row_index}{Logical; if \code{TRUE}, include the source row index
  in the output of \code{convert.standard.counting()}.}

  \item{sorted}{Logical; indicates whether the counting-process input is
  already sorted by subject and time.}

  \item{id.default}{Default subject identifier column name used when
  \code{formula} omits an explicit \code{id} term.}

  \item{eps}{Small positive tolerance used in internal time comparisons.}

  \item{landmark.time}{Optional landmark time passed to
  \code{convert.standard.counting()}.}

  \item{landmark.use.tminus}{Logical; controls how landmark covariates are
  selected when \code{landmark.time} is supplied.}

  \item{return.type}{Character string specifying whether
  \code{convert.standard.counting()} returns a standard survival outcome plus
  covariates (\code{"survival"}) or only the covariate frame (\code{"x"}).}

  \item{keep.landmark.cols}{Logical; if \code{TRUE}, keep auxiliary landmark
  columns in the output of \code{convert.standard.counting()}.}

  \item{type}{Simulation scenario. May be given numerically
  (\code{1}, \code{2}, \code{3}) or by name.}

  \item{n, p, nrecords, ngrid}{Simulation controls used by
  \code{hazard.simulation()}.}

  \item{...}{Additional arguments passed to the selected simulation engine.}

  \item{parallel}{Logical; if \code{TRUE}, allow parallel work in
  \code{xvar.wt.rhf()}.}
}

\details{
\code{convert.counting()} converts standard right-censored survival data to the
counting-process representation expected by \code{rhf()}.

\code{convert.standard.counting()} converts counting-process style RHF data to
a conventional survival-analysis data frame or to a predictor-only frame.

\code{hazard.simulation()} generates synthetic counting-process data together
with the corresponding theoretical hazard function.

\code{xvar.wt.rhf()} computes variable weights that can be supplied to
\code{rhf(..., xvar.wt = ...)}.
}

\value{
The returned value depends on the helper:

\itemize{
  \item \code{convert.counting()}: a data frame with columns
  \code{id}, \code{start}, \code{stop}, \code{event}, followed by the predictor
  columns from the input data.
  \item \code{convert.standard.counting()}: a data frame in the format requested
  by \code{return.type}. With \code{return.type = "survival"}, the leading
  columns are \code{time} and \code{event}; additional covariate and optional
  helper columns may also be present. The function may return an empty data
  frame when no valid rows remain after filtering.
  \item \code{hazard.simulation()}: a list with components \code{dta}
  (simulated counting-process data frame), \code{haz}
  (a function that evaluates the true hazard on a supplied time grid), and
  \code{scale} (the time-scaling factor used in the simulation).
  \item \code{xvar.wt.rhf()}: a named numeric vector of predictor weights for
  the \code{xvar.wt} argument of \code{rhf()}.
}
}

\seealso{
  \code{\link{rhf}},
  \code{\link{predict.rhf}}
}

\keyword{internal}

\name{auct.rhf}
\alias{auct.rhf}
\alias{auct}
\alias{plot.auct.rhf}
\alias{print.auct.rhf}
\title{Time-varying AUC(t) and iAUC for Random Hazard Forests (TDC)}

\description{
Compute the time-varying AUC(t) and its time-aggregated
summaries for Random Hazard Forests (RHF) with time-dependent covariates
(TDC).
}

\usage{
auct.rhf(
  object,
  marker = c("cumhaz", "hazard", "chf", "haz"),
  method = c("cumulative", "incident"),
  tau = NULL,
  riskset = c("subject", "record"),
  min.controls = 25,
  nfrac.controls = 0.10,
  min.cases = 1,
  g.floor = 0.10,
  g.floor.q = NULL,
  power = 2,
  ydata = NULL,
  winsor.q = NULL,
  eps = 1e-12,
  bootstrap.rep = 0L,
  bootstrap.refit = FALSE,
  bootstrap.conf = 0.95,
  bootstrap.seed = NULL,
  verbose = TRUE
)

\method{print}{auct.rhf}(x, digits = 4, max.rows = 8, \dots)
\method{plot}{auct.rhf}(x, bass = 10, xlab = "Time", ylab = NULL,
         main = NULL, ylim = NULL, pch = 16, alpha = .05, \dots)
}

\arguments{

  \item{object}{An RHF object. For grow/restore objects (class includes \code{"rhf"} and \code{"grow"}),
  OOB prediction matrices are used. For predict objects (class includes \code{"rhf"} and \code{"predict"}),
  test data is used (supply \code{ydata=} if the object does not include \code{id}/\code{yvar}).}

  \item{marker}{Which time-varying score to evaluate: \code{"cumhaz"} (alias \code{"chf"})
    or \code{"hazard"} (alias \code{"haz"}).}

  \item{method}{Target estimand. \code{"incident"} (incident/dynamic) uses cases who fail at time \eqn{t}
  and controls who are at risk at \eqn{t}; this targets instantaneous hazard discrimination.
  \code{"cumulative"} (cumulative/dynamic) uses cases who have failed by \eqn{t} and controls who are event-free at \eqn{t};
  this matches the more commonly used time-dependent AUC (Heagerty and Zheng, 2005).}

  \item{tau}{Optional time horizon. Evaluation is restricted to times \code{t <= tau}.}

  \item{riskset}{Risk-set definition for \code{method = "incident"}:
    \code{"subject"} treats a subject as at risk when \code{entry < t <= stop};
    \code{"record"} uses counting-process rows. Ignored when \code{method = "cumulative"}.}

  \item{min.controls}{Minimum number of controls required at each evaluated time.}

  \item{nfrac.controls}{Minimum fraction of the total sample to be available as controls at each evaluated time.}

  \item{min.cases}{Minimum number of cases required at each evaluated time.}

  \item{g.floor}{Lower bound used to stabilize inverse-probability-of-censoring weights \code{G(t)} (global KM).}

  \item{g.floor.q}{Optional quantile-based floor (computed over the evaluation grid) to further protect tails.}

  \item{power}{Exponent in the Uno-style time weight, effectively using weights proportional to \code{1 / G(t)^{power}}.
  The classical choice is \code{power = 2}.}

  \item{ydata}{Counting-process data frame \code{(id, start, stop, event, ...)}.
  Required when \code{object} is a predict object that does not carry \code{id}/\code{yvar}.}

  \item{winsor.q}{Optional winsorization quantile (e.g., \code{0.99})
    applied to the time weights to minimize extreme influence.}

  \item{eps}{Small positive constant for numerical stability in weight
    calculations.}

  \item{bootstrap.rep}{Number of bootstrap replicates for standard errors. Default is \code{0} (no bootstrap).}

  \item{bootstrap.refit}{Logical. If \code{FALSE} (default), uses a subject-level pairs bootstrap with the
  fitted marker held fixed ("plug-in"). If \code{TRUE}, performs a \emph{full refit} at each replicate,
  using the original RHF tuning parameters.}

  \item{bootstrap.conf}{Confidence level for pointwise normal bands on AUC(t) when \code{bootstrap.rep > 0}
    (e.g., \code{0.95}). Set to \code{NA} to suppress bands.}

  \item{verbose}{Logical; if \code{TRUE}, enable bootstrap progress messages.
  Messages can be silenced with \code{suppressMessages()}.}

  \item{bootstrap.seed}{Optional integer seed used before bootstrap resampling for reproducibility.}
  
  \item{x, digits, max.rows, bass, xlab, ylab, main, ylim, pch, alpha, \dots}{
    Standard arguments for the \code{print} and \code{plot} methods.}
}

\details{
\strong{What is estimated.}
At each evaluation time \eqn{t}, the function computes an AUC for the chosen marker
(either cumulative hazard or hazard), using the specified case/control definition:
\itemize{
  \item \emph{Incident/dynamic} (\code{method = "incident"}): cases fail at \eqn{t}; controls are at risk at \eqn{t}.
  \item \emph{Cumulative/dynamic} (\code{method = "cumulative"}): cases have failed by \eqn{t}; controls are event-free at \eqn{t}.
}

The per-time AUC(t) values are combined in two ways:
\itemize{
  \item \code{iAUC.uno}: an inverse-probability-of-censoring weighted (IPCW) average of AUC(t) over time,
        using a global KM estimate \code{G(t)} and a weight proportional to \code{1 / G(t)^{power}}
        (Uno-style time weighting). Stabilization parameters include \code{g.floor}, \code{g.floor.q}, and \code{winsor.q}.
  \item \code{iAUC.std}: a simple time-standardized mean of AUC(t) over the evaluation grid (trapezoidal rule divided by span).
        This quantity is sensitive to the time window and the shape of AUC(t) across time.
}

\strong{Markers.}
For \code{method = "cumulative"}, using \code{marker = "cumhaz"} matches standard
time-dependent AUCs. Using \code{marker = "hazard"} with \code{"cumulative"}
often yields smaller AUC(t), as it targets instantaneous risk rather than cumulative risk.

\strong{Risk-set choice.}
When \code{method = "incident"} there are two ways to define controls at time \code{t}:
\itemize{

  \item \code{riskset = "subject"}: at risk when \code{entry < t <= Tstop} (one interval per subject).
    Fast and appropriate when subject rows tile continuous follow-up (no gaps).
  \item \code{riskset = "record"}: at risk when a counting-process row satisfies \code{start < t <= stop}.
        Exact under gaps but may yield fewer controls at sparse times.

}
If rows tile the entire follow-up, the two definitions coincide (up to ties). With gaps,
\code{"subject"} can over-include controls (lower variance but a conceptual mismatch),
while \code{"record"} is exact but can be sparser. For \code{method = "cumulative"} the risk-set choice is ignored.

\strong{Bootstrap.}
When \code{bootstrap.rep > 0}, two modes are available:
\itemize{
  \item \emph{Plug-in} (\code{bootstrap.refit = FALSE}): subject-level pairs bootstrap holding the fitted marker matrix fixed.
        Weights the within-time AUC(t) using resampled multiplicities and recomputes Uno's time weights via a weighted KM.
  \item \emph{Refit} (\code{bootstrap.refit = TRUE}): for each replicate, resample subjects, \emph{refit} the RHF using the original
        tuning parameters, and re-evaluate AUC(t). The returned per-time SEs are matched to the original
        evaluation grid. For predict objects, the refit is performed using the data carried in the object; predictions are then
        recomputed for those data. This is computationally heavier but captures variability from model fitting.
}
The \code{plot} method draws a base-R shaded band for AUC(t) when bootstrap SEs are present.
}

\value{
An object of class \code{"auct.rhf"} with elements:
\itemize{
  \item \code{AUC.by.time}: data frame with columns \code{time}, \code{AUC}, \code{n.cases}, \code{n.ctrl}, \code{G}, \code{W}.
  \item \code{iAUC.uno}: IPCW (Uno-style) time-averaged AUC.
  \item \code{iAUC.std}: time-standardized mean of AUC(t).
  \item \code{marker}, \code{method}, \code{riskset}, \code{power}, \code{g.floor}, \code{g.floor.q}, \code{winsor.q}, \code{times}: metadata.
  \item \code{diag.riskset}: Quick diagnostic comparing control counts under
        \code{riskset = "subject"} vs \code{"record"} on a small subset of times (list with
        \code{times}, \code{n.ctrl.subject}, \code{n.ctrl.record}, \code{n.diff}, \code{prop.times.different}).
  \item \code{boot}: present when \code{bootstrap.rep > 0} with
        \code{AUC.se} (per-time SE matched to the reporting grid), \code{iAUC.uno.se}, \code{iAUC.std.se},
        \code{conf.level}, \code{AUC.lower}, \code{AUC.upper}, \code{rep}, and \code{mode} (\code{"plug-in"} or \code{"refit"}).
}
Generic methods \code{print} and \code{plot} are provided for compact summaries and visualization.
}

\references{
Heagerty, P. J., Lumley, T., and Pepe, M. S. (2000).
Time-dependent ROC curves for censored survival data and a diagnostic marker.
\emph{Biometrics}, 56(2), 337--344.

Heagerty, P. J., and Zheng, Y. (2005).
Survival model predictive accuracy and ROC curves.
\emph{Biometrics}, 61(1), 92--105.

Uno, H., Tian, L., Cai, T., Kohane, I. S., Wei, L.-J. (2013).
A unified inference procedure for a class of measures to assess improvement in risk prediction systems with survival data.
\emph{Statistics in Medicine}, 32(14), 2430--2442.
}

\seealso{
  \code{\link{rhf}},
  \code{\link{predict.rhf}}
}

\author{
Hemant Ishwaran and Udaya B. Kogalur
}

\examples{
\donttest{

## ------------------------------------------------------------
## Peak VO2 example
## ------------------------------------------------------------

data(peakVO2, package = "randomForestSRC")
d <- convert.counting(Surv(ttodead, died) ~ ., peakVO2)
f <- "Surv(id, start, stop, event) ~ ."

o <- rhf(f, d, ntree = 25, nodesize = 5)

## AUC(t) with cumulative/dynamic definition and cumhaz marker
a.chf <- auct.rhf(o, marker = "cumhaz", method = "cumulative")

## AUC(t) with incident/dynamic definition and hazard marker
a.haz <- auct.rhf(o, marker = "hazard", method = "incident")

print(a.chf)
print(a.haz)

oldpar <- par(mfrow = c(1, 2))
plot(a.chf, main = "AUC(t): cumulative + cumhaz")
plot(a.haz, main = "AUC(t): incident + hazard")
par(oldpar)

## ------------------------------------------------------------
##  TDC illustration with training/testing
## ------------------------------------------------------------

trn <- hazard.simulation(1)$dta
tst <- hazard.simulation(1)$dta
f <- "Surv(id, start, stop, event) ~ ."
o <- rhf(f, trn, ntree = 25)
p <- predict(o, tst)
a.trn <- auct.rhf(o)
a.tst <- auct.rhf(p)

oldpar <- par(mfrow = c(1, 2))
plot(a.trn, main = "AUC(t): chf marker, train", ylim = c(0.5,1))
plot(a.tst, main = "AUC(t): chf marker, test",  ylim = c(0.5,1))
par(oldpar)

## ------------------------------------------------------------
## Bootstrap SEs and shaded band
## ------------------------------------------------------------
d <- hazard.simulation(1)$dta
f <- "Surv(id, start, stop, event) ~ ."
o <- rhf(f, d)


oldpar <- par(mfrow = c(1, 2))


## plug-in bootstrap
a.bs1 <- auct.rhf(o, marker = "cumhaz", method = "cumulative",
                  bootstrap.rep = 20, bootstrap.seed = 123)
plot(a.bs1, main = "AUC(t) with bootstrap band (plug-in)")

## refit bootstrap (can be slow)
a.bs2 <- auct.rhf(o, marker = "cumhaz", method = "cumulative",
                  bootstrap.rep = 10, bootstrap.refit = TRUE, bootstrap.seed = 7)
plot(a.bs2, main = "AUC(t) with bootstrap band (refit)")


par(oldpar)
}
}

\keyword{survival}
\keyword{ROC}
\keyword{AUC}

\name{tune.treesize.rhf}
\alias{tune.treesize.rhf}
\alias{tune.rhf}
\alias{tune.iAUC.rhf}
\alias{tune.iAUC}
\alias{plot.tune.treesize.rhf}
\title{Tune tree size for Random Hazard Forests}

\description{

  Tune the \code{treesize} for a Random Hazard Forest using either
  out-of-bag (OOB) empirical risk or an OOB integrated AUC computed by
  \code{\link{auct.rhf}}.

}

\usage{
tune.treesize.rhf(
  formula,
  data,
  ntree = 500,
  nsplit = 10,
  nodesize = NULL,
  perf = c("risk", "iAUC"),
  auct.args = NULL,
  lower = 2L,
  upper = NULL,
  C = 3,
  method = c("golden", "bisect"),
  max.evals = 20L,
  bracket.tol = 2L,
  seed = NULL,
  verbose = TRUE,
  forest = TRUE,
  ...
)


tune.iAUC.rhf(
  formula,
  data,
  auct.args = NULL,
  ...
)

\method{plot}{tune.treesize.rhf}(x, ylab = NULL, main = NULL,
    se.band = TRUE, se.mult = 1, ylim = NULL, \dots)
}


\arguments{
  
  \item{formula}{Survival formula for \code{\link{rhf}}, typically
    \code{Surv(id, start, stop, event) ~ .}.}

  \item{data}{Counting-process data frame containing the variables in
    \code{formula}.}

  \item{ntree}{Number of trees grown at each candidate \code{treesize}.}

  \item{nsplit}{Number of random split points tried per variable.}

  \item{nodesize}{Minimum number of events in a terminal node; passed to
    \code{\link{rhf}}.}

  \item{perf}{Tuning criterion. \code{"risk"} minimizes mean OOB risk;
    \code{"iAUC"} maximizes integrated AUC obtained using
    \code{\link{auct.rhf}}.}

  \item{auct.args}{Optional list of arguments passed to
    \code{\link{auct.rhf}} when \code{perf = "iAUC"}. }

  \item{lower}{Smallest \code{treesize} considered.}

  \item{upper}{Largest \code{treesize} considered.} 

  \item{C}{Multiplier used when computing the default upper bound if
    \code{upper} is \code{NULL}.}

  \item{method}{Discrete search strategy: \code{"golden"} (golden-section
    search) or \code{"bisect"} (bisection-style search).}

  \item{max.evals}{Maximum number of distinct \code{treesize} values
    evaluated by the search.}

  \item{bracket.tol}{Tolerance for the bracketing interval width used by
    the search algorithms.}

  \item{seed}{Optional random seed for reproducible comparisons across
    tree sizes.}

  \item{verbose}{Logical; if \code{TRUE}, print progress messages when
    evaluating new \code{treesize} values.}

  \item{forest}{Logical; if \code{TRUE}, return the RHF object at the
    best \code{treesize}.}

  \item{\dots}{Additional arguments passed to \code{\link{rhf}}.}

  \item{x, ylab, main, se.band, se.mult, ylim}{Arguments to
     customize \code{plot}.}
   
}

\details{

The alias \code{tune.rhf} is provided for convenience.

  
Rpeatedly fits a hazard forest with different values of \code{treesize}
and selects the value that optimizes the chosen criterion
\code{perf}. When \code{perf = "risk"}, the criterion is the mean OOB
risk. When \code{perf = "iAUC"}, \code{\link{auct.rhf}} is applied to
each fit and the criterion is \code{1 - iAUC.uno}.  \code{tune.iAUC} is
a wrapper around \code{tune.rhf} with \code{perf = "iAUC"}, using
default \code{auct.rhf} settings when \code{auct.args = NULL}.

}

\value{
An object of class \code{"tune.treesize.rhf"} with components
\itemize{
  \item \code{best.size}: optimal \code{treesize};
  \item \code{best.err}: minimal criterion value (mean OOB risk if
    \code{perf = "risk"}, or \code{1 - iAUC.uno} if \code{perf = "iAUC"});
  \item \code{path}: data frame with evaluated \code{treesize} values and
    corresponding criterion values (and an \code{iAUC} column when
    \code{perf = "iAUC"});
  \item \code{forest}: RHF fit at \code{best.size} when
    \code{forest = TRUE}.}
}

\seealso{
\code{\link{rhf}}, \code{\link{auct.rhf}}
}

\examples{
\donttest{
## ------------------------------------------------------------
## Example: tuning treesize by OOB empirical risk
## ------------------------------------------------------------

set.seed(7)
d <- hazard.simulation(1)$dta
f <- "Surv(id, start, stop, event) ~ ."

## tune.rhf is an alias for tune.treesize.rhf
tuneRisk <- tune.rhf(f, d)  ## same as tune.treesize.rhf(f, d)

oldpar <- par(mfrow = c(1, 1))
plot(tuneRisk)
par(oldpar)

tuneRisk$best.size    # treesize minimizing mean OOB risk
tuneRisk$best.err

## Access the tuned forest and refit AUC diagnostics if desired
best.forest <- tuneRisk$forest
a.risk <- auct.rhf(best.forest)
print(a.risk$iAUC.uno)


## ------------------------------------------------------------
## Example: tuning treesize by iAUC
## ------------------------------------------------------------

set.seed(7)
tuneiAUC <- tune.iAUC(f, d)

oldpar <- par(mfrow = c(1, 1))
plot(tuneiAUC)
par(oldpar)

print(tuneiAUC$best.size)
print(1 - tuneiAUC$best.err)

best.forest.iauc <- tuneiAUC$forest
a.iauc <- auct.rhf(best.forest.iauc)
print(1 - a.iauc$iAUC.uno)


## ------------------------------------------------------------
## Example: tuning treesize by iAUC with bootstrap s.e.
## ------------------------------------------------------------

set.seed(7)
tuneiAUC.boot <- tune.iAUC(f, d, auct.args=list(bootstrap.rep=25))
oldpar <- par(mfrow = c(1, 1))
plot(tuneiAUC.boot)        
par(oldpar)

## ------------------------------------------------------------
## Example: tuning by hazard-based incident AUC
## ------------------------------------------------------------

set.seed(7)
tuneHazInc <- tune.iAUC(
  f, d,
  auct.args = list(
    marker       = "hazard",
    method       = "incident"
  )
)

oldpar <- par(mfrow = c(1, 1))
plot(tuneHazInc)
par(oldpar)

print(tuneHazInc$best.size)
print(tuneHazInc$best.err)

## Confirm the tuned incident AUC on the selected forest
best.forest.haz <- tuneHazInc$forest
aHazInc <- auct.rhf(
  best.forest.haz,
  marker  = "hazard",
  method  = "incident"
)
print(aHazInc)
print(1 - aHazInc$iAUC.uno)

}}

\keyword{survival}
\keyword{tuning}

\name{rhf}
\alias{rhf}
\title{Random Hazard Forests}
\description{

  Random Hazard Forest (RHF) is a tree-ensemble survival method that
  estimates case-specific hazards and cumulative hazard functions on a
  working time grid and supports time-dependent covariates using
  counting-process (start/stop) format.

}

\usage{
rhf(formula,
      data,
      ntree = 500,
      nsplit = 10,
      treesize = NULL,
      nodesize = NULL,
      block.size = 10,
      bootstrap = c("by.root", "none", "by.user"),
      samptype = c("swor", "swr"),
      samp = NULL,
      case.wt = NULL,
      membership = TRUE,
      sampsize = if (samptype == "swor") function(x){x * .632} else function(x){x},
      xvar.wt = NULL,
      ntime = 50,
      min.events.per.gap = 10,
      seed = NULL,
      do.trace = FALSE,
      ...)
}

\arguments{

  \item{formula}{Model formula specifying the response and predictors.}

  \item{data}{Data frame containing the variables referenced in \code{formula}.}

  \item{ntree}{Number of trees to grow in the ensemble.}

  \item{nsplit}{Non-negative integer specifying the number of random
  split points to consider for each variable. When set to 0, all
  possible split points are evaluated (deterministic splitting), which
  may be slower. The default is 10.}

  \item{treesize}{Specifies the desired tree size, defined as the number
  of terminal nodes. Can be provided as a function of sample size
  \code{n}. If unspecified, an internal default is used.}

  \item{nodesize}{Minimum number of cases required in a terminal
    node. If not specified, an internal default is used.}
  
  \item{block.size}{Controls the granularity of cumulative error rate
  reporting. Setting this to an integer between 1 and \code{ntree}
  yields the cumulative error every \code{block.size} trees.}

  \item{bootstrap}{Bootstrap strategy for generating inbag samples. The
  default is \code{"by.root"}, which bootstraps by sampling with or
  without replacement (default is without; see \code{samptype}). If set
  to \code{"none"}, no bootstrapping is performed (OOB predictions and
  errors are then unavailable). The option \code{"by.user"} uses a
  user-defined bootstrap specified by \code{samp}.}

  \item{samptype}{Sampling type used when \code{bootstrap =
  "by.root"}. Options are \code{"swor"} (sampling without replacement,
  default) and \code{"swr"} (sampling with replacement).}

  \item{samp}{User-specified bootstrap when \code{bootstrap =
  "by.user"}. Should be an \code{n} by \code{ntree} array, where each
  entry gives the number of times a case appears in the inbag sample for
  a given tree.}

  \item{case.wt}{Non-negative vector of case weights (not required to
  sum to 1). Higher weights increase the probability of an observation
  being selected during bootstrapping or subsampling. Using real-valued
  weights is preferred over integer weights.}

  \item{membership}{Logical flag indicating whether to return terminal
    node membership and inbag information.}
  
  \item{sampsize}{Specifies the size of the bootstrap sample when
  \code{bootstrap = "by.root"}. For sampling without replacement, it
  represents the requested subsample size (default is 0.632 times the
  sample size). For sampling with replacement, it equals the sample
  size. Can be supplied as a numeric value or a function.}

  \item{xvar.wt}{Non-negative vector of variable selection probabilities
  for splitting. Values do not need to sum to 1. Defaults to uniform
  selection probability.}


  \item{ntime}{Controls the working time grid used for all
   ensemble calculations. Can be:
   \describe{

     \item{\code{ntime >= 1} (integer):}{Request approximately \code{ntime} grid
        points chosen from the observed event times. When used with
       \code{min.events.per.gap}, grid points are selected adaptively so that each
       interval contains at least \code{min.events.per.gap} events (when possible);
       therefore the resulting grid may contain fewer than \code{ntime} points when
       events are sparse (especially in the tail).}

     \item{numeric vector:}{A user-supplied set of time points. Each value is
       aligned (snapped) to the nearest observed event time and duplicates are
       removed.}

     \item{\code{0} or \code{NULL}:}{Use all observed (unique) event times.}

 }}
 
 \item{min.events.per.gap}{Minimum number of observed events required in
   each time interval when \code{ntime} is specified as an integer.
   Together with ntime, this provides an automatic event-balanced grid
   that allocates more resolution where events are dense and avoids 
   sparse tail intervals.}

  \item{seed}{Negative integer setting the random seed for reproducibility.}

  \item{do.trace}{Time in seconds between progress updates printed to the console.}

  \item{...}{Additional arguments (currently unused).}

}

\details{
\code{rhf()} grows an ensemble of hazard trees for survival
outcomes specified in counting-process notation
\code{Surv(id, start, stop, event)}. Predictors may include both time-static
variables and time-dependent covariates (TDCs). The fitted object contains
case-specific ensemble estimates of the hazard and cumulative hazard on the
working time grid \code{time.interest}.

Random Hazard Forests (RHF) extends Random Survival Forests (RSF) in two
key ways: (1) it directly estimates the hazard function, and (2) it
accommodates time-dependent covariates.

Tree construction uses best-first splitting (BFS) guided by empirical
risk reduction. The risk is based on a smooth convex surrogate for the
nonparametric log-likelihood functional, as described in Lee, Chen, and
Ishwaran (2021).

Splits can occur on both static and time-dependent covariates (TDCs),
but not on time itself. Time splitting is unnecessary, as terminal node
hazard estimators already incorporate time appropriately. To encourage
selection of time-dependent covariates, use the \code{xvar.wt} option to
weight their importance.


The case-specific hazard estimator uses a "stitched" active-record rule
on the \code{time.interest} grid: each record is treated as active from
just after its start time up to (and including) the next record's start
time, with the final record extended to the maximum time in
\code{time.interest}.


\strong{Data Format:} Input data must follow standard counting process
format and include the following four columns (column names must match
those specified in the counting process formula; see examples):


\enumerate{

  \item \code{id}: A unique integer identifier for each
  individual. Repeated for multiple rows corresponding to that
  individual.
  

  \item \code{start}: The start time for each interval. All time values
  must be scaled to the interval [0, 1].
  

  \item \code{stop}: The stop time for each interval. Must satisfy
  \code{stop > start}.
  

  \item \code{event}: A binary event indicator (0 = censored, 1 =
  event). Only one event is allowed per individual; otherwise, the
  individual is treated as censored.
  

}

Use the helper function \code{convert.counting} to convert conventional
survival data to counting process format. See the examples section for
further illustration and guidance on data preparation.


}

\value{

  An object of class \code{"rhf"} containing the fitted forest and related
  results. Components include (among others) the working time grid
  \code{time.interest} and case-specific ensemble estimates of the hazard
  and cumulative hazard. When bootstrapping is used, these are typically
  returned as \code{hazard.oob}/\code{chf.oob} (out-of-bag) and
  \code{hazard.inbag}/\code{chf.inbag} (in-bag). Use \code{\link{predict.rhf}}
  to obtain estimates for new data.

}

\author{

  Hemant Ishwaran and Udaya B. Kogalur

}

\references{
  
  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
  
  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.

  Lee, D.K. and Chen N. and Ishwaran H (2021). Boosted nonparametric
  hazards with time-dependent covariates. \emph{Annals of Statistics},
  49: 2101-2128.

  Ishwaran H. (2025). Multivariate Statistics: Classical Foundations and
  Modern Machine Learning.  Chapman and Hall.

  Ishwaran H., Kogalur U.B., Hsich E.M. and Lee D.K. (2026).  Random
  hazard forests.
  
}

\seealso{
  \code{\link{auct.rhf}},
  \code{\link{plot.rhf}},
  \code{\link{predict.rhf}},
  \code{\link{smoothed.hazard.rhf}},
  \code{\link{tune.treesize.rhf}}
}


\examples{

  
## ------------------------------------------------------------
##  time-static pbc: parameters set for fast CRAN run
## ------------------------------------------------------------

## load the data
data(pbc, package = "randomForestSRC")
## convert the data to counting process
d <- convert.counting(Surv(days, status) ~ ., na.omit(pbc))
## set the formula
f <- "Surv(id, start, stop, event) ~ ."

## rhf call
print((o <- rhf(f, d, ntree = 3)))

## smooth hazard estimator for specific cases
plot(o, idx=c(1,5,10))
plot(o, idx=c(1,5,10), hazard.only=TRUE)
plot(o, idx=c(1,5,10), hazard.only=TRUE, lwd=0)
plot(o, idx=1:10, lwd=0, hazard.only=TRUE, legend.show=FALSE)

\donttest{

## ------------------------------------------------------------
##  time-static pbc: compare RHF to RSF
## ------------------------------------------------------------

data(pbc, package = "randomForestSRC")
d <- convert.counting(Surv(days, status) ~ ., na.omit(pbc))

## first we run rhf
o.rhf <- rhf("Surv(id, start, stop, event) ~ .", d)
h <- o.rhf$hazard.oob
time <- o.rhf$time.interest
delta <- c(0, diff(time))
H <- apply(h, 1, function(x) {cumsum(delta * x)})
S.rhf <- exp(-H)

## next we run rsf, using the same time.interest grid
o.rsf <- randomForestSRC::rfsrc(Surv(days, status) ~ ., pbc, ntime = time)
S.rsf <- t(o.rsf$survival.oob)

## graphical parameters
bass <- 3
oldpar <- par(mfrow=c(3,2))

## plot survival results 
matplot(time,S.rhf,pch=16)
matplot(time,S.rsf,pch=16)
matplot(time,S.rhf-S.rsf,pch=16)
boxplot(S.rhf-S.rsf,ylab="S.rhf-S.rsf",xaxt="n",outline=FALSE,range=1e-10)
abline(h=0,lwd=3,col=2)

## plot chf and hazard results
matplot(time,H,ylab="CHF",pch=16)
matplot(time,t(h),ylab="hazard",pch=16,type="n",ylim=c(0,quantile(c(h),.95)))
nO <- lapply(1:nrow(h), function(i) {
  lines(supsmu(time, h[i,], bass=bass),type="s",col=gray(.4))
})

par(oldpar)

## ------------------------------------------------------------
##  TDC illustration (using built in hazard simulation function)
## ------------------------------------------------------------

d1 <- hazard.simulation(1)$dta
d2 <- hazard.simulation(2)$dta
d3 <- hazard.simulation(3)$dta
f <- "Surv(id, start, stop, event) ~ ."

o1 <- rhf(f, d1)
o2 <- rhf(f, d2)
o3 <- rhf(f, d3)

plot(o1,idx=4)
plot(o2,idx=1:3)
plot(o3,idx=2)


## ------------------------------------------------------------
##  peakVO2: demonstrates time-varying auc
## ------------------------------------------------------------

data(peakVO2, package = "randomForestSRC")
d <- convert.counting(Surv(ttodead, died)~., peakVO2)
f <- "Surv(id, start, stop, event) ~ ."

## build the forest
o1 <- rhf(f, d, nodesize=1)
o2 <- rhf(f, d, nodesize=15)

## acquire auc-t
a1.chf <- auct.rhf(o1) ## same as auct.rhf(o1, marker = "chf")
a1.haz <- auct.rhf(o1, marker = "haz")
a2.chf <- auct.rhf(o2) ## same as auct.rhf(o2, marker = "chf")
a2.haz <- auct.rhf(o2, marker = "haz")

## print auc-t
print(a1.chf)
print(a2.chf)
print(a1.haz)
print(a2.haz)

## plot auc-t
oldpar <- par(mfrow=c(2,2))
plot(a1.chf, main="nodesize 1")
plot(a1.haz, main="nodesize 1")
plot(a2.chf, main="nodesize 15")
plot(a2.haz, main="nodesize 15")
par(oldpar)

## ------------------------------------------------------------
##  more detailed example of time-varying performance 
## ------------------------------------------------------------

d <- hazard.simulation(1)$dta
f <- "Surv(id, start, stop, event) ~ ."
treesize <- c(10, 30, 100)

oldpar <- par(mfrow=c(3,2))
perf <- lapply(treesize, function(tz) {
  o <- rhf(f, d, treesize=tz, nodesize=1)
  print(o)
  rO <- list(chf = auct.rhf(o, marker="chf"),
             haz = auct.rhf(o, marker="haz"))
  plot(rO$chf, main=paste("chf marker, treesize=", tz))
  plot(rO$haz, main=paste("hazard marker, treesize=", tz))
  rO
})
par(oldpar)

## ------------------------------------------------------------
##  example illustrating tuning the tree size
##  using function 'rhf.tune.treesize'
## ------------------------------------------------------------

d <- hazard.simulation(1)$dta
f <- "Surv(id, start, stop, event) ~ ."

tune <- tune.rhf(f, d)  ## same as rhf.tune.treesize(f,d)
oldpar <- par(mfrow=c(1,1))
plot(tune)
par(oldpar)

## ------------------------------------------------------------
##  example illustrating guided feature selection
##  using built in 'xvar.wt.rhf' helper
## ------------------------------------------------------------

d <- hazard.simulation(1)$dta
f <- "Surv(id, start, stop, event) ~ ."

o <- rhf(f,d)
o.gfs <- rhf(f,d, xvar.wt = xvar.wt.rhf(f, d))

print(o)
print(o.gfs)

}
}


\keyword{rhf}

\name{plot.rhf}
\alias{plot.rhf}
\title{Plot smoothed hazard and cumulative hazard plots from RHF analysis}

\description{
  Plot case specific hazard and cumulative hazard (CHF) from a fitted
  random hazard forest (RHF) object. Hazards are smoothed with
  \code{stats::supsmu}. Optional scaling places hazards on the same
  scale as the CHF for easier comparison.
}

\usage{\method{plot}{rhf}(x, idx = NULL, scale.hazard = FALSE, 
         ngrid = 30, bass = 0, q = 0.99, grid = FALSE,
         col = NULL, lty = NULL, legend.loc = "topright",
         jitter.factor = 1, lwd = 4,
         hazard.only = TRUE, legend.show = TRUE, ...)
}

\arguments{

  \item{x}{An \code{rhf} fit object returned by \code{\link{rhf}}.}

  \item{idx}{Subject identifiers selecting which cases to plot. Values must
  match those in \code{unique(x$id)}; by default the first value is used.}

  \item{scale.hazard}{Logical or numeric. If \code{TRUE}, hazards are
  multiplied by the local time step so they approximate CHF increments.
  If numeric, hazards are multiplied by the given constant.}

  \item{ngrid}{Number of equally spaced points used when \code{grid = TRUE}.}

  \item{bass}{Smoothing parameter passed to \code{stats::supsmu}.}

  \item{q}{Quantile used to trim extreme hazard values before smoothing.}

  \item{grid}{If \code{TRUE}, overlays smoothed CHF and its derivative.}

  \item{col}{Colors for selected cases. Recycled as needed.}

  \item{lty}{Line types for hazard and CHF curves.}

  \item{legend.loc}{Legend location. Passed to \code{\link[graphics]{legend}}.}

  \item{jitter.factor}{Amount of horizontal jitter applied to hazard spikes.}

  \item{lwd}{Line width for hazard spikes.}

  \item{hazard.only}{If \code{TRUE}, only the hazard is drawn.}

  \item{legend.show}{If \code{TRUE}, show a legend when plotting multiple cases.}
  
  \item{...}{Additional graphics parameters passed to \code{\link[graphics]{plot}}.}

}

\details{
  
The function displays out of bag hazard and CHF estimates on the grid
\code{x$time.interest}. Hazards are smoothed using
\code{\link[stats]{supsmu}} after trimming large values for stability.
Scaling the hazard allows direct visual comparison with CHF.
Plots are drawn to the current device; no model re-estimation is
performed.

}

\value{
  Invisibly returns \code{NULL}. The function is used for its plotting side effects.
}

\author{
  Hemant Ishwaran and Udaya B. Kogalur
}

\references{

  Ishwaran H. and Kogalur U.B. (2007).
  Random survival forests for R.
  \emph{R News}, 7(2): 25--31.

  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S. (2008).
  Random survival forests.
  \emph{Annals of Applied Statistics}, 2: 841--860.

  Lee D.K., Chen N., and Ishwaran H. (2021).
  Boosted nonparametric hazards with time dependent covariates.
  \emph{Annals of Statistics}, 49: 2101--2128.

  Ishwaran H., Lee D.K. and Hsich E.M. (2025).
  Random hazard forests.
}

\seealso{
  \code{\link{hazard.simulation}},
  \code{\link{predict.rhf}},
  \code{\link{rhf}},
  \code{\link{smoothed.hazard.rhf}}
}

\examples{
\donttest{

## ------------------------------------------------------------
##  time static peakVO2
## ------------------------------------------------------------
data(peakVO2, package = "randomForestSRC")
d <- convert.counting(Surv(ttodead, died)~., peakVO2)
f <- "Surv(id, start, stop, event) ~ ."

set.seed(1)
o <- rhf(f, d)

oldpar <- par(mfrow=c(1,1))

## plot selected cases
ids <- o$ensemble.id[1:3]
plot(o, idx = ids)

## hazard only
plot(o, idx = ids)

## scaled hazard with CHF
plot(o, idx = ids, scale.hazard = TRUE, hazard.only = FALSE)

## auxiliary grid with smoother control
plot(o, idx = ids, grid = TRUE, ngrid = 60, bass = 2, hazard.only = FALSE)

## multiple cases, no legend
plot(o, idx = o$ensemble.id[1:10], lwd = 0, legend.show = FALSE)

## lowess median smoothed hazard
s <- smoothed.hazard(o)  ## default method="median.loess"
plot(s, idx = o$ensemble.id[1:10])
plot(s, idx = o$ensemble.id[1:10], lwd = 0)


par(oldpar)

## ------------------------------------------------------------
##  complex simulated time dependent covariate hazards
## ------------------------------------------------------------
f <- "Surv(id, start, stop, event) ~ ."
sim2 <- hazard.simulation(2)$dta
o2 <- rhf(f, sim2)

oldpar <- par(mfrow=c(1,1))
plot(o2, 1)
par(oldpar)

## ------------------------------------------------------------
##  same complex simulation, but using smoothed hazard
## ------------------------------------------------------------
f <- "Surv(id, start, stop, event) ~ ."
sim2 <- hazard.simulation(2)$dta
o2 <- rhf(f, sim2)

s2.loess <- smoothed.hazard(o2, method="loess")
s2.med.loess <- smoothed.hazard(o2, method="median.loess")

oldpar <- par(mfrow=c(1,1))
plot(o2, 1)
plot(s2.loess, 1)
plot(s2.med.loess, 1)
plot(s2.med.loess, 1:10, lwd = 0)
par(oldpar)


}

}

\keyword{plot}
\keyword{rhf}

\name{importance.rhf}
\alias{importance.rhf}
\alias{varpro.cache.rhf}
\alias{varpro.cache}
\alias{print.importance.rhf}
\alias{as.data.frame.importance.rhf}
\alias{plot.importance.rhf}
\alias{dotmatrix.importance.rhf}
\alias{dotmatrix.importance}
\title{Time-localized VarPro importance for random hazard forests}

\description{
Computes time-localized variable importance for a random hazard forest
(RHF) by using VarPro (variable priority) importance and restricting
rule and near-miss memberships to pseudo-individuals whose start-stop
intervals overlap selected windows from the master time grid
\code{time.interest}. The working response used for VarPro importance is
now taken directly from the RHF object as the logarithm of the upstream
integrated hazard exposure, giving a fast localized view of how variable
importance evolves over time.
}

\usage{
importance.rhf(o,
 cache = NULL,
 time.index = NULL,
 trim = 0.1,
 sort = TRUE,
 max.rules.tree,
 max.tree,
 eps = 1e-6,
 y.external = NULL,
 verbose = FALSE,
 ...)

varpro.cache.rhf(o,
 max.rules.tree = 150L,
 max.tree = 150L,
 y.external = NULL,
 eps = 1e-6,
 verbose = FALSE)

\method{print}{importance.rhf}(x,
 top = 10L,
 rank.by = c("q90", "median", "mean", "max"),
 digits = 4L,
 scientific.threshold = 1e4,
 ...)

\method{as.data.frame}{importance.rhf}(x,
  row.names = NULL,
  optional = FALSE,
  format = c("long", "variable_by_time", "time_by_variable"),
  ...)

dotmatrix.importance.rhf(x,
 vars = NULL,
 top_n_union = 15L,
 variable.labels = NULL,
 time.labels = NULL,
 sort_by = c("q90", "sum", "max", "mean", "median", "alphabetical", "cluster", "none"),
 sort_abs = TRUE,
 transform = c("none", "log10"),
 color_by = c("value", "sign", "single", "none"),
 point_color = "steelblue4",
 value_colors = c("grey85", "steelblue4"),
 sign_colors = c("firebrick3", "grey90", "steelblue4"),
 cex.range = c(0.6, 3.2),
 size.cap = 0.99,
 color.cap = 0.99,
 alpha = 0.9,
 show.grid = TRUE,
 grid.col = "grey92",
 legend = TRUE,
 display.note = TRUE,
 xlab = "",
 ylab = "",
 main = "RHF time-localized VarPro importance",
 axis.cex = 0.7,
 var.cex = 0.7,
 time.label.srt = 45,
 save_plot = FALSE,
 out.file = "rhf_time_varpro_dotmatrix.pdf",
 width = 11,
 height = NULL,
 mar = NULL,
 legend.width = 0.7,
 ...)

\method{plot}{importance.rhf}(x,
 type = c("dotmatrix", "lines"),
 vars = NULL,
 top = 10L,
 rank.by = c("q90", "median", "mean", "max"),
 curve = c("step", "line", "lowess"),
 smooth.f = 2/3,
 display.cap = 0.99,
 display.note = TRUE,
 xlab = NULL,
 ylab = NULL,
 lty = 1,
 lwd = 2,
 ...)
}

\arguments{

  \item{o}{A RHF object with class \code{"rhf"}.}

  \item{cache}{Optional cache object returned by \code{varpro.cache()}.
  If \code{NULL}, the cache is built internally. Supplying a cache is
  useful when repeated calls are made.}

  \item{time.index}{Optional vector identifying which windows of the time
  grid \code{o$time.interest} are to be analyzed. This may be an integer index
  vector or a logical vector of length \code{length(o$time.interest)}. If
  omitted, all windows are used.}

  \item{trim}{Tuning parameter passed to the underlying VarPro
  importance workhorse. \code{trim} controls winsorized aggregation
  across trees.}

  \item{sort}{Logical. If \code{TRUE}, variables are ordered within each
  window in decreasing importance before the long-format output is assembled.}

  \item{max.rules.tree, max.tree}{Arguments controlling rule extraction when
  the cache is built.}

  \item{y.external}{Optional externally supplied working response. When
  \code{NULL}, the working response is built internally from the RHF object's
  integrated hazard exposure values.}

  \item{eps}{Nonnegative value added before taking the logarithm of the
  integrated hazard exposure when \code{y.external} is not supplied.}

  \item{verbose}{Logical. If \code{TRUE}, reports cache construction and
  per-window progress.}

  \item{x}{An object of class \code{"importance.rhf"}.}

  \item{top, rank.by}{Arguments used by \code{print()} and by
  \code{plot(type = "lines")}. Printing now ranks variables robustly over
  time by default using \code{rank.by = "q90"}. The line plot also uses
  \code{rank.by} when \code{vars} is omitted.}

  \item{digits, scientific.threshold}{Formatting controls for
  \code{print()}, used to keep very large importance values readable.}

  \item{row.names, optional}{Included for compatibility with
  \code{as.data.frame()}.}

  \item{format}{Output format for \code{as.data.frame()}. \code{"long"}
  returns the long-format table, \code{"variable_by_time"} returns a data
  frame whose rows are variables and columns are times, and
  \code{"time_by_variable"} returns the transpose with window metadata.}

  \item{type, vars, top_n_union}{Arguments controlling which
  variables are displayed and which plot is produced. \code{type = "dotmatrix"}
  gives the time-by-variable dot-matrix display; \code{type = "lines"} gives a
  line, step, or smoothed view for selected variables. When \code{vars} is
  omitted, the line plot chooses \code{top} variables using \code{rank.by},
  while the dot-matrix plot uses the union of the top
  \code{top_n_union} variables across time.}

  \item{curve, smooth.f, lty, lwd, display.cap, display.note}{Arguments for
  \code{type = "lines"}. \code{curve} chooses between step, ordinary line,
  and lowess-smoothed displays; \code{smooth.f} is passed to
  \code{stats::lowess()} when needed; \code{lty} and \code{lwd} control line
  type and line width. \code{display.cap} applies display-only quantile
  capping to stabilize the vertical scale in the presence of extreme spikes,
  and \code{display.note} toggles the on-plot note when capping is applied.
  The same \code{display.note} flag is also used by the dot-matrix plot.}

  \item{variable.labels, time.labels, sort_by, sort_abs}{Arguments
  controlling variable labeling and ordering in the dot-matrix plot. Variable
  labels may be supplied as a named vector or a two-column data frame.
  Variables may be ordered by robust aggregate importance, alphabetically,
  hierarchical clustering, or left in their existing order.}

  \item{transform, color_by, point_color, value_colors, sign_colors,
  cex.range, size.cap, color.cap, alpha, legend}{Arguments
  controlling dot size, color encoding, display-only quantile capping,
  transparency, and the optional right-side legend in the dot-matrix plot.}

  \item{xlab, ylab, main, axis.cex, var.cex, time.label.srt, show.grid,
  grid.col, mar, legend.width, width, height, save_plot, out.file}{Display,
  layout, and export options for the plotting helpers. By default the
  dot-matrix plot uses blank axis labels, draws light guide lines, computes
  margins automatically, and can optionally be written to file with
  \code{save_plot = TRUE}.}

  \item{...}{Additional arguments passed to internal calculations or plotting
  routines.}
}

\details{

This routine implements a \emph{fast localization} strategy for RHF
VarPro importance. The master time grid is taken from
\code{time.interest}. For a window corresponding to a selected grid
index, the method keeps only those pseudo-individuals whose start-stop
interval overlaps that window, while reusing the same sampled rules and
near-miss sets already obtained from the RHF fit.

For RHF objects the underlying VarPro importance calculation follows a
regression-style approach in which the working response is the logarithm
of the integrated hazard exposure, and local rule importance is computed
by comparing this working response in a rule versus its near-miss
set. Time localization is achieved by restricting those memberships
within each window rather than rebuilding the entire rule structure
repeatedly.

The helper \code{varpro.cache()} stores the minimum information needed
for repeated localized importance calculations: a regression-style rule
template, window metadata, the working response source, and precomputed
window-local rule statistics. During cache construction, raw OOB and
complementary memberships are converted into compact per-window rule
summaries, so the later window sweep does not need to rescan membership
vectors.

The returned importance matrix has variables in rows and selected time
windows in columns. Column names correspond to the right endpoints of
the selected windows. The long-format table contains the same values
together with window metadata such as start, stop, midpoint, number at
risk, and number of active rules.

Printing and plotting share a robust strategy.  Summaries default to a
robust over-time ranking based on the 90th percentile, and the plotting
helpers apply optional quantile capping for display only. This prevents
rare extreme spikes from flattening curves while preserving the original
importance matrix for downstream analyses.

}

\value{

\code{varpro.cache()} returns an object containing cached rule
memberships, the working response used for importance, start-stop
information for pseudo-individuals, time-window metadata, and the rule
extraction settings.

\code{importance.rhf()} returns a list including:

\itemize{
  \item \code{importance.matrix}: matrix of localized importance values with
  variables in rows and selected time windows in columns.
  \item \code{importance.long}: long-format data frame containing variable,
  time, window metadata, and localized importance.
  \item \code{window.info}: data frame describing the analyzed windows,
  including start, stop, midpoint, \code{n.risk}, and \code{n.rules}.
  \item \code{y.source}: source of the working response. This is
  \code{"int.haz.oob"}, \code{"int.haz.test"}, or \code{"y.external"}.
  \item \code{trim}: tuning value used in importance aggregation.
}

\code{print()} returns its input invisibly after displaying a short summary
that includes robust over-time summaries for the leading variables.

\code{as.data.frame()} returns one of the supported data-frame views.

\code{dotmatrix.importance.rhf()} produces a base-R dot-matrix plot and
returns plotting metadata invisibly.

\code{plot()} returns invisibly the result of the underlying plotting helper.
}


\seealso{
  \code{\link{rhf}}
}

\examples{
\donttest{
################################################################
##
## simulation model
##
################################################################

## draw simulation (can be modified)
n <- 400
p <- 10
simid <- 2
d <- hazard.simulation(type = simid, n = n, p = p, nrecords = 4)$dta

## fit a RHF model with weighted mtry (use for high-dimension)
f <- "Surv(id, start, stop, event) ~ ."
o <- rhf(f, d, ntree = 50, nsplit = 5, xvar.wt = xvar.wt.rhf(f, d))
print(o)

## time-localized RHF importance across the full time grid
imp.t <- importance.rhf(o)
print(imp.t)

## extract the variable-by-time matrix
print(head(imp.t$importance.matrix))

oldpar <- par(mfrow=c(1,1))

## dot-matrix importance plot (default)
plot(imp.t)

## step-style importance line plot for the top variables
## (ranked by the 90th percentile over time and display-capped at q99)
plot(imp.t, type = "lines", top = 10)

## smoothed importance plot for all variables with display capping
plot(imp.t, type = "lines", curve = "lowess", smooth.f = 0.5,
     display.cap = 0.95)

## dot-matrix plot with robust ordering and display capping
plot(imp.t, sort_by = "q90", size.cap = 0.99, color.cap = 0.99)

par(oldpar)


## reuse a cache for repeated calls on subsets of the time grid
cache <- varpro.cache(o)
imp.t.sub <- importance.rhf(
  o,
  cache = cache,
  time.index = seq(1, length(o$time.interest), by = 5),
  verbose = TRUE
)

## long-format export
print(head(as.data.frame(imp.t.sub)))
}
}

\keyword{rhf importance}

\name{smoothed.hazard.rhf}
\alias{smoothed.hazard.rhf}
\alias{smoothed.hazard}
\title{Smoothed hazard and cumulative hazard estimates from a random hazard forest}

\description{

Compute smoothed case-specific hazard curves (and corresponding
cumulative hazards) from a Random Hazard Forest (RHF) object. The
returned object is a lightweight list designed to be directly usable by
\code{\link{plot.rhf}}.

}

\usage{
smoothed.hazard.rhf(
  o,
  method = c("median.loess", "loess"),
  oob = TRUE,
  span = 0.35,
  degree = 1,
  family = NULL,
  eps = 1e-12,
  suppress.warnings = TRUE,
  trace = FALSE
)
}

\arguments{

  \item{o}{A fitted RHF object or predict object.}


  \item{method}{
    Smoothing strategy.
    See \bold{Details} for definitions of \code{"loess"} and \code{"median.loess"}.
  }

  \item{oob}{By default out-of-bag (OOB) values are returned for RHF
    objects, but inbag values can be requested by setting this option
    to \code{FALSE}.  Does not apply to RHF predict objects.}

  \item{span}{Loess span passed to \code{\link[stats]{loess}}.}

  \item{degree}{Loess degree passed to \code{\link[stats]{loess}}.}

  \item{family}{
    Loess family passed to \code{\link[stats]{loess}}.
    Must be \code{"gaussian"} or \code{"symmetric"}.
    If \code{NULL}, defaults to \code{"gaussian"} for \code{method="median.loess"}
    and \code{"symmetric"} for \code{method="loess"}.
  }

  \item{eps}{
    Small positive constant added before log-transforming hazards: \code{log(h + eps)}.
  }

  \item{suppress.warnings}{
    If \code{TRUE}, warnings arising from \code{loess()} fitting/prediction are suppressed.
  }

  \item{trace}{If \code{TRUE}, print progress messages.}
}

\details{

Hazard curves are computed/smoothed on the grid \code{o$time.interest}.
Smoothing is performed on the log-hazard scale: \code{log(hazard +
eps)}, and then back-transformed to the hazard scale.


\describe{ 


  \item{\code{method="loess"}}{
    Applies loess smoothing to an already-constructed hazard matrix
    stored in \code{o} (e.g. \code{o$hazard.oob} and/or
    \code{o$hazard.inbag}), independently for each subject.  }

  \item{\code{method="median.loess"}}{
    For each subject and time bin, computes the median across trees of
    the terminal-node (leaf) log-hazard \code{log(h + eps)} (using a
    native C routine), then applies loess smoothing to the resulting
    median log-hazard curve and back-transforms.  This method requires
    the fitted object to retain tree-specific hazard and membership
    structures (e.g., \code{o$forest$t.hazard} and related arrays).  }

}

}


\value{

A list containing:
  
\itemize{
  \item \code{id}, \code{ensemble.id}, \code{yvar}, \code{time.interest} copied from \code{o}.
  \item Depending on \code{target}:
  \itemize{
    \item \code{hazard.oob}, \code{chf.oob} (out-of-bag), matrices of dimension
    \code{nSubj x length(time.interest)}.
    \item \code{hazard.inbag}, \code{chf.inbag} (in-bag), matrices of dimension
    \code{nSubj x length(time.interest)}.
    \item \code{hazard.test}, \code{chf.test} (test), matrices of dimension
    \code{nSubjTest x length(time.interest)}.
  }
}
}


\seealso{
  \code{\link{plot.rhf}},
  \code{\link{predict.rhf}},
  \code{\link{rhf}}
}

\examples{
\donttest{

## ------------------------------------------------------------
## canonical example using synthetic data
## includes both train/test scenarios
## ------------------------------------------------------------

simID <- 1
trn <- hazard.simulation(simID)$dta
tst <- hazard.simulation(simID)$dta
f <- "Surv(id, start, stop, event) ~ ."

## training/testing
o <- rhf(f, trn)
p <- predict(o, tst)

## default: median(log leaf hazard) + loess
so <- smoothed.hazard(o)
sp <- smoothed.hazard(p)

## the returned object can be passed directly to plot.rhf()
oldpar <- par(mfrow=c(1,1))
plot.rhf(so)
plot.rhf(sp)
par(oldpar)

## ------------------------------------------------------------
##  peak vo2
## ------------------------------------------------------------

data(peakVO2, package = "randomForestSRC")
d <- convert.counting(Surv(ttodead, died)~., peakVO2)
f <- "Surv(id, start, stop, event) ~ ."

## training
o <- rhf(f, d)

oldpar <- par(mfrow=c(1,1))


## median loess
s0 <- smoothed.hazard(o) 
ids <- o$ensemble.id[1:3]
plot.rhf(s0, idx = ids)

## loess smoothing 
s1 <- smoothed.hazard(o, method = "loess")
plot.rhf(s1, idx = ids)

par(oldpar)
}
}

\keyword{survival}
\keyword{models}
